using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using UnityEngine.EventSystems;
using UnityEngine.Events;
using System.Security.AccessControl;

namespace JayJayWon
{
    public abstract class CustomUIElement
    {
        protected UIDynamic _spacerLeft { get; set; }
        protected UIDynamic _spacerRight { get; set; }
        protected UIDynamic _spacerActive { get; set; }

        protected int spacerHeight;
        public float height { get { return spacerHeight; } }
        protected HashSet<UIDynamicPopup> uidPopuplist;


        public CustomUIElement(int height)
        {
            spacerHeight = height;
            uidPopuplist = new HashSet<UIDynamicPopup>();
        }

        public void CloseAllPopups(UIDynamicPopup exception)
        {
            foreach (UIDynamicPopup popup in uidPopuplist)
            {
                if (popup != exception) popup.popup.visible = false;
            }
        }

        public virtual void ClearUI()
        {
            if (_spacerLeft != null) UIAGlobals.mvrScript.RemoveSpacer(_spacerLeft);
            if (_spacerRight != null) UIAGlobals.mvrScript.RemoveSpacer(_spacerRight);
            _spacerActive = null;
            uidPopuplist.Clear();
        }
    }
    public class CustomRowUIElement : CustomUIElement
    {
        private int buttonPadding = 0;
        private int borderWidth = 20;

        private Dictionary<int, UIDynamicButton> buttonDict;
        private bool activeButtonHighlight;

        public CustomRowUIElement(int height = 50, bool _activeButtonHighlight = true) : base(height)
        {
            activeButtonHighlight = _activeButtonHighlight;

            buttonDict = new Dictionary<int, UIDynamicButton>();
        }


        private UIDynamicPopup CreateUIPopup(JSONStorableStringChooser jssc, bool scrollable)
        {
            Transform originalTransform = UIAGlobals.mvrScript.manager.configurableScrollablePopupPrefab.transform;
            if (!scrollable) originalTransform = UIAGlobals.mvrScript.manager.configurablePopupPrefab.transform;

            Transform popup = UnityEngine.Object.Instantiate(originalTransform, _spacerActive.gameObject.transform, false);
            UIDynamicPopup selectTargetPopup = popup.GetComponent<UIDynamicPopup>();

            selectTargetPopup.popup.showSlider = false;
            popup.SetParent(_spacerActive.gameObject.transform, false);
            jssc.popup = selectTargetPopup.popup;
            return (selectTargetPopup);
        }


        public UIDynamicPopup AddPopupToRow(JSONStorableStringChooser jssc, UIPopup.OnOpenPopup callbackMethod = null, float panelHeight = 400f, bool forceRightParentPopup = false, bool scrollable = true)
        {
            UIDynamicPopup newPopup = CreateUIPopup(jssc, scrollable);
            newPopup.label = jssc.label;
            newPopup.popup.onOpenPopupHandlers += callbackMethod;
            newPopup.popup.onOpenPopupHandlers += () =>
            {
                UIAPluginUI.CloseAllPopups(newPopup);
                UIAGlobals.mvrScript.StartCoroutine(ReparentPopup(newPopup, forceRightParentPopup));
            };
            newPopup.popup.popupPanelHeight = panelHeight;
            newPopup.popup.selectColor = UIAConsts.UIhighlightColor;

            uidPopuplist.Add(newPopup);

            return newPopup;
        }

        // This prevents the popup panel from being rendered behind other items. Needs to be set at the end of the Frame to override the VAM default behaviour.
        private IEnumerator ReparentPopup(UIDynamicPopup popupUID, bool forceRightParentPopup)
        {
            yield return new WaitForEndOfFrame();

            if (popupUID.popup.visible)
            {
                if (forceRightParentPopup) popupUID.popup.popupPanel.transform.SetParent(_spacerRight.gameObject.transform.parent);
                else popupUID.popup.popupPanel.transform.SetParent(_spacerActive.gameObject.transform.parent);
            }
        }

        private UIDynamicButton CreateUIButton()
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(UIAGlobals.mvrScript.manager.configurableButtonPrefab.transform, _spacerActive.gameObject.transform, false);
            UIDynamicButton selectTargetButton = buttonTransform.GetComponent<UIDynamicButton>();
            buttonTransform.SetParent(_spacerActive.gameObject.transform, false);
            return (selectTargetButton);
        }

        private UIDynamicButton CreateGridlessUIButton(float width, float xOffsetPosition, float yOffsetPostion = 0f)
        {
            RectTransform spacerRT = _spacerActive.GetComponent<RectTransform>();

            spacerRT.anchorMin = new Vector2(0f, 0.5f);
            spacerRT.anchorMax = new Vector2(0f, 0.5f);
            spacerRT.pivot = new Vector2(0f, 0.5f);

            UIDynamicButton targetButton = CreateUIButton();

            RectTransform childRect = targetButton.button.GetComponent<RectTransform>();
            childRect.anchorMin = new Vector2(0f, 0.5f);
            childRect.anchorMax = new Vector2(0f, 0.5f);
            childRect.pivot = new Vector2(0f, 0.5f);

            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, spacerHeight);
            childRect.transform.localPosition = new Vector3(xOffsetPosition, yOffsetPostion, 0f);

            return (targetButton);
        }

        private void ClearAllButtonColors()
        {
            foreach (KeyValuePair<int, UIDynamicButton> kvp in buttonDict)
            {
                kvp.Value.buttonColor = new Color(0.85f, 0.85f, 0.85f, 1f);
            }
        }

        private void RegisterButtonAction(UIDynamicButton newButton, string label, int buttonRef, Action<int> callbackMethod)
        {
            newButton.label = label;

            newButton.button.onClick.AddListener(delegate
            {
                try
                {
                    if (activeButtonHighlight) SetButtonColorActive(buttonRef);
                    callbackMethod.Invoke(buttonRef);
                }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });

            buttonDict.Add(buttonRef, newButton);
        }

        public UIDynamicButton AddButtonToRow(string label, int buttonRef, Action<int> callbackMethod)
        {
            UIDynamicButton newButton = CreateUIButton();
            RegisterButtonAction(newButton, label, buttonRef, callbackMethod);

            return newButton;
        }

        public UIDynamicButton AddGridlessButtonToRow(string label, int buttonRef, Action<int> callbackMethod, float width, float xOffsetPostion, float yOffsetPostion = 0, string pointerEnterLabel = "")
        {
            UIDynamicButton newButton = CreateGridlessUIButton(width, xOffsetPostion, yOffsetPostion);

            RegisterButtonAction(newButton, label, buttonRef, callbackMethod);
            if (pointerEnterLabel != "") newButton.label = pointerEnterLabel;

            if (pointerEnterLabel != "")
            {
                EventTrigger trigger = newButton.button.gameObject.AddComponent<EventTrigger>();
                var pointerEnter = new EventTrigger.Entry();
                pointerEnter.eventID = EventTriggerType.PointerEnter;
                pointerEnter.callback.AddListener((e) => PointerEnterButton(newButton, pointerEnterLabel));
                trigger.triggers.Add(pointerEnter);

                var pointerExit = new EventTrigger.Entry();
                pointerExit.eventID = EventTriggerType.PointerExit;
                pointerExit.callback.AddListener((e) => PointerExitButton(newButton, label));
                trigger.triggers.Add(pointerExit);
            }

            return newButton;
        }
        public UIDynamicButton AddGridlessButtonToRow(string label, int buttonRef, Action<int> callbackMethod, float width, float xOffsetPostion, UIDynamicButton targetLabelButton, string pointerEnterLabel, string pointerExitLabel)
        {
            UIDynamicButton newButton = AddGridlessButtonToRow(label, buttonRef, callbackMethod, width, xOffsetPostion);

            if (pointerEnterLabel != "")
            {
                EventTrigger trigger = newButton.button.gameObject.AddComponent<EventTrigger>();
                var pointerEnter = new EventTrigger.Entry();
                pointerEnter.eventID = EventTriggerType.PointerEnter;
                pointerEnter.callback.AddListener((e) => PointerEnterButton(targetLabelButton, pointerEnterLabel));
                trigger.triggers.Add(pointerEnter);

                var pointerExit = new EventTrigger.Entry();
                pointerExit.eventID = EventTriggerType.PointerExit;
                pointerExit.callback.AddListener((e) => PointerExitButton(targetLabelButton, pointerExitLabel));
                trigger.triggers.Add(pointerExit);
            }

            return newButton;
        }

        public void PointerEnterButton(UIDynamicButton button, string label)
        {
            button.label = label;
        }
        public void PointerExitButton(UIDynamicButton button, string label)
        {
            button.label = label;
        }


        public void SetButtonColorActive(int buttonRef)
        {
            ClearAllButtonColors();
            buttonDict[buttonRef].buttonColor = UIAConsts.UIhighlightColor;
        }

        public override void ClearUI()
        {
            base.ClearUI();
            buttonDict.Clear();
        }

        public void InitUI(int elementsCount, bool glgEnabled = true, bool fullWidth = true, bool isRight = false, bool includeRightSpacer = true)
        {
            if (fullWidth || !isRight)
            {
                _spacerLeft = UIAGlobals.mvrScript.CreateSpacer(false);
                _spacerActive = _spacerLeft;
                _spacerLeft.height = spacerHeight;


            }
            if (includeRightSpacer || isRight)
            {
                _spacerRight = UIAGlobals.mvrScript.CreateSpacer(true);
                _spacerRight.height = spacerHeight;
            }
            if (isRight)
            {
                _spacerActive = _spacerRight;
                //               if (isRight && !fullWidth) _spacerActive = _spacerRight;
                //               else _spacerActive = _spacerLeft;
            }

            if (glgEnabled)
            {
                int width = 525;
                if (fullWidth) width = 1050;

                GridLayoutGroup glg = _spacerActive.gameObject.AddComponent<GridLayoutGroup>();
                glg.cellSize = new Vector2((width - ((elementsCount - 1) * buttonPadding)) / elementsCount, spacerHeight);

                glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
                glg.constraintCount = 1;
                glg.padding = new RectOffset(buttonPadding, buttonPadding, buttonPadding, buttonPadding);
                glg.spacing = new Vector2(buttonPadding, 0);
            }
        }
    }

    public class CustomTreeNavUIElement : CustomUIElement
    {
        private static List<CustomTreeNavUIRow> customTreeNavUIRows;
        private static float latestTopScrollRow = 0f;
        private static UIAButton copyButton = null;
        private static UIAButtonOperation copyButtonOp = null;

        public static int treeNavMode
        {
            get; private set;
        } = NORMAL;
        public const int NORMAL = 1;
        private const int MOVEGRID = 2;
        private const int DELETEGRIDCONFIRM = 3;
        private const int MOVEBUTTON = 4;
        private const int COPYBUTTON = 5;
        private const int CLEARBUTTONCONFIRM = 6;
        private const int MOVEBUTTONOP = 7;
        private const int COPYBUTTONOP = 8;
        private const int DELETEBUTTONOPCONFIRM = 9;

        private const float treeBrowserWidth = 530f;
        private const float treeBrowserRowHeight = 30f;

        private static GameObject buttonListContentGO;

        public class CustomTreeNavUIRow
        {
            private int rowType;

            private const int GRID = 1;
            private const int ROW = 2;
            private const int BUTTON = 3;
            private const int OPERATION = 4;

            private UIAButtonGrid uiaButtonGrid;
            private UIAButton uiaButton;
            private UIAButtonOperation uiaButtonOperation;
            private bool setCurrent;
            private int row;


            public CustomTreeNavUIRow(UIAButtonGrid bg, bool _setCurrent)
            {
                uiaButtonGrid = bg;
                rowType = GRID;
                setCurrent = _setCurrent;
                InitUI();
            }
            public CustomTreeNavUIRow(UIAButton button, bool _setCurrent)
            {
                uiaButton = button;
                rowType = BUTTON;
                setCurrent = _setCurrent;
                InitUI();
            }
            public CustomTreeNavUIRow(UIAButtonOperation operation, bool _setCurrent)
            {
                uiaButtonOperation = operation;
                rowType = OPERATION;
                setCurrent = _setCurrent;
                InitUI();
            }
            public CustomTreeNavUIRow(UIAButtonGrid bg, int _row)
            {
                uiaButtonGrid = bg;
                rowType = ROW;
                row = _row;
                setCurrent = false;

                InitUI();
            }

            public bool IsCollapsible()
            {
                if (treeNavMode == CLEARBUTTONCONFIRM && rowType != GRID && rowType != ROW) return false;
                if (treeNavMode == MOVEGRID) return false;
                if (treeNavMode == MOVEBUTTON && rowType == BUTTON) return false;
                if (treeNavMode == COPYBUTTON && rowType == BUTTON) return false;
                if (treeNavMode == MOVEBUTTON && rowType == GRID && uiaButtonGrid.gridRef != ButtonSetupPluginUI.gridSelectorJSSC.val) return false;
                if (rowType == GRID || rowType == ROW) return true;
                if (rowType == BUTTON && PatreonFeatures.patreonContentEnabled && uiaButton.buttonOperations.Count > 1) return true;
                return false;
            }

            private bool IsCollapsed()
            {
                if (rowType == GRID) return uiaButtonGrid.treeBrowserCollapsed;
                if (rowType == ROW) return uiaButtonGrid.GetTreeBrowserRowCollapsedRow(row);
                if (rowType == BUTTON) return uiaButton.treeBrowserCollapsed;
                return true;
            }
            private bool IsAddItemActive()
            {
                if (treeNavMode != NORMAL) return false;
                if (!setCurrent) return false;
                if (rowType == GRID && uiaButtonGrid != UIAStorables.quickLaunchButtonGrid) return true;
                if ((rowType == BUTTON || rowType == OPERATION) && PatreonFeatures.patreonContentEnabled) return true;
                return false;
            }
            private bool IsInsertItemActive()
            {
                if (treeNavMode == MOVEBUTTONOP && !setCurrent && rowType == OPERATION && (ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpIndex != uiaButtonOperation.buttonOpIndex + 1 || ButtonSetupPluginUI.currentUIAButton != uiaButtonOperation.parentButton)) return true;
                if (treeNavMode == MOVEBUTTONOP && rowType == BUTTON && ButtonSetupPluginUI.currentUIAButton != uiaButton) return true;
                if (treeNavMode == MOVEBUTTON && !setCurrent && rowType == BUTTON && ButtonSetupPluginUI.currentUIAButton.buttonGridIndex != uiaButton.buttonGridIndex + 1) return true;
                if (treeNavMode == MOVEBUTTON && rowType == GRID && uiaButtonGrid.gridRef != ButtonSetupPluginUI.gridSelectorJSSC.val) return true;
                if (treeNavMode == MOVEGRID && !setCurrent && rowType == GRID && ButtonSetupPluginUI.currentUIAButtonGrid.gridIndex != uiaButtonGrid.gridIndex + 1 && uiaButtonGrid != UIAStorables.quickLaunchButtonGrid) return true;
                return false;
            }

            private bool IsInteractable()
            {
                if (treeNavMode == CLEARBUTTONCONFIRM) return false;
                if (treeNavMode == DELETEBUTTONOPCONFIRM && rowType != OPERATION) return false;
                if (rowType == ROW) return false;
                if (rowType == BUTTON && !uiaButton.treeBrowserCollapsed && uiaButton.buttonOperations.Count > 1 && treeNavMode != MOVEBUTTON && treeNavMode != COPYBUTTON) return false;
                if (rowType == BUTTON && treeNavMode == MOVEBUTTONOP) return false;
                if (rowType == GRID && (treeNavMode == MOVEBUTTON || treeNavMode == COPYBUTTON || treeNavMode == MOVEBUTTONOP || treeNavMode == COPYBUTTONOP)) return false;
                return true;
            }

            private int indentation { get
                {
                    if (rowType == ROW) return 40;
                    if (rowType == BUTTON) return 70;
                    if (rowType == OPERATION) return 100;
                    return 10;
                }
            }

            private string GetRowOperationLabel(UIAButtonOperation operation)
            {
                return operation.buttonOpTypeJSEnum.displayVal;
            }
            private string rowLabel { get
                {
                    if (rowType == GRID)
                    {
                        string gridLabel = uiaButtonGrid.gridLabel;
                        if (uiaButtonGrid == UIAStorables.quickLaunchButtonGrid) gridLabel = "Quick Launch Bar";
                        if (ButtonSetupPluginUI.currentComponentTab != ButtonComponentTypes.grid && ButtonSetupPluginUI.gridSelectorJSSC.val == uiaButtonGrid.gridRef) return "<b> GRID " + gridLabel + " *</b>";
                        return " GRID " + gridLabel;
                    }
                    if (rowType == ROW)
                    {
                        if (ButtonSetupPluginUI.currentComponentTab != ButtonComponentTypes.grid && ButtonSetupPluginUI.currentUIAButton.buttonRow == row && ButtonSetupPluginUI.gridSelectorJSSC.val == uiaButtonGrid.gridRef)
                        {
                            return "<b>ROW " + row.ToString() + " *</b>";
                        }
                        return "ROW " + row.ToString();
                    }
                    if (rowType == BUTTON)
                    {
                        string manualLabel = uiaButton.GetManualLabel();
                        if (manualLabel != "") return " " + uiaButton.buttonRef + ": " + manualLabel;
                        if (uiaButton.buttonOperations.Count == 1) return " " + uiaButton.buttonRef + ": " + GetRowOperationLabel(uiaButton.buttonOperations[0]);
                        return " " + uiaButton.buttonRef;
                    }
                    if (rowType == OPERATION) return " " + uiaButtonOperation.buttonOpRef + ": " + GetRowOperationLabel(uiaButtonOperation);
                    return "";
                }
            }
            private void AddTreeRowItemSelected()
            {
                if (rowType == GRID) AddGrid();
                if (rowType == BUTTON) AddOperation(uiaButton, 0);
                if (rowType == OPERATION) AddOperation(uiaButtonOperation.parentButton, uiaButtonOperation.buttonOpIndex + 1);
            }

            private void InsertButtonMove()
            {
                string buttonRef = "R1C1";
                UIAButton movingButton = ButtonSetupPluginUI.currentUIAButton;
                int currentIndex = movingButton.buttonGridIndex;

                if (rowType == BUTTON)
                {
                    int newIndex = uiaButton.buttonGridIndex;

                    movingButton.parentGrid.buttonList.RemoveAt(currentIndex);
                    if (newIndex > currentIndex) newIndex--;
                    movingButton.parentGrid.buttonList.Insert(newIndex + 1, movingButton);
                    buttonRef = movingButton.buttonRef;
                }
                else if (rowType == GRID)
                {
                    int newIndex = uiaButtonGrid.GetFirstBlankButtonIndex();
                    UIAButton swappingButton = uiaButtonGrid.buttonList[newIndex];

                    uiaButtonGrid.buttonList[newIndex] = movingButton;
                    movingButton.parentGrid = uiaButtonGrid;
                    ButtonSetupPluginUI.currentUIAButtonGrid.buttonList[currentIndex] = swappingButton;
                    swappingButton.parentGrid = ButtonSetupPluginUI.currentUIAButtonGrid;
                    movingButton.parentGrid.treeBrowserCollapsed = false;
                    movingButton.parentGrid.SetTreeBrowserRowCollapsedRow(movingButton.buttonRow, false);
                    buttonRef = movingButton.buttonRef;
                    ButtonSetupPluginUI.gridSelectorJSSC.val = movingButton.parentGrid.gridRef;
                }

                ButtonSetupPluginUI.buttonSelectorJSSC.val = buttonRef;
                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
                GameControlUI.RefreshWristUIButtonGrid();
            }


            private void InsertButtonOpMove()
            {
                UIAButtonOperation movingButtonOp = ButtonSetupPluginUI.currentUIAButtonOperation;
                int currentIndex = movingButtonOp.buttonOpIndex;


                if (rowType == OPERATION)
                {
                    int newIndex = uiaButtonOperation.buttonOpIndex;

                    movingButtonOp.parentButton.buttonOperations.Remove(movingButtonOp);
                    if (movingButtonOp.parentButton == uiaButtonOperation.parentButton && newIndex > currentIndex) newIndex--;
                    uiaButtonOperation.parentButton.buttonOperations.Insert(newIndex + 1, movingButtonOp);
                    movingButtonOp.parentButton = uiaButtonOperation.parentButton;
                }
                else if (rowType == BUTTON)
                {
                    movingButtonOp.parentButton.buttonOperations.Remove(movingButtonOp);
                    uiaButton.buttonOperations.Add(movingButtonOp);
                    movingButtonOp.parentButton = uiaButton;

                    uiaButton.treeBrowserCollapsed = false;
                    ButtonSetupPluginUI.buttonSelectorJSSC.val = uiaButton.buttonRef;
                }
                ButtonSetupPluginUI.gridSelectorJSSC.val = movingButtonOp.parentButton.parentGrid.gridRef;
                ButtonSetupPluginUI.buttonSelectorJSSC.val = movingButtonOp.parentButton.buttonRef;
                ButtonSetupPluginUI.buttonOperationSelectorJSSC.val = movingButtonOp.buttonOpRef;

                treeNavMode = MOVEBUTTONOP;


                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
                GameControlUI.RefreshWristUIButtonGrid();
            }

            private void InsertGridMove()
            {
                int currentIndex = ButtonSetupPluginUI.currentUIAButtonGrid.gridIndex;
                int newIndex = uiaButtonGrid.gridIndex;
                UIAButtonGrid movingButtonGrid = ButtonSetupPluginUI.currentUIAButtonGrid;

                UIAStorables.buttonGridsList.RemoveAt(currentIndex);
                if (newIndex > currentIndex) newIndex--;
                UIAStorables.buttonGridsList.Insert(newIndex + 1, movingButtonGrid);
                ButtonSetupPluginUI.gridSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButtonGrid.gridRef;
                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
                GameControlUI.RefreshWristUIButtonGrid();
            }

            private void AddGrid()
            {
                int currentGridIndex = uiaButtonGrid.gridIndex;

                UIAButtonGrid newGrid = new UIAButtonGrid(3, 3);
                UIAStorables.buttonGridsList.Insert(currentGridIndex + 1, newGrid);
                ButtonSetupPluginUI.gridSelectorJSSC.val = newGrid.gridRef;

                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);

                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit) GameControlUI.RefreshWristUIButtonGrid();
            }
            private void AddOperation(UIAButton currentUIAButton, int opIndex)
            {

                UIAButtonOperation newOp = new UIAButtonOperation(currentUIAButton);
                currentUIAButton.buttonOperations.Insert(opIndex, newOp);
                ButtonSetupPluginUI.buttonOperationSelectorJSSC.valNoCallback = newOp.buttonOpRef;
                ButtonSetupPluginUI.OperationSelected(newOp.buttonOpRef);
                currentUIAButton.treeBrowserCollapsed = false;

                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
            }

            private void ToggleCollapsedButton()
            {
                if (rowType == GRID) uiaButtonGrid.treeBrowserCollapsed = !uiaButtonGrid.treeBrowserCollapsed;
                if (rowType == ROW) uiaButtonGrid.SetTreeBrowserRowCollapsedRow(row, !uiaButtonGrid.GetTreeBrowserRowCollapsedRow(row));
                if (rowType == BUTTON) uiaButton.treeBrowserCollapsed = !uiaButton.treeBrowserCollapsed;
                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
            }
            private void TreeRowItemSelected()
            {
                if (!setCurrent)
                {
                    if (rowType == GRID)
                    {
                        ButtonSetupPluginUI.gridSelectorJSSC.val = uiaButtonGrid.gridRef;
                        UIAPluginUI.buttonSetupUI.ComponentTabSelected(ButtonComponentTypes.grid);
                    }
                    if (rowType == BUTTON)
                    {
                        ButtonSetupPluginUI.gridSelectorJSSC.val = uiaButton.parentGrid.gridRef;
                        ButtonSetupPluginUI.buttonSelectorJSSC.val = uiaButton.buttonRef;
                        if (ButtonSetupPluginUI.currentComponentTab == ButtonComponentTypes.grid) UIAPluginUI.buttonSetupUI.ComponentTabSelected(ButtonComponentTypes.operation);
                    }
                    if (rowType == OPERATION)
                    {
                        ButtonSetupPluginUI.gridSelectorJSSC.val = uiaButtonOperation.parentButton.parentGrid.gridRef;
                        ButtonSetupPluginUI.buttonSelectorJSSC.val = uiaButtonOperation.parentButton.buttonRef;
                        ButtonSetupPluginUI.buttonOperationSelectorJSSC.val = uiaButtonOperation.buttonOpRef;

                        if (ButtonSetupPluginUI.currentComponentTab == ButtonComponentTypes.grid) UIAPluginUI.buttonSetupUI.ComponentTabSelected(ButtonComponentTypes.operation);

                    }
                }
                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            }

            private void InitUI()
            {
                GameObject treeItemContainerImageGO = UITools.CreateUIContainerImageInParentGO(buttonListContentGO.transform, treeBrowserWidth, treeBrowserRowHeight, new Color(0.1f, 0.1f, 0.1f, 0.0f));

                int extraIndentation = 0;
                if (IsCollapsible())
                {
                    GameObject collapseButtonGO = UITools.CreateUIButtonInParent(treeItemContainerImageGO.transform);
                    UITools.SetSizeAndAnchors(treeItemContainerImageGO, collapseButtonGO, 30f, treeBrowserRowHeight, "left", indentation);
                    extraIndentation = 30;
                    UIDynamicButton uidCollapseButton = collapseButtonGO.transform.GetComponent<UIDynamicButton>();
                    if (IsCollapsed()) uidCollapseButton.label = "+";
                    else uidCollapseButton.label = "-";
                    uidCollapseButton.buttonText.fontSize = 24;

                    uidCollapseButton.button.onClick.AddListener(delegate
                    {
                        try
                        {
                            ToggleCollapsedButton();
                        }
                        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

                    });

                }
                int widthAdjustment = 0;
                if (IsAddItemActive() || IsInsertItemActive())
                {
                    widthAdjustment += 30;
                    GameObject addButtonGO = UITools.CreateUIButtonInParent(treeItemContainerImageGO.transform);
                    UITools.SetSizeAndAnchors(treeItemContainerImageGO, addButtonGO, 30f, treeBrowserRowHeight, "left", treeBrowserWidth - 40f);
                    UIDynamicButton uidAddButton = addButtonGO.transform.GetComponent<UIDynamicButton>();
                    uidAddButton.buttonText.fontSize = 26;
                    if (IsInsertItemActive())
                    {
                        uidAddButton.label = ">";
                        uidAddButton.buttonColor = Color.yellow;
                    }
                    else
                    {
                        uidAddButton.label = "+";
                        uidAddButton.buttonColor = Color.green;
                    }

                    uidAddButton.button.onClick.AddListener(delegate
                    {
                        try
                        {
                            if (treeNavMode == MOVEBUTTON) InsertButtonMove();
                            else if (treeNavMode == MOVEGRID) InsertGridMove();
                            else if (treeNavMode == MOVEBUTTONOP) InsertButtonOpMove();
                            else AddTreeRowItemSelected();
                        }
                        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

                    });
                }

                GameObject buttonGO = UITools.CreateUIButtonInParent(treeItemContainerImageGO.transform);
                UITools.SetSizeAndAnchors(treeItemContainerImageGO, buttonGO, treeBrowserWidth - indentation - extraIndentation - 10f - widthAdjustment, treeBrowserRowHeight, "left", indentation + extraIndentation);
                UIDynamicButton uidButton = buttonGO.transform.GetComponent<UIDynamicButton>();
                if (!IsInteractable())
                {
                    uidButton.button.interactable = false;
                    uidButton.textColor = new Color(1f, 1f, 1f);
                }
                uidButton.buttonText.fontSize = 24;
                uidButton.label = rowLabel;
                uidButton.buttonText.alignment = TextAnchor.MiddleLeft;
                if (rowType == BUTTON && copyButton == uiaButton) uidButton.buttonColor = Color.yellow;
                if (rowType == OPERATION && copyButtonOp == uiaButtonOperation) uidButton.buttonColor = Color.yellow;
                else if (setCurrent) uidButton.buttonColor = Color.cyan;
                uidButton.button.onClick.AddListener(delegate
                {
                    try
                    {
                        TreeRowItemSelected();
                    }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

                });

            }

        }

        public static void PluginUIGridChange()
        {
            if (treeNavMode == DELETEGRIDCONFIRM || treeNavMode == CLEARBUTTONCONFIRM || treeNavMode == DELETEBUTTONOPCONFIRM) treeNavMode = NORMAL;
        }
        public static void PluginUIButtonChange()
        {
            if (treeNavMode == CLEARBUTTONCONFIRM || treeNavMode == DELETEBUTTONOPCONFIRM || (treeNavMode == MOVEBUTTONOP && ButtonSetupPluginUI.currentUIAButton.buttonOperations.Count < 2)) treeNavMode = NORMAL;
        }

        public static void PluginUIGridComponentTabSelected()
        {
            if (treeNavMode != MOVEGRID && treeNavMode != DELETEGRIDCONFIRM) treeNavMode = NORMAL;
        }

        public CustomTreeNavUIElement() : base(0)
        {
            customTreeNavUIRows = new List<CustomTreeNavUIRow>();
        }

        public override void ClearUI()
        {
            base.ClearUI();
        }

        private void MoveGridUp()
        {
            int index = UIAStorables.buttonGridsList.IndexOf(ButtonSetupPluginUI.currentUIAButtonGrid);
            index += -1;
            UIAStorables.buttonGridsList.Reverse(index, 2);

            ButtonSetupPluginUI.RefreshGridChoices();
            ButtonSetupPluginUI.gridSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButtonGrid.gridRef;
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            GameControlUI.RefreshWristUIButtonGrid();
        }
        private void MoveGridDown()
        {
            int index = UIAStorables.buttonGridsList.IndexOf(ButtonSetupPluginUI.currentUIAButtonGrid);
            UIAStorables.buttonGridsList.Reverse(index, 2);
            ButtonSetupPluginUI.RefreshGridChoices();
            ButtonSetupPluginUI.gridSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButtonGrid.gridRef;
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            GameControlUI.RefreshWristUIButtonGrid();
        }

        private void MoveButtonUp()
        {
            int index = ButtonSetupPluginUI.currentUIAButton.buttonGridIndex;
            index += -1;
            ButtonSetupPluginUI.currentUIAButtonGrid.buttonList.Reverse(index, 2);

            ButtonSetupPluginUI.buttonSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButton.buttonRef;

            ButtonSetupPluginUI.currentUIAButton.parentGrid.SetTreeBrowserRowCollapsedRow(ButtonSetupPluginUI.currentUIAButton.buttonRow, false);

            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            GameControlUI.RefreshWristUIButtonGrid();
        }

        private void MoveButtonDown()
        {
            int index = ButtonSetupPluginUI.currentUIAButton.buttonGridIndex;
            ButtonSetupPluginUI.currentUIAButtonGrid.buttonList.Reverse(index, 2);

            ButtonSetupPluginUI.buttonSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButton.buttonRef;

            ButtonSetupPluginUI.currentUIAButton.parentGrid.SetTreeBrowserRowCollapsedRow(ButtonSetupPluginUI.currentUIAButton.buttonRow, false);

            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            GameControlUI.RefreshWristUIButtonGrid();
        }


        private void MoveButtonOpUp()
        {
            int index = ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpIndex;
            index += -1;
            ButtonSetupPluginUI.currentUIAButton.buttonOperations.Reverse(index, 2);

            ButtonSetupPluginUI.buttonOperationSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpRef;
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            GameControlUI.RefreshWristUIButtonGrid();
        }

        private void MoveButtonOpDown()
        {
            int index = ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpIndex;
            ButtonSetupPluginUI.currentUIAButton.buttonOperations.Reverse(index, 2);

            ButtonSetupPluginUI.buttonOperationSelectorJSSC.valNoCallback = ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpRef;
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            GameControlUI.RefreshWristUIButtonGrid();
        }
        private void DeleteCurrentGrid()
        {
            treeNavMode = NORMAL;
            int currentGridIndex = ButtonSetupPluginUI.currentUIAButtonGrid.gridIndex;
            int currentGameUIGridIndex = GameControlUI._currentGrid.gridIndex;

            UIAStorables.buttonGridsList.Remove(ButtonSetupPluginUI.currentUIAButtonGrid);
            ButtonSetupPluginUI.RefreshGridChoices();
            string gridRef = "1";
            if (currentGridIndex > 0) gridRef = currentGridIndex.ToString();

            ButtonSetupPluginUI.gridSelectorJSSC.valNoCallback = gridRef;
            ButtonSetupPluginUI.GridSelected(gridRef);

            if (currentGameUIGridIndex == currentGridIndex) GameControlUI._currentGrid = ButtonSetupPluginUI.currentUIAButtonGrid;

            GameControlUI.RefreshWristUIButtonGrid();
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
        }
        private void DeleteCurrentOperation()
        {
            treeNavMode = NORMAL;
            UIAButtonOperation buttonOp = ButtonSetupPluginUI.currentUIAButtonOperation;
            int currentOpIndex = buttonOp.buttonOpIndex;

            buttonOp.parentButton.buttonOperations.Remove(buttonOp);
            string operationValue = "1";
            if (currentOpIndex > 0) operationValue = currentOpIndex.ToString();

            ButtonSetupPluginUI.buttonOperationSelectorJSSC.valNoCallback = operationValue;
            ButtonSetupPluginUI.OperationSelected(operationValue);

            GameControlUI.RefreshWristUIButtonGrid();
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
        }


        private void PasteButton()
        {
            ButtonSetupPluginUI.currentUIAButton.CopyFrom(copyButton);
            ButtonSetupPluginUI.currentUIAButtonOperation = ButtonSetupPluginUI.currentUIAButton.buttonOperations[0];
            GameControlUI.RefreshWristUIButtonGrid();

            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
        }

        private void PasteButtonOp()
        {
            ButtonSetupPluginUI.currentUIAButtonOperation.CopyFrom(copyButtonOp);

            GameControlUI.RefreshWristUIButtonGrid();
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
        }

        private void ClearButton()
        {
            treeNavMode = NORMAL;
            ButtonSetupPluginUI.buttonOperationSelectorJSSC.val = "1";
            var list = ButtonSetupPluginUI.currentUIAButton.buttonOperations;
            if (list.Count > 1)
                list.RemoveRange(1, list.Count - 1);
            list[0].buttonOpCategoryJSEnum.val = UIAButtonCategory.misc;
            list[0].buttonOpTypeJSEnum.val = UIAButtonOpType.blank;
            ButtonSetupPluginUI.currentUIAButtonOperation = ButtonSetupPluginUI.currentUIAButton.buttonOperations[0];
            ButtonSetupPluginUI.buttonOperationSelectorJSSC.valNoCallback = "1";
            GameControlUI.RefreshWristUIButtonGrid();
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
        }

        private UIDynamicButton CreateButtonInContainer(GameObject containerImage, float width, float height, string placement, string label, float xOffset, float yOffset)
        {
            GameObject buttonGO = UITools.CreateUIButtonInParent(containerImage.transform);
            UIDynamicButton uidButton = buttonGO.transform.GetComponent<UIDynamicButton>();
            UITools.SetSizeAndAnchors(containerImage.gameObject, buttonGO, width, height, placement);
            buttonGO.transform.localPosition = new Vector3(buttonGO.transform.localPosition.x + xOffset, buttonGO.transform.localPosition.y + yOffset, buttonGO.transform.localPosition.z);
            if (label != "") uidButton.label = label;
            uidButton.buttonText.fontSize = 24;
            return uidButton;
        }

        public UIDynamicButton CreateButtonInContainer(GameObject containerImage, float width, float height, string placement, string label = "", int newMode = NORMAL)
        {
            UIDynamicButton uidButton = CreateButtonInContainer(containerImage, width, height, placement, label, 0f, 0f);

            uidButton.button.onClick.AddListener(delegate
            {
                treeNavMode = newMode;
                UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);
            });

            return uidButton;
        }
        public UIDynamicButton CreateButtonInContainer(GameObject containerImage, float width, float height, string placement, string label = "", UnityAction callback = null, float xOffset = 0f, float yOffset = 0f)
        {
            UIDynamicButton uidButton = CreateButtonInContainer(containerImage, width, height, placement, label, xOffset, yOffset);

            uidButton.button.onClick.AddListener(callback);

            return uidButton;
        }

        private void CreateTreeRowButtonOp(UIAButtonOperation operation)
        {
            bool setCurrent = false;
            if (ButtonSetupPluginUI.currentComponentTab != ButtonComponentTypes.grid && ButtonSetupPluginUI.gridSelectorJSSC.val == operation.parentButton.parentGrid.gridRef && ButtonSetupPluginUI.buttonSelectorJSSC.val == operation.parentButton.buttonRef && ButtonSetupPluginUI.buttonOperationSelectorJSSC.val == operation.buttonOpRef) setCurrent = true;
            customTreeNavUIRows.Add(new CustomTreeNavUIRow(operation, setCurrent));
        }
        private void CreateTreeRowButton(UIAButton button)
        {
            bool setCurrent = false;
            if (ButtonSetupPluginUI.currentComponentTab != ButtonComponentTypes.grid && ButtonSetupPluginUI.gridSelectorJSSC.val == button.parentGrid.gridRef && ButtonSetupPluginUI.buttonSelectorJSSC.val == button.buttonRef && (button.buttonOperations.Count < 2 || button.treeBrowserCollapsed || treeNavMode == MOVEBUTTON || treeNavMode == COPYBUTTON)) setCurrent = true;

            customTreeNavUIRows.Add(new CustomTreeNavUIRow(button, setCurrent));
            if (button.buttonOperations.Count > 1 && treeNavMode != MOVEBUTTON && treeNavMode != COPYBUTTON)
            {
                if (!button.treeBrowserCollapsed)
                {
                    foreach (UIAButtonOperation operation in button.buttonOperations)
                    {
                        if (IsOperationDisplayed(operation)) CreateTreeRowButtonOp(operation);
                    }
                }

            }
        }

        private void CreateTreeRowButtonRow(UIAButtonGrid bg, int row)
        {
            bool rowCollapser = (bg.buttonCount) > 9;
            if (rowCollapser) customTreeNavUIRows.Add(new CustomTreeNavUIRow(bg, row));
            if (!bg.GetTreeBrowserRowCollapsedRow(row) || !rowCollapser)
            {
                for (int column = 1; column <= bg.gridColumns; column++)
                {
                    UIAButton button = bg.buttonList[((row - 1) * bg.gridColumns) + column - 1];
                    if (IsButtonDisplayed(button)) CreateTreeRowButton(button);
                }
            }
        }
        private void CreateTreeRowGrid(UIAButtonGrid bg)
        {
            bool setCurrent = false;
            if (ButtonSetupPluginUI.currentComponentTab == ButtonComponentTypes.grid && ButtonSetupPluginUI.gridSelectorJSSC.val == bg.gridRef) setCurrent = true;

            CustomTreeNavUIRow cTreeNavRow = new CustomTreeNavUIRow(bg, setCurrent);
            customTreeNavUIRows.Add(cTreeNavRow);

            if (!bg.treeBrowserCollapsed && cTreeNavRow.IsCollapsible())
            {
                for (int row = 1; row <= bg.gridRows; row++)
                {
                    if (IsButtonRowDisplay(row)) CreateTreeRowButtonRow(bg, row);
                }
            }
        }

        public void InitUI(int height, bool isRight = false)
        {
            if (treeNavMode != COPYBUTTON) copyButton = null;
            if (treeNavMode != COPYBUTTONOP) copyButtonOp = null;
            spacerHeight = height;
            if (!isRight)
            {
                _spacerLeft = UIAGlobals.mvrScript.CreateSpacer(false);
                _spacerActive = _spacerLeft;
                _spacerLeft.height = spacerHeight;
            }
            else
            {
                _spacerRight = UIAGlobals.mvrScript.CreateSpacer(true);
                _spacerActive = _spacerRight;
                _spacerRight.height = spacerHeight;
            }

            GameObject scrollableItemList = UITools.CreateScrollableButtonList(_spacerActive.transform, 0f, 45f, treeBrowserWidth, spacerHeight - 30, new Color(0.5f, 0.5f, 0.5f, 0.5f));

            buttonListContentGO = scrollableItemList.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;

            GameObject containerImage = UITools.CreateUIContainerImageInParentGO(_spacerActive.transform, treeBrowserWidth, 60f, new Color(0.5f, 0.5f, 0.5f, 0.5f));
            UITools.SetSizeAndAnchors(_spacerActive.gameObject, containerImage, treeBrowserWidth, 60f, "bottom");

            switch (treeNavMode)
            {
                case NORMAL:
                    if (ButtonSetupPluginUI.currentComponentTab == ButtonComponentTypes.grid)
                    {
                        if (ButtonSetupPluginUI.gridSelectorJSSC.val != UIAStorables.quickLaunchButtonGrid.gridRef && UIAStorables.buttonGridsList.Count > 1)
                        {
                            CreateButtonInContainer(containerImage, 150f, 50f, "left", "Move Grid", MOVEGRID);
                            CreateButtonInContainer(containerImage, 150f, 50f, "right", "Delete Grid", DELETEGRIDCONFIRM);
                        }

                    }
                    else if (ButtonSetupPluginUI.currentUIAButton.buttonOperations.Count < 2 || ButtonSetupPluginUI.currentUIAButton.treeBrowserCollapsed)
                    {
                        CreateButtonInContainer(containerImage, 150f, 50f, "left", "Move Button", MOVEBUTTON);
                        CreateButtonInContainer(containerImage, 150f, 50f, "center", "Copy Button", COPYBUTTON);
                        if (!ButtonSetupPluginUI.currentUIAButton.ContainsOneBlankOperations()) CreateButtonInContainer(containerImage, 150f, 50f, "right", "Clear Button", CLEARBUTTONCONFIRM);
                    }
                    else
                    {
                        CreateButtonInContainer(containerImage, 160f, 50f, "left", "Move Op", MOVEBUTTONOP);
                        CreateButtonInContainer(containerImage, 160f, 50f, "center", "Copy Op", COPYBUTTONOP);
                        CreateButtonInContainer(containerImage, 160f, 50f, "right", "Delete Op", DELETEBUTTONOPCONFIRM);
                    }

                    break;
                case MOVEGRID:
                    int index = UIAStorables.buttonGridsList.IndexOf(ButtonSetupPluginUI.currentUIAButtonGrid);
                    if (index != 0) CreateButtonInContainer(containerImage, 60f, 50f, "left", "^", MoveGridUp);
                    if (index < UIAStorables.buttonGridsList.Count - 1) CreateButtonInContainer(containerImage, 60f, 50f, "left", "v", MoveGridDown, 80f, 0f);
                    CreateButtonInContainer(containerImage, 200f, 50f, "right", "End Grid Move", NORMAL);
                    break;
                case DELETEGRIDCONFIRM:
                    UIDynamicButton button = CreateButtonInContainer(containerImage, 250f, 50f, "left", "Confirm Grid Deletion", DeleteCurrentGrid);
                    button.buttonColor = Color.red;
                    button.textColor = Color.white;
                    button = CreateButtonInContainer(containerImage, 250f, 50f, "right", "Cancel Grid Deletion", NORMAL);
                    button.buttonColor = Color.green;
                    break;
                case MOVEBUTTON:
                    int buttonIndex = ButtonSetupPluginUI.currentUIAButton.buttonGridIndex;
                    if (buttonIndex != 0) CreateButtonInContainer(containerImage, 60f, 50f, "left", "^", MoveButtonUp);
                    if (buttonIndex < ButtonSetupPluginUI.currentUIAButtonGrid.buttonCount - 1) CreateButtonInContainer(containerImage, 60f, 50f, "left", "v", MoveButtonDown, 80f, 0f);
                    CreateButtonInContainer(containerImage, 200f, 50f, "right", "End Button Move", NORMAL);
                    break;
                case COPYBUTTON:
                    if (copyButton == null) copyButton = ButtonSetupPluginUI.currentUIAButton;

                    if (copyButton != ButtonSetupPluginUI.currentUIAButton)
                    {
                        UIDynamicButton pasteButton = CreateButtonInContainer(containerImage, 150f, 50f, "left", "Paste Button", PasteButton);
                        pasteButton.buttonColor = Color.red;
                        pasteButton.textColor = Color.white;
                    }
                    CreateButtonInContainer(containerImage, 200f, 50f, "right", "End Button Copy", NORMAL);
                    break;
                case CLEARBUTTONCONFIRM:
                    UIDynamicButton clearButtonConfirm = CreateButtonInContainer(containerImage, 250f, 50f, "left", "Confirm Clear Button", ClearButton);
                    clearButtonConfirm.buttonColor = Color.red;
                    clearButtonConfirm.textColor = Color.white;
                    button = CreateButtonInContainer(containerImage, 250f, 50f, "right", "Cancel Clear Button", NORMAL);
                    button.buttonColor = Color.green;
                    break;
                case MOVEBUTTONOP:
                    int buttonOpIndex = ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpIndex;
                    if (buttonOpIndex != 0) CreateButtonInContainer(containerImage, 60f, 50f, "left", "^", MoveButtonOpUp);
                    if (buttonOpIndex < ButtonSetupPluginUI.currentUIAButton.buttonOperations.Count - 1) CreateButtonInContainer(containerImage, 60f, 50f, "left", "v", MoveButtonOpDown, 80f, 0f);
                    CreateButtonInContainer(containerImage, 250f, 50f, "right", "End Operation Move", NORMAL);
                    break;
                case COPYBUTTONOP:
                    if (copyButtonOp == null) copyButtonOp = ButtonSetupPluginUI.currentUIAButtonOperation;

                    if (copyButtonOp != ButtonSetupPluginUI.currentUIAButtonOperation)
                    {
                        UIDynamicButton pasteButtonOp = CreateButtonInContainer(containerImage, 200f, 50f, "left", "Paste Operation", PasteButtonOp);
                        pasteButtonOp.buttonColor = Color.red;
                        pasteButtonOp.textColor = Color.white;
                    }
                    CreateButtonInContainer(containerImage, 250f, 50f, "right", "End Operation Copy", NORMAL);
                    break;
                case DELETEBUTTONOPCONFIRM:
                    UIDynamicButton confirmOpDeletebutton = CreateButtonInContainer(containerImage, 250f, 50f, "left", "Confirm Op Deletion", DeleteCurrentOperation);
                    confirmOpDeletebutton.buttonColor = Color.red;
                    confirmOpDeletebutton.textColor = Color.white;
                    UIDynamicButton cancelDeletionbutton = CreateButtonInContainer(containerImage, 250f, 50f, "right", "Cancel Op Deletion", NORMAL);
                    cancelDeletionbutton.buttonColor = Color.green;
                    break;
            }

            customTreeNavUIRows.Clear();
            if (IsGridDisplayed(UIAStorables.quickLaunchButtonGrid) && treeNavMode != MOVEGRID && GameControlSettings.activateQuickLaunchBarJSB.val && PatreonFeatures.patreonContentEnabled) CreateTreeRowGrid(UIAStorables.quickLaunchButtonGrid);
            foreach (UIAButtonGrid bg in UIAStorables.buttonGridsList)
            {
                if (IsGridDisplayed(bg)) CreateTreeRowGrid(bg);
            }

            ScrollRect scrollRect = scrollableItemList.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            sb.onValueChanged.AddListener((float value) => {
                InitSB(sb);
            });

        }

        private bool IsGridDisplayed(UIAButtonGrid bg)
        {
            if (treeNavMode == DELETEGRIDCONFIRM && ButtonSetupPluginUI.gridSelectorJSSC.val != bg.gridRef) return false;
            if (treeNavMode == DELETEBUTTONOPCONFIRM && ButtonSetupPluginUI.gridSelectorJSSC.val != bg.gridRef) return false;
            if (treeNavMode == CLEARBUTTONCONFIRM && ButtonSetupPluginUI.gridSelectorJSSC.val != bg.gridRef) return false;
            if (treeNavMode == MOVEBUTTON && ButtonSetupPluginUI.gridSelectorJSSC.val != bg.gridRef && !bg.ContainsBlankButtons()) return false;
            return true;
        }

        private bool IsButtonRowDisplay(int row)
        {
            if ((treeNavMode == CLEARBUTTONCONFIRM || treeNavMode == DELETEBUTTONOPCONFIRM) && ButtonSetupPluginUI.currentUIAButton.buttonRow != row) return false;

            return true;
        }

        private bool IsButtonDisplayed(UIAButton button)
        {
            if ((treeNavMode == CLEARBUTTONCONFIRM || treeNavMode == DELETEBUTTONOPCONFIRM) && ButtonSetupPluginUI.currentUIAButton != button) return false;
            return true;
        }
        private bool IsOperationDisplayed(UIAButtonOperation operation)
        {
            if (treeNavMode == DELETEBUTTONOPCONFIRM && ButtonSetupPluginUI.currentUIAButtonOperation != operation) return false;
            return true;
        }

        private void InitSB(Scrollbar sb)
        {
            //            yield return new WaitForEndOfFrame();
            float notVisibleRowsA = customTreeNavUIRows.Count() - (customTreeNavUIRows.Count() * sb.size);
            if (notVisibleRowsA == 0) sb.value = 1;
            else sb.value = 1 - (latestTopScrollRow / notVisibleRowsA);

            sb.onValueChanged.RemoveAllListeners();
            sb.onValueChanged.AddListener((float value) => {
                float notVisibleRows = customTreeNavUIRows.Count() - (customTreeNavUIRows.Count() * sb.size);
                float notVisbleAtTop = (1 - value) * notVisibleRows;
                latestTopScrollRow = notVisbleAtTop;
            });
        }
    }

    public abstract class BaseUITab
    {
        protected class GroupLR
        {
            public int groupID;
            public int leftRight;
            public BaseUITab uiTab;

            public void RefreshGroupUIStringChange(string val)
            {
                uiTab.RefreshGroupUI(leftRight, groupID);
            }
        }

        public class UIDynamicElementLists
        {
            public HashSet<UIDynamic> UIDSpacersList;
            public HashSet<UIDynamicButton> UIDButtonsList;
            public HashSet<UIDynamicSlider> UIDSlidersList;
            public HashSet<UIDynamicPopup> UIDPopupList;
            public HashSet<UIDynamicColorPicker> UIDColorPickerList;
            public HashSet<UIDynamicTextField> UIDTextFieldList;
            public HashSet<UIDynamicToggle> UIDToggleList;
            public HashSet<CustomUIElement> UIDCustomElementList;
            public HashSet<BaseUITab> childUITabs;

            public UIDynamicElementLists()
            {
                UIDSpacersList = new HashSet<UIDynamic>();
                UIDButtonsList = new HashSet<UIDynamicButton>();
                UIDSlidersList = new HashSet<UIDynamicSlider>();
                UIDPopupList = new HashSet<UIDynamicPopup>();
                UIDColorPickerList = new HashSet<UIDynamicColorPicker>();
                UIDTextFieldList = new HashSet<UIDynamicTextField>();
                UIDToggleList = new HashSet<UIDynamicToggle>();
                UIDCustomElementList = new HashSet<CustomUIElement>();
                childUITabs = new HashSet<BaseUITab>();
            }

            public void ClearUIDs()
            {
                MVRScript mvrs = UIAGlobals.mvrScript;
                foreach (UIDynamic spacer in UIDSpacersList) if (spacer != null) mvrs.RemoveSpacer(spacer);
                foreach (UIDynamicButton button in UIDButtonsList) if (button != null) mvrs.RemoveButton(button);
                foreach (UIDynamicSlider slider in UIDSlidersList) if (slider != null) mvrs.RemoveSlider(slider);
                foreach (UIDynamicPopup popup in UIDPopupList) if (popup != null) mvrs.RemovePopup(popup);
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) if (colorpicker != null) mvrs.RemoveColorPicker(colorpicker);
                foreach (UIDynamicTextField textfield in UIDTextFieldList) if (textfield != null) mvrs.RemoveTextField(textfield);
                foreach (UIDynamicToggle toggle in UIDToggleList) if (toggle != null) mvrs.RemoveToggle(toggle);
                foreach (CustomUIElement Element in UIDCustomElementList) Element.ClearUI();

                foreach (BaseUITab childUITab in childUITabs) childUITab.ClearAllGroupsUI(LeftRight.both);
                UIDSpacersList.Clear();
                UIDButtonsList.Clear();
                UIDSlidersList.Clear();
                UIDPopupList.Clear();
                UIDColorPickerList.Clear();
                UIDTextFieldList.Clear();
                UIDToggleList.Clear();
                UIDCustomElementList.Clear();
            }

            public float GetUIElementsHeight(bool isRight)
            {
                float height = 0f;
                foreach (UIDynamic spacer in UIDSpacersList) height += spacer.height;
                foreach (UIDynamicButton button in UIDButtonsList) height += button.height;
                foreach (UIDynamicSlider slider in UIDSlidersList) height += slider.height;
                foreach (UIDynamicPopup popup in UIDPopupList) height += popup.height;
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) height += colorpicker.height;
                foreach (UIDynamicTextField textfield in UIDTextFieldList) height += textfield.height;
                foreach (UIDynamicToggle toggle in UIDToggleList) height += toggle.height;
                foreach (CustomUIElement customElement in UIDCustomElementList) height += customElement.height;
                foreach (BaseUITab childUITab in childUITabs) height += childUITab.GetTotalUIElementHeights(isRight);

                return height;
            }
            public int GetUIElementsCount()
            {
                int count = 0;
                foreach (UIDynamic spacer in UIDSpacersList) count += 1;
                foreach (UIDynamicButton button in UIDButtonsList) count += 1;
                foreach (UIDynamicSlider slider in UIDSlidersList) count += 1;
                foreach (UIDynamicPopup popup in UIDPopupList) count += 1;
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) count += 1;
                foreach (UIDynamicTextField textfield in UIDTextFieldList) count += 1;
                foreach (UIDynamicToggle toggle in UIDToggleList) count += 1;
                foreach (CustomUIElement customElement in UIDCustomElementList) count += 1;
                foreach (BaseUITab childUITab in childUITabs) count += 1;

                return count;
            }
            public void CloseAllPopup(UIDynamicPopup exception)
            {
                foreach (UIDynamicPopup UIDpopup in UIDPopupList)
                {
                    if (UIDpopup != exception) UIDpopup.popup.visible = false;
                }
                foreach (CustomUIElement customElement in UIDCustomElementList)
                {
                    customElement.CloseAllPopups(exception);
                }
                foreach (BaseUITab childUITab in childUITabs)
                {
                    childUITab.CloseAllPopup(exception);
                }
            }
        }

        private Dictionary<int, UIDynamicElementLists> leftUIElementGroups;
        private Dictionary<int, UIDynamicElementLists> rightUIElementGroups;
        private Dictionary<int, List<JSONStorableParam>> leftJSPsGroups;
        private Dictionary<int, List<JSONStorableParam>> rightJSPsGroups;

        protected int uiGroupCount = 1;

        public BaseUITab()
        {
            leftUIElementGroups = new Dictionary<int, UIDynamicElementLists>();
            rightUIElementGroups = new Dictionary<int, UIDynamicElementLists>();
            leftJSPsGroups = new Dictionary<int, List<JSONStorableParam>>();
            rightJSPsGroups = new Dictionary<int, List<JSONStorableParam>>();
        }
        protected void ForceUIRefreshOnValChange(JSONStorableParam jsp, int leftRight, int uiRefreshGroup, BaseUITab uiTab = null)
        {
            GroupLR groupLR = new GroupLR();
            groupLR.groupID = uiRefreshGroup;
            groupLR.leftRight = leftRight;
            if (uiTab == null) groupLR.uiTab = this;
            else groupLR.uiTab = uiTab;

            switch (jsp.type)
            {
                case JSONStorable.Type.StringChooser:
                    RegisterJSSCRefreshUIonValChange(jsp as JSONStorableStringChooser, groupLR);
                    break;
                case JSONStorable.Type.String:
                    RegisterJSSRefreshUIonValChange(jsp as JSONStorableString, groupLR);
                    break;
                case JSONStorable.Type.Float:
                    RegisterJSFRefreshUIonValChange(jsp as JSONStorableFloat, groupLR);
                    break;
                case JSONStorable.Type.Bool:
                    RegisterJSBRefreshUIonValChange(jsp as JSONStorableBool, groupLR);
                    break;
                default:
                    if (UIAConsts.debugLogging) SuperController.LogMessage("UIA.BaseUITab.RegisterJSRefreshUIonValChange: unsupported JSONStorableParam type " + jsp.type.ToString());
                    break;
            }
        }
        private void RegisterJSSCRefreshUIonValChange(JSONStorableStringChooser jssc, GroupLR groupLR)
        {

            if (jssc.popup == null)
            {
                if (UIAConsts.debugLogging) SuperController.LogMessage("UIA.BaseUITab.RegisterJSSCRefreshUIonValChange - attempt to register JSSC with no associated UI Popup");
                return;
            }
            jssc.popup.onValueChangeHandlers += groupLR.RefreshGroupUIStringChange;
        }

        private void RegisterJSBRefreshUIonValChange(JSONStorableBool jsb, GroupLR groupLR)
        {
            if (jsb.toggle == null)
            {
                if (UIAConsts.debugLogging) SuperController.LogMessage("UIA.BaseUITab.RegisterJSBRefreshUIonValChange - attempt to register JSB with no associated UI Toggle");
                return;
            }
            jsb.toggle.onValueChanged.AddListener((value) => groupLR.RefreshGroupUIStringChange(""));
        }

        private void RegisterJSSRefreshUIonValChange(JSONStorableString jss, GroupLR groupLR)
        {
            if (jss.inputField == null)
            {
                if (UIAConsts.debugLogging) SuperController.LogMessage("UIA.BaseUITab.RegisterJSSRefreshUIonValChange - attempt to register JSS with no associated inputfield");
                return;
            }
            jss.inputField.onValueChanged.AddListener((text) => groupLR.RefreshGroupUIStringChange(text));

        }

        private void RegisterJSFRefreshUIonValChange(JSONStorableFloat jsf, GroupLR groupLR)
        {

            if (jsf.slider == null)
            {
                if (UIAConsts.debugLogging) SuperController.LogMessage("UIA.BaseUITab.RegisterJSFRefreshUIonValChange - attempt to register JSF with no associated Slider");
                return;
            }
            jsf.slider.onValueChanged.AddListener((floatVal) => groupLR.RefreshGroupUIStringChange(floatVal.ToString()));

        }

        protected abstract void CreateGroupUI(bool isRight, int groupID);

        public void RefreshGroupUI(int leftRight, int startGroupID)
        {
            if (leftRight == LeftRight.both)
            {
                RefreshGroupUI(LeftRight.right, startGroupID);
                RefreshGroupUI(LeftRight.left, startGroupID);

                return;
            }
            if (leftRight == LeftRight.left) CloseAllLeftPopup(null);
            if (leftRight == LeftRight.right) CloseAllRightPopup(null);
            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                ClearGroupUI(leftRight, groupID);
            }

            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                CreateGroupUI(leftRight == LeftRight.right, groupID);
            }
        }

        private UIDynamicElementLists GetUIElementsFromGroup(int groupID, Dictionary<int, UIDynamicElementLists> uiElementsGroups)
        {
            UIDynamicElementLists uiElementGroup;

            if (uiElementsGroups.ContainsKey(groupID)) uiElementGroup = uiElementsGroups[groupID];
            else
            {
                uiElementGroup = new UIDynamicElementLists();
                uiElementsGroups.Add(groupID, uiElementGroup);
            }
            return (uiElementGroup);
        }

        protected void RegisterCustomUIElement(int groupID, CustomUIElement customElement, bool isRight = false)
        {
            UIDynamicElementLists uiElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            if (isRight) uiElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            uiElements.UIDCustomElementList.Add(customElement);
        }

        protected void RegisterChildUITab(int groupID, BaseUITab childUITab, bool rightSide = false)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            if (rightSide) rightUIElements.childUITabs.Add(childUITab);
            else leftUIElements.childUITabs.Add(childUITab);
        }

        protected UIDynamic CreateSpacer(int groupID, float height = 10f, bool rightSide = false)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamic spacer = UIAGlobals.mvrScript.CreateSpacer(rightSide);
            if (rightSide) rightUIElements.UIDSpacersList.Add(spacer);
            else leftUIElements.UIDSpacersList.Add(spacer);
            spacer.height = height;
            return (spacer);
        }

        protected UIDynamicButton CreateButton(int groupID, string label, bool rightSide = false, UnityEngine.Events.UnityAction callback = null, float height = 50f)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicButton button = UIAGlobals.mvrScript.CreateButton(label, rightSide);
            button.height = height;
            if (rightSide) rightUIElements.UIDButtonsList.Add(button);
            else leftUIElements.UIDButtonsList.Add(button);

            if (callback != null) button.button.onClick.AddListener(callback);

            return (button);
        }

        protected UIDynamicSlider CreateSlider(int groupID, JSONStorableFloat jsonFloat, bool rightSide = false, string label = "")
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicSlider slider = UIAGlobals.mvrScript.CreateSlider(jsonFloat, rightSide);

            if (label != "") slider.label = label;

            if (rightSide)
            {
                rightUIElements.UIDSlidersList.Add(slider);
                if (!rightJSPsGroups.ContainsKey(groupID)) rightJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                rightJSPsGroups[groupID].Add(jsonFloat);
            }
            else
            {
                leftUIElements.UIDSlidersList.Add(slider);
                if (!leftJSPsGroups.ContainsKey(groupID)) leftJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                leftJSPsGroups[groupID].Add(jsonFloat);
            }
            return (slider);
        }
        protected UIDynamicColorPicker CreateColorPicker(int groupID, JSONStorableColor jsc, bool rightSide = false, string label = "")
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicColorPicker colorPicker = UIAGlobals.mvrScript.CreateColorPicker(jsc, rightSide);
            if (rightSide)
            {
                rightUIElements.UIDColorPickerList.Add(colorPicker);
                if (!rightJSPsGroups.ContainsKey(groupID)) rightJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                rightJSPsGroups[groupID].Add(jsc);
            }
            else
            {
                leftUIElements.UIDColorPickerList.Add(colorPicker);
                if (!leftJSPsGroups.ContainsKey(groupID)) leftJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                leftJSPsGroups[groupID].Add(jsc);
            }
            colorPicker.label = label;

            return (colorPicker);
        }
        protected UIDynamicToggle CreateToggle(int groupID, JSONStorableBool jBool, bool rightSide = false, string label = "", float height = 50f)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);
            UIDynamicToggle toggle = UIAGlobals.mvrScript.CreateToggle(jBool, rightSide);
            toggle.height = height;
            if (label != "") toggle.label = label;

            if (rightSide)
            {
                rightUIElements.UIDToggleList.Add(toggle);
                if (!rightJSPsGroups.ContainsKey(groupID)) rightJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                rightJSPsGroups[groupID].Add(jBool);
            }
            else
            {
                leftUIElements.UIDToggleList.Add(toggle);
                if (!leftJSPsGroups.ContainsKey(groupID)) leftJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                leftJSPsGroups[groupID].Add(jBool);
            }

            return (toggle);
        }
        protected UIDynamicTextField CreateTextField(int groupID, JSONStorableString jss, bool rightSide = false, float height = 100f, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "")
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicTextField textfield = UIAGlobals.mvrScript.CreateTextField(jss, rightSide);

            textfield.height = height;
            if (input)
            {
                textfield.backgroundColor = Color.white;
                var inputComponent = textfield.gameObject.AddComponent<InputField>();

                inputComponent.textComponent = textfield.UItext;
                inputComponent.lineType = inputFieldType;
                inputComponent.textComponent.resizeTextMaxSize = 30;
                inputComponent.textComponent.resizeTextForBestFit = true;
                jss.inputField = inputComponent;

                if (placeHolderText != "")
                {
                    var placeholderGo = new GameObject();
                    placeholderGo.transform.SetParent(textfield.gameObject.transform, false);

                    var placeholderRect = placeholderGo.AddComponent<RectTransform>();
                    placeholderRect.anchorMin = new Vector2(0f, 0f);
                    placeholderRect.anchorMax = new Vector2(1f, 1f);
                    placeholderRect.sizeDelta = new Vector2(1f, 1f);

                    Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");

                    Text placeholderUIText = placeholderGo.AddComponent<Text>();
                    placeholderUIText.raycastTarget = false;
                    placeholderUIText.font = selectedFont;
                    placeholderUIText.color = Color.gray;
                    placeholderUIText.fontStyle = FontStyle.Italic;
                    placeholderUIText.fontSize = 30;
                    placeholderUIText.alignment = TextAnchor.UpperLeft;
                    if (jss.val == "") placeholderUIText.text = placeHolderText;

                    inputComponent.onValueChanged.AddListener((text) =>
                    {
                        TextFieldValueChanged(inputComponent, placeholderUIText, jss, placeHolderText);
                    });
                }

            }

            if (rightSide)
            {
                rightUIElements.UIDTextFieldList.Add(textfield);
                if (!rightJSPsGroups.ContainsKey(groupID)) rightJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                rightJSPsGroups[groupID].Add(jss);
            }
            else
            {
                leftUIElements.UIDTextFieldList.Add(textfield);
                if (!leftJSPsGroups.ContainsKey(groupID)) leftJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                leftJSPsGroups[groupID].Add(jss);
            }

            return (textfield);
        }

        private void TextFieldValueChanged(InputField inputField, Text placeholderUIText, JSONStorableString jss, string placeHolderText)
        {
            if (inputField.text == "") placeholderUIText.text = placeHolderText;
            else placeholderUIText.text = "";
        }

        private List<string> GetChoices(JSONStorableStringChooser jssc, List<string> ignoreVals = null, List<string> ignoreValPatterns = null)
        {
            List<string> choices = new List<string>();
            foreach (string choice in jssc.choices)
            {
                if (ignoreVals != null && ignoreVals.Contains(choice)) continue;
                if (ignoreValPatterns != null)
                {
                    bool found = false;
                    foreach (string pattern in ignoreValPatterns)
                    {
                        if (choice.StartsWith(pattern))
                        {
                            found = true;
                            break;
                        }

                    }
                    if (found) continue;
                }
                choices.Add(choice);
            }
            return choices;
        }

        private UIDynamicPopup CreatePopupAuto(JSONStorableStringChooser jssc, bool rightSide = false, bool scrollable = false, float labelWidth = 150f, List<string> ignoreVals = null, List<string> ignoreValPatterns = null)
        {
            UIDynamicPopup popup;
            if (scrollable) popup = UIAGlobals.mvrScript.CreateScrollablePopup(jssc, rightSide);
            else popup = UIAGlobals.mvrScript.CreatePopup(jssc, rightSide);

            popup.popup.labelText.alignment = TextAnchor.UpperCenter;
            popup.popup.labelText.GetComponent<RectTransform>().anchorMax = new Vector2(0, 0.89f);

            {
                var btn = MVRScript.Instantiate(UIAGlobals.mvrScript.manager.configurableButtonPrefab);
                btn.SetParent(popup.transform, false);
                MVRScript.Destroy(btn.GetComponent<LayoutElement>());
                btn.GetComponent<UIDynamicButton>().label = "<";
                btn.GetComponent<UIDynamicButton>().button.onClick.AddListener(() =>
                {
                    List<string> choices = GetChoices(jssc, ignoreVals, ignoreValPatterns);
                    if (choices.Contains(jssc.val) && choices.IndexOf(jssc.val) > 0)
                    {
                        jssc.val = choices[choices.IndexOf(jssc.val) - 1];
                        if (jssc.displayChoices != null) jssc.popup.currentValue = jssc.displayChoices[jssc.choices.IndexOf(jssc.val)];
                        else jssc.popup.currentValue = jssc.val;
                        jssc.popup.onValueChangeHandlers.Invoke(jssc.val);
                    }
                });
                var prevBtnRect = btn.GetComponent<RectTransform>();
                prevBtnRect.pivot = new Vector2(0, 0);
                prevBtnRect.anchoredPosition = new Vector2(10f, 0);
                prevBtnRect.sizeDelta = new Vector2(0f, 0f);
                prevBtnRect.offsetMin = new Vector2(5f, 15f);
                prevBtnRect.offsetMax = new Vector2(80f, 90f);
                prevBtnRect.anchorMin = new Vector2(0f, 0f);
                prevBtnRect.anchorMax = new Vector2(0f, 0f);
            }

            {
                float widthOffset = labelWidth - 150f;
                var btn = MVRScript.Instantiate(UIAGlobals.mvrScript.manager.configurableButtonPrefab);
                btn.SetParent(popup.transform, false);
                MVRScript.Destroy(btn.GetComponent<LayoutElement>());
                btn.GetComponent<UIDynamicButton>().label = ">";
                btn.GetComponent<UIDynamicButton>().button.onClick.AddListener(() =>
                {
                    List<string> choices = GetChoices(jssc, ignoreVals, ignoreValPatterns);
                    if (choices.Contains(jssc.val) && choices.IndexOf(jssc.val) < choices.Count - 1)
                    {
                        jssc.val = choices[choices.IndexOf(jssc.val) + 1];
                        if (jssc.displayChoices != null) jssc.popup.currentValue = jssc.displayChoices[jssc.choices.IndexOf(jssc.val)];
                        else jssc.popup.currentValue = jssc.val;
                        jssc.popup.onValueChangeHandlers.Invoke(jssc.val);
                    }
                    //                   popup.popup.SetNextValue();
                });
                var prevBtnRect = btn.GetComponent<RectTransform>();
                prevBtnRect.pivot = new Vector2(0, 0);
                prevBtnRect.anchoredPosition = new Vector2(10f, 0);
                prevBtnRect.sizeDelta = new Vector2(0f, 0f);
                prevBtnRect.offsetMin = new Vector2(82f + widthOffset, 15f);
                prevBtnRect.offsetMax = new Vector2(157f + widthOffset, 90f);
                prevBtnRect.anchorMin = new Vector2(0f, 0f);
                prevBtnRect.anchorMax = new Vector2(0f, 0f);
            }

            return popup;
        }

        protected UIDynamicPopup CreatePopup(int groupID, JSONStorableStringChooser jssc, bool rightUI = false, float panelHeight = 400f, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool scrollable = true, float labelWidth = 150f, bool labelScrollButtons = false, List<string> labelScrollIgnoreVals = null, List<string> ignoreValPatterns = null, bool closeOtherPopups = true, bool setValFromChoices = false, float popupHeight = 80f)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicPopup dynamicPopup;
            if (labelScrollButtons) dynamicPopup = CreatePopupAuto(jssc, rightUI, scrollable, labelWidth, labelScrollIgnoreVals, ignoreValPatterns);
            else if (scrollable) dynamicPopup = UIAGlobals.mvrScript.CreateScrollablePopup(jssc, rightUI);
            else dynamicPopup = UIAGlobals.mvrScript.CreatePopup(jssc, rightUI);

            LayoutElement layoutElement = dynamicPopup.GetComponent<LayoutElement>();
            layoutElement.minHeight = popupHeight;
            dynamicPopup.height = popupHeight;

            dynamicPopup.popup.showSlider = false;
            dynamicPopup.popupPanelHeight = panelHeight;
            dynamicPopup.popup.selectColor = UIAConsts.UIhighlightColor;
            dynamicPopup.labelWidth = labelWidth;


            if (callbackMethod != null) dynamicPopup.popup.onOpenPopupHandlers += callbackMethod;

            if (closeOtherPopups) dynamicPopup.popup.onOpenPopupHandlers += () =>
            {
                UIAPluginUI.CloseAllPopups(dynamicPopup);
            };

            if (setValFromChoices)
            {
                if (jssc.choices != null && !jssc.choices.Contains(jssc.val) && jssc.choices.Count > 0)
                {
                    jssc.val = jssc.choices[0];
                }
            }

            if (label != null) { dynamicPopup.label = label; }

            if (rightUI)
            {
                rightUIElements.UIDPopupList.Add(dynamicPopup);
                if (!rightJSPsGroups.ContainsKey(groupID)) rightJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                rightJSPsGroups[groupID].Add(jssc);
            }
            else
            {
                leftUIElements.UIDPopupList.Add(dynamicPopup);
                if (!leftJSPsGroups.ContainsKey(groupID)) leftJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                leftJSPsGroups[groupID].Add(jssc);
            }


            return (dynamicPopup);
        }

        protected UIDynamicPopup CreateFilterablePopup(int groupID, JSONStorableStringChooser stringChooserJSON, bool rightUI = false, float panelHeight = 400f, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool closeOtherPopups = true, bool setValFromChoices = false)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicPopup dynamicPopup = UIAGlobals.mvrScript.CreateFilterablePopup(stringChooserJSON, rightUI);
            dynamicPopup.popupPanelHeight = panelHeight;

            if (callbackMethod != null) { dynamicPopup.popup.onOpenPopupHandlers += callbackMethod; }
            if (closeOtherPopups) dynamicPopup.popup.onOpenPopupHandlers += () =>
            {
                UIAPluginUI.CloseAllPopups(dynamicPopup);
            };

            if (label != null) { dynamicPopup.label = label; }

            if (setValFromChoices)
            {
                if (stringChooserJSON.choices != null && !stringChooserJSON.choices.Contains(stringChooserJSON.val) && stringChooserJSON.choices.Count > 0)
                {
                    stringChooserJSON.val = stringChooserJSON.choices[0];
                }
            }

            if (rightUI)
            {
                rightUIElements.UIDPopupList.Add(dynamicPopup);
                if (!rightJSPsGroups.ContainsKey(groupID)) rightJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                rightJSPsGroups[groupID].Add(stringChooserJSON);
            }
            else
            {
                leftUIElements.UIDPopupList.Add(dynamicPopup);
                if (!leftJSPsGroups.ContainsKey(groupID)) leftJSPsGroups.Add(groupID, new List<JSONStorableParam>());
                leftJSPsGroups[groupID].Add(stringChooserJSON);
            }

            return (dynamicPopup);
        }

        public void ClearAllGroupsUI(int leftRight)
        {
            CloseAllPopup(null);
            if (leftRight == LeftRight.right || leftRight == LeftRight.both)
            {
                foreach (KeyValuePair<int, UIDynamicElementLists> kvp in rightUIElementGroups) ClearGroupUI(LeftRight.right, kvp.Key);
            }
            if (leftRight == LeftRight.left || leftRight == LeftRight.both)
            {
                foreach (KeyValuePair<int, UIDynamicElementLists> kvp in leftUIElementGroups) ClearGroupUI(LeftRight.left, kvp.Key);
            }
        }
        public virtual void ClearGroupUI(int leftRight, int groupID)
        {
            if (leftRight == LeftRight.both)
            {
                ClearGroupUI(LeftRight.left, groupID);
                ClearGroupUI(LeftRight.right, groupID);
                return;
            }
            if (leftRight == LeftRight.neither) return;

            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            List<JSONStorableParam> jspListForGroup = null;

            if (leftRight == LeftRight.right)
            {
                rightUIElements.ClearUIDs();
                if (rightJSPsGroups.ContainsKey(groupID)) jspListForGroup = rightJSPsGroups[groupID];
            }
            if (leftRight == LeftRight.left)
            {
                leftUIElements.ClearUIDs();
                if (leftJSPsGroups.ContainsKey(groupID)) jspListForGroup = leftJSPsGroups[groupID];
            }

            if (jspListForGroup != null)
            {
                jspListForGroup.Clear();
            }

        }
        public void CloseAllLeftPopup(UIDynamicPopup exception)
        {
            foreach (var leftUIElements in leftUIElementGroups.Values.ToList()) leftUIElements.CloseAllPopup(exception);
        }
        public void CloseAllRightPopup(UIDynamicPopup exception)
        {
            foreach (var rightUIElements in rightUIElementGroups.Values.ToList()) rightUIElements.CloseAllPopup(exception);
        }

        public void CloseAllPopup(UIDynamicPopup exception)
        {
            CloseAllLeftPopup(exception);
            CloseAllRightPopup(exception);
        }

        protected float GetTotalUIElementHeights(bool isRight)
        {
            Dictionary<int, UIDynamicElementLists> uiElementsGroup;

            if (!isRight) uiElementsGroup = leftUIElementGroups;
            else uiElementsGroup = rightUIElementGroups;

            float height = 0f;
            int count = 0;

            foreach (UIDynamicElementLists uidElementsList in uiElementsGroup.Values.ToList())
            {
                height += uidElementsList.GetUIElementsHeight(isRight);
                count += uidElementsList.GetUIElementsCount();
            }
            height += (count) * 15f;

            return (height);
        }

    }

    public abstract class BaseUIHelpTab : BaseUITab
    {
        protected JSONStorableString currentHelpTextJSS;
        protected string defaultHelpText = "";

        public BaseUIHelpTab()
        {
            currentHelpTextJSS = new JSONStorableString("currentHelpText", "");
        }
        protected void CreateHelpTextUI(int groupID, int height = 990)
        {
            if (groupID == 1)
            {
                base.CreateTextField(groupID, currentHelpTextJSS, true, height);
                RegisterContextualHelpForTopTabs();
            }
        }

        protected void RegisterContextualHelpForTopTabs()
        {
            RegisterContextualHelp(UIAPluginUI.buttonSetupUID.gameObject, TabButtonsUIHelp.buttonSetupTab);
            RegisterContextualHelp(UIAPluginUI.loadProfileUID.gameObject, TabButtonsUIHelp.loadProfileTab);
            RegisterContextualHelp(UIAPluginUI.saveProfileUID.gameObject, TabButtonsUIHelp.saveProfileTab);
            RegisterContextualHelp(UIAPluginUI.toolsUID.gameObject, TabButtonsUIHelp.toolsTab);
            RegisterContextualHelp(UIAPluginUI.settingsUID.gameObject, TabButtonsUIHelp.settingsTab);
        }

        public void RegisterContextualHelp(GameObject go, string helpText)
        {
            EventTrigger trigger = go.AddComponent<EventTrigger>();
            var pointerEnter = new EventTrigger.Entry();
            pointerEnter.eventID = EventTriggerType.PointerEnter;
            pointerEnter.callback.AddListener((e) => { currentHelpTextJSS.val = helpText; });
            trigger.triggers.Add(pointerEnter);

            var pointerExit = new EventTrigger.Entry();
            pointerExit.eventID = EventTriggerType.PointerExit;
            pointerExit.callback.AddListener((e) => { currentHelpTextJSS.val = defaultHelpText; });
            trigger.triggers.Add(pointerExit);
        }

        protected UIDynamicButton CreateButton(int groupID, string label, bool rightSide = false, UnityAction callback = null, float height = 50, string helpText = "")
        {
            var button = base.CreateButton(groupID, label, rightSide, callback, height);
            if (helpText != "") RegisterContextualHelp(button.gameObject, helpText);
            return button;
        }

        protected UIDynamicSlider CreateSlider(int groupID, JSONStorableFloat jsonFloat, bool rightSide = false, string label = "", string helpText = "")
        {
            var slider = base.CreateSlider(groupID, jsonFloat, rightSide, label);
            if (helpText != "") RegisterContextualHelp(slider.gameObject, helpText);
            return slider;
        }

        protected UIDynamicColorPicker CreateColorPicker(int groupID, JSONStorableColor jsc, bool rightSide = false, string label = "", string helpText = "")
        {
            var colorPicker = base.CreateColorPicker(groupID, jsc, rightSide, label);
            if (helpText != "") RegisterContextualHelp(colorPicker.gameObject, helpText);
            return colorPicker;
        }

        protected UIDynamicToggle CreateToggle(int groupID, JSONStorableBool jBool, bool rightSide = false, string label = "", float height = 50, string helpText = "")
        {
            var toggle = base.CreateToggle(groupID, jBool, rightSide, label, height);
            if (helpText != "") RegisterContextualHelp(toggle.gameObject, helpText);

            return (toggle);
        }

        protected UIDynamicTextField CreateTextField(int groupID, JSONStorableString jss, bool rightSide = false, float height = 100, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "", string helpText = "")
        {
            var textField = base.CreateTextField(groupID, jss, rightSide, height, input, inputFieldType, placeHolderText);
            if (helpText != "") RegisterContextualHelp(textField.gameObject, helpText);

            return (textField);
        }

        protected UIDynamicPopup CreateFilterablePopup(int groupID, JSONStorableStringChooser stringChooserJSON, bool rightUI = false, float panelHeight = 400, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool closeOtherPopups = true, bool setValFromChoices = false, string helpText = "")
        {
            var dynamicPopup = base.CreateFilterablePopup(groupID, stringChooserJSON, rightUI, panelHeight, callbackMethod, label, closeOtherPopups, setValFromChoices);

            if (helpText != "")
            {
                RegisterContextualHelp(dynamicPopup.gameObject, helpText);
                RegisterContextualHelp(dynamicPopup.popup.popupPanel.gameObject, helpText);
            }
            return dynamicPopup;
        }

        protected UIDynamicPopup CreatePopup(int groupID, JSONStorableStringChooser jssc, bool rightUI = false, float popupPanelHeight = 400, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool scrollable = true, string helpText = "", float labelWidth = 150, bool labelScrollButtons = false, List<string> labelScrollIgnoreVals = null, List<string> ignoreValPatterns = null, bool closeOtherPopups = true, bool setValFromChoices = false, float popupHeight = 80f)
        {
            var dynamicPopup = base.CreatePopup(groupID, jssc, rightUI, popupPanelHeight, callbackMethod, label, scrollable, labelWidth, labelScrollButtons, labelScrollIgnoreVals, ignoreValPatterns, closeOtherPopups, setValFromChoices, popupHeight);

            if (helpText != "")
            {
                RegisterContextualHelp(dynamicPopup.gameObject, helpText);
                RegisterContextualHelp(dynamicPopup.popup.popupPanel.gameObject, helpText);
            }
            return dynamicPopup;
        }
    }

    public abstract class BaseUIButtonTreeTab : BaseUIHelpTab
    {
        protected int helpExpandedMode { get { return UIAStorables.uiHelpExpandMode; } set { UIAStorables.uiHelpExpandMode = value; } }

        private CustomRowUIElement toggleHelpTreeRatioButtonUI;
        protected CustomTreeNavUIElement treeNavigationUI;
        private const int toggleHelpTreeRatioButtonHeight = 30;

        public BaseUIButtonTreeTab()
        {
            toggleHelpTreeRatioButtonUI = new CustomRowUIElement(toggleHelpTreeRatioButtonHeight, false);
            treeNavigationUI = new CustomTreeNavUIElement();
        }
        protected void CreateHelpAndButtonTree(int groupID, int height = 990)
        {

            int helpControlHeight = 0;
            int treeControlHeight = 0;

            switch (helpExpandedMode)
            {
                case UIHelpExpandMode.minimized:
                    treeControlHeight = height - toggleHelpTreeRatioButtonHeight + 15;
                    break;
                case UIHelpExpandMode.half:
                    treeControlHeight = (int)(Mathf.Round(height - toggleHelpTreeRatioButtonHeight) * 0.6f);
                    helpControlHeight = height - treeControlHeight - toggleHelpTreeRatioButtonHeight;
                    break;
                case UIHelpExpandMode.full:
                    helpControlHeight = height - toggleHelpTreeRatioButtonHeight + 15;
                    break;

            }

            if (groupID == 1)
            {
                if (helpExpandedMode != UIHelpExpandMode.minimized) CreateTextField(groupID, currentHelpTextJSS, true, helpControlHeight);
                toggleHelpTreeRatioButtonUI.InitUI(2, false, false, true);
                if (helpExpandedMode != UIHelpExpandMode.full)
                {
                    UIDynamicButton button = toggleHelpTreeRatioButtonUI.AddGridlessButtonToRow("v", 2, AlterMode, 40, 435, 15f);
                    button.buttonText.fontSize = 24;

                }
                if (helpExpandedMode != UIHelpExpandMode.minimized)
                {
                    UIDynamicButton button = toggleHelpTreeRatioButtonUI.AddGridlessButtonToRow("^", 1, AlterMode, 40, 480, 15f);
                    button.buttonText.fontSize = 24;
                }

                RegisterCustomUIElement(groupID, toggleHelpTreeRatioButtonUI, true);

                if (helpExpandedMode != UIHelpExpandMode.full)
                {
                    treeNavigationUI.InitUI(treeControlHeight, true);
                    RegisterCustomUIElement(groupID, treeNavigationUI, true);
                }

                RegisterContextualHelpForTopTabs();
                RegisterContextualHelp(ButtonSetupPluginUI.gridSelector.gameObject, TabButtonsUIHelp.gridSelector);
                RegisterContextualHelp(ButtonSetupPluginUI.buttonSelector.gameObject, TabButtonsUIHelp.buttonSelector);
                if (ButtonSetupPluginUI.operationSelector != null) RegisterContextualHelp(ButtonSetupPluginUI.operationSelector.gameObject, TabButtonsUIHelp.operationSelector);
            }
        }

        private void AlterMode(int buttonID)
        {
            if (buttonID == 1)
            {
                if (helpExpandedMode == UIHelpExpandMode.full) helpExpandedMode = UIHelpExpandMode.half;
                else if (helpExpandedMode == UIHelpExpandMode.half) helpExpandedMode = UIHelpExpandMode.minimized;
            }
            else if (buttonID == 2)
            {
                if (helpExpandedMode == UIHelpExpandMode.half) helpExpandedMode = UIHelpExpandMode.full;
                else if (helpExpandedMode == UIHelpExpandMode.minimized) helpExpandedMode = UIHelpExpandMode.half;
            }
            RefreshGroupUI(LeftRight.both, 1);
        }





    }
    public static class UIAPluginUI
    {
        public static CustomRowUIElement topTabMenuOptions;

        private static int currentTopTabSelection = TopUITabs.buttonSetup;

        private static BaseUITab currentTabUI;
        public static ButtonSetupPluginUI buttonSetupUI;
        private static SettingsPluginUI settingsUI;
        private static ToolsPluginUI toolsUI;
        private static SaveProfileUITab saveProfileUITab;
        private static LoadProfileUITab loadProfileUITab;

        public static UIDynamicButton buttonSetupUID;
        public static UIDynamicButton loadProfileUID;
        public static UIDynamicButton saveProfileUID;
        public static UIDynamicButton toolsUID;
        public static UIDynamicButton settingsUID;

        public static void InitUI()
        {
            topTabMenuOptions = new CustomRowUIElement();
            buttonSetupUI = new ButtonSetupPluginUI();
            settingsUI = new SettingsPluginUI();
            toolsUI = new ToolsPluginUI();
            saveProfileUITab = new SaveProfileUITab();
            loadProfileUITab = new LoadProfileUITab();

            topTabMenuOptions.InitUI(5);

            buttonSetupUID = topTabMenuOptions.AddButtonToRow("Button Setup", TopUITabs.buttonSetup, TopTabSelected);
            loadProfileUID = topTabMenuOptions.AddButtonToRow("Load Profile", TopUITabs.loadConfig, TopTabSelected);
            saveProfileUID = topTabMenuOptions.AddButtonToRow("Save Profile", TopUITabs.saveConfig, TopTabSelected);
            toolsUID = topTabMenuOptions.AddButtonToRow("Tools", TopUITabs.tools, TopTabSelected);
            settingsUID = topTabMenuOptions.AddButtonToRow("Settings", TopUITabs.settings, TopTabSelected);

            buttonSetupUI.RefreshGroupUI(LeftRight.left, 1);

            topTabMenuOptions.SetButtonColorActive(currentTopTabSelection);

            currentTabUI = buttonSetupUI;
        }

        public static void TopTabSelected(int tabRef)
        {
            if (tabRef != currentTopTabSelection)
            {
                ClearCurrentTopTabsUI();
                switch (tabRef)
                {
                    case TopUITabs.buttonSetup:
                        buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case TopUITabs.settings:
                        settingsUI.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case TopUITabs.tools:
                        toolsUI.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case TopUITabs.loadConfig:
                        loadProfileUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case TopUITabs.saveConfig:
                        saveProfileUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                }
                currentTopTabSelection = tabRef;
            }
        }

        private static void ClearCurrentTopTabsUI()
        {
            if (currentTopTabSelection == TopUITabs.buttonSetup)
            {
                currentTabUI.ClearAllGroupsUI(LeftRight.both);
            }
            if (currentTopTabSelection == TopUITabs.settings)
            {
                settingsUI.ClearAllGroupsUI(LeftRight.both);
            }
            if (currentTopTabSelection == TopUITabs.tools)
            {
                toolsUI.ClearAllGroupsUI(LeftRight.both);
            }
            if (currentTopTabSelection == TopUITabs.loadConfig)
            {
                loadProfileUITab.ClearAllGroupsUI(LeftRight.both);
            }
            if (currentTopTabSelection == TopUITabs.saveConfig)
            {
                saveProfileUITab.ClearAllGroupsUI(LeftRight.both);
            }
        }

        public static void CloseAllPopups(UIDynamicPopup exception)
        {
            currentTabUI.CloseAllPopup(exception);
        }
    }

    public class ButtonSetupPluginUI : BaseUITab
    {
        private static CustomRowUIElement buttonSelectorsUI;

        public static JSONStorableStringChooser gridSelectorJSSC;
        public static JSONStorableStringChooser buttonSelectorJSSC;
        public static JSONStorableStringChooser buttonOperationSelectorJSSC;

        private static CustomRowUIElement componentsTabSelectorUI;

        public static UIAButtonGrid currentUIAButtonGrid;
        public static UIAButton currentUIAButton;
        public static UIAButtonOperation currentUIAButtonOperation;

        public static UIDynamicPopup gridSelector;
        public static UIDynamicPopup buttonSelector;
        public static UIDynamicPopup operationSelector;

        public static int currentComponentTab { get; private set; }

        private static ButtonComponentUITab buttonComponentUITab;
        private static GridComponentUITab gridComponentUITab;
        private static SpawnAtomComponentUITab spawnAtomComponentUITab;
        private static RelativePositionComponentUITab relativePositionComponentUITab;
        private static MoveAtomComponentUITab moveAtomComponentUITab;
        private static FileReferencesComponentUITab fileReferencesComponentUITab;
        private static HairColorComponentUITab hairColorComponentUITab;
        private static PresetLockComponentUITab presetLockComponentUITab;
        private static PluginLoadComponentUITab pluginLoadComponentUITab;
        private static ButtonSkinComponentUITab buttonSkinComponentUITab;
        private static MergeClothingComponentUITab mergeClothingComponentUITab;
        private static DecalMakerComponentUITab decalMakerComponentUITab;
        private static PluginSettingsUITab pluginSettingsUITab;
        private static NodeControlComponentUITab nodeControlONComponentUITab;
        private static NodeControlComponentUITab nodeControlOFFComponentUITab;
        private static NodePhysicsComponentUITab nodePhysicsONComponentUITab;
        private static NodePhysicsComponentUITab nodePhysicsOFFComponentUITab;
        private static NodeSelectionComponentUITab nodeSelectionComponentUITab;

        public ButtonSetupPluginUI() : base()
        {
            buttonSelectorsUI = new CustomRowUIElement(80);
            componentsTabSelectorUI = new CustomRowUIElement();
            uiGroupCount = 3;

            gridSelectorJSSC = new JSONStorableStringChooser("gridSelector", null, "1", "Grid Selector", GridSelected);
            buttonSelectorJSSC = new JSONStorableStringChooser("buttonSelector", null, "R1C1", "Button Selector", ButtonSelected);
            buttonOperationSelectorJSSC = new JSONStorableStringChooser("buttonOperationSelector", null, "1", "Operation Selector", OperationSelected);
            ResetCurrentButtonSelections();

            buttonComponentUITab = new ButtonComponentUITab();
            gridComponentUITab = new GridComponentUITab();

            spawnAtomComponentUITab = new SpawnAtomComponentUITab();
            relativePositionComponentUITab = new RelativePositionComponentUITab();
            moveAtomComponentUITab = new MoveAtomComponentUITab();
            fileReferencesComponentUITab = new FileReferencesComponentUITab();
            hairColorComponentUITab = new HairColorComponentUITab();
            presetLockComponentUITab = new PresetLockComponentUITab();
            pluginLoadComponentUITab = new PluginLoadComponentUITab();
            pluginSettingsUITab = new PluginSettingsUITab();
            buttonSkinComponentUITab = new ButtonSkinComponentUITab();
            mergeClothingComponentUITab = new MergeClothingComponentUITab();
            decalMakerComponentUITab = new DecalMakerComponentUITab();

            nodeControlONComponentUITab = new NodeControlComponentUITab(false);
            nodeControlOFFComponentUITab = new NodeControlComponentUITab(true);

            nodePhysicsONComponentUITab = new NodePhysicsComponentUITab(false);
            nodePhysicsOFFComponentUITab = new NodePhysicsComponentUITab(true);
            nodeSelectionComponentUITab = new NodeSelectionComponentUITab();

            currentComponentTab = ButtonComponentTypes.button;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (groupID == 1 && !isRight)
            {
                buttonSelectorsUI.InitUI(3);

                RefreshGridChoices();
                gridSelector = buttonSelectorsUI.AddPopupToRow(gridSelectorJSSC, GridSelectorPopupOpen, 600f);
                buttonSelector = buttonSelectorsUI.AddPopupToRow(buttonSelectorJSSC, ButtonSelectorPopupOpen, 600f, true);
                if (currentUIAButton.buttonOperations.Count > 1)
                {
                    operationSelector = buttonSelectorsUI.AddPopupToRow(buttonOperationSelectorJSSC, ButtonOpSelectorPopupOpen, 600f, true);
                    ForceUIRefreshOnValChange(buttonOperationSelectorJSSC, LeftRight.left, 2);
                }
                else operationSelector = null;

                RegisterCustomUIElement(groupID, buttonSelectorsUI);

                ForceUIRefreshOnValChange(gridSelectorJSSC, LeftRight.left, 1);
                ForceUIRefreshOnValChange(buttonSelectorJSSC, LeftRight.left, 1);

            }
            if (groupID == 2 && !isRight)
            {
                Dictionary<int, string> componentUITabs = GetComponentUITabs();
                componentsTabSelectorUI.InitUI(componentUITabs.Count);

                foreach (KeyValuePair<int, string> kvp in componentUITabs)
                {
                    componentsTabSelectorUI.AddButtonToRow(kvp.Value, kvp.Key, ComponentTabSelected);
                }
                if (!componentUITabs.Keys.Contains(currentComponentTab)) currentComponentTab = ButtonComponentTypes.button;

                RegisterCustomUIElement(groupID, componentsTabSelectorUI);
            }
            if (groupID == 3 && !isRight)
            {
                Dictionary<int, string> componentUITabs = GetComponentUITabs();
                if (!componentUITabs.Keys.Contains(currentComponentTab)) currentComponentTab = ButtonComponentTypes.button;
                componentsTabSelectorUI.SetButtonColorActive(currentComponentTab);
                switch (currentComponentTab)
                {
                    case ButtonComponentTypes.button:
                        RegisterChildUITab(groupID, buttonComponentUITab, false);
                        buttonComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.grid:
                        RegisterChildUITab(groupID, gridComponentUITab, false);
                        gridComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.spawnAtomComponent:
                        RegisterChildUITab(groupID, spawnAtomComponentUITab, false);
                        spawnAtomComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.relativePositionComponent:
                        RegisterChildUITab(groupID, relativePositionComponentUITab, false);
                        relativePositionComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.moveAtomComponent:
                        RegisterChildUITab(groupID, moveAtomComponentUITab, false);
                        moveAtomComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.fileRefs:
                        RegisterChildUITab(groupID, fileReferencesComponentUITab, false);
                        fileReferencesComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.hairColorComponent:
                        RegisterChildUITab(groupID, hairColorComponentUITab, false);
                        hairColorComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.presetLockComponent:
                        RegisterChildUITab(groupID, presetLockComponentUITab, false);
                        presetLockComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.pluginsLoadComponent:
                        RegisterChildUITab(groupID, pluginLoadComponentUITab, false);
                        pluginLoadComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.pluginSettingComponent:
                        RegisterChildUITab(groupID, pluginSettingsUITab, false);
                        pluginSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.buttonSkinComponent:
                        RegisterChildUITab(groupID, buttonSkinComponentUITab, false);
                        buttonSkinComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.clothingComponent:
                        RegisterChildUITab(groupID, mergeClothingComponentUITab, false);
                        mergeClothingComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.decalMakerComponent:
                        RegisterChildUITab(groupID, decalMakerComponentUITab, false);
                        decalMakerComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.nodeControlOnComponent:
                        RegisterChildUITab(groupID, nodeControlONComponentUITab, false);
                        nodeControlONComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.nodeControlOffComponent:
                        RegisterChildUITab(groupID, nodeControlOFFComponentUITab, false);
                        nodeControlOFFComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.nodePhysicsOnComponent:
                        RegisterChildUITab(groupID, nodePhysicsONComponentUITab, false);
                        nodePhysicsONComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.nodePhysicsOffComponent:
                        RegisterChildUITab(groupID, nodePhysicsOFFComponentUITab, false);
                        nodePhysicsOFFComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case ButtonComponentTypes.nodeSelectionComponent:
                        RegisterChildUITab(groupID, nodeSelectionComponentUITab, false);
                        nodeSelectionComponentUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                }
            }
        }

        private Dictionary<int, string> GetComponentUITabs()
        {
            List<int> componentTypes = currentUIAButtonOperation.GetComponentTypes();
            Dictionary<int, string> componentUITabs = new Dictionary<int, string>();
            componentUITabs.Add(ButtonComponentTypes.grid, "Grid");


            componentUITabs.Add(ButtonComponentTypes.button, "Operation");

            if (componentTypes.Contains(ButtonComponentTypes.spawnAtomComponent)) componentUITabs.Add(ButtonComponentTypes.spawnAtomComponent, "Atom");
            if (componentTypes.Contains(ButtonComponentTypes.relativePositionComponent)) componentUITabs.Add(ButtonComponentTypes.relativePositionComponent, "Position");

            List<int> fileRefTypes = currentUIAButtonOperation.GetFileReferenceTypes();
            if (fileRefTypes.Count > 0)
            {
                int fileRefCategory = FileReferenceTypes.GetFileRefCategory(fileRefTypes[0]);
                if (fileRefCategory == FileReferenceCategory.sceneCat) componentUITabs.Add(ButtonComponentTypes.fileRefs, "Scene");
                else if (fileRefCategory == FileReferenceCategory.uiapCat) componentUITabs.Add(ButtonComponentTypes.fileRefs, "UIA Profile");
                else if (fileRefCategory == FileReferenceCategory.pluginCat) componentUITabs.Add(ButtonComponentTypes.fileRefs, "Plugins");
                else componentUITabs.Add(ButtonComponentTypes.fileRefs, "Presets");
            }

            if (componentTypes.Contains(ButtonComponentTypes.moveAtomComponent)) componentUITabs.Add(ButtonComponentTypes.moveAtomComponent, "Lock Movement");
            if (componentTypes.Contains(ButtonComponentTypes.hairColorComponent)) componentUITabs.Add(ButtonComponentTypes.hairColorComponent, "Hair Color");
            if (componentTypes.Contains(ButtonComponentTypes.presetLockComponent)) componentUITabs.Add(ButtonComponentTypes.presetLockComponent, "Preset Locks");

            if (componentTypes.Contains(ButtonComponentTypes.pluginsLoadComponent)) componentUITabs.Add(ButtonComponentTypes.pluginsLoadComponent, "Plugins Load");
            if (componentTypes.Contains(ButtonComponentTypes.pluginSettingComponent)) componentUITabs.Add(ButtonComponentTypes.pluginSettingComponent, "Plugin Control");
            if (componentTypes.Contains(ButtonComponentTypes.decalMakerComponent)) componentUITabs.Add(ButtonComponentTypes.decalMakerComponent, "Decal Maker");

            if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.mergeClothingPreset) componentUITabs.Add(ButtonComponentTypes.clothingComponent, "Remove Clothing");


            if (currentUIAButtonOperation.targetComponent.isPersonTargetType)
            {
                if (componentTypes.Contains(ButtonComponentTypes.nodeControlOnComponent) && currentUIAButtonOperation.nodeControlOnComponent.multiNodeControlStateJSEnum.val == MultiNodeControlState.custom)
                {
                    if (componentTypes.Contains(ButtonComponentTypes.nodeControlOnComponent))
                    {
                        if (componentTypes.Contains(ButtonComponentTypes.nodeControlOffComponent))
                            componentUITabs.Add(ButtonComponentTypes.nodeControlOnComponent, "On Node States");
                        else componentUITabs.Add(ButtonComponentTypes.nodeControlOnComponent, "Node States");
                    }
                }

                if (componentTypes.Contains(ButtonComponentTypes.nodeControlOffComponent) && currentUIAButtonOperation.nodeControlOffComponent.multiNodeControlStateJSEnum.val == MultiNodeControlState.custom)
                    componentUITabs.Add(ButtonComponentTypes.nodeControlOffComponent, "Off Node States");
            }
            if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.nodePhysics)
            {
                componentUITabs.Add(ButtonComponentTypes.nodePhysicsOnComponent, "Node Physics");
                componentUITabs.Add(ButtonComponentTypes.nodeSelectionComponent, "Nodes");
            }
            if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodePhysics)
            {
                componentUITabs.Add(ButtonComponentTypes.nodePhysicsOnComponent, "Physics ON");
                componentUITabs.Add(ButtonComponentTypes.nodePhysicsOffComponent, "Physics OFF");
                componentUITabs.Add(ButtonComponentTypes.nodeSelectionComponent, "Nodes");
            }

            if (componentTypes.Contains(ButtonComponentTypes.buttonSkinComponent)) componentUITabs.Add(ButtonComponentTypes.buttonSkinComponent, "Display");

            return (componentUITabs);
        }

        public static void ResetCurrentButtonSelections()
        {
            if (gridSelectorJSSC != null)
            {
                gridSelectorJSSC.valNoCallback = "1";
                GridSelected("1");
            }
        }
        private void GridSelectorPopupOpen()
        {
            RefreshGridChoices();
        }
        private void ButtonSelectorPopupOpen()
        {
            RefreshButtonChoices();
        }

        private void ButtonOpSelectorPopupOpen()
        {
            RefreshButtonOpChoices();
        }

        public static void RefreshGridChoices()
        {
            List<string> choices = new List<string>();
            List<string> displayChoices = new List<string>();

            if (PatreonFeatures.patreonContentEnabled && GameControlSettings.activateQuickLaunchBarJSB.val)
            {
                choices.Add("QL");
                displayChoices.Add("Quick Launch");
            }
            int i = 1;
            foreach (UIAButtonGrid bg in UIAStorables.buttonGridsList)
            {
                choices.Add(i.ToString());
                if (bg.gridLabelJSS.val != "") displayChoices.Add(bg.gridLabelJSS.val);
                else displayChoices.Add(i.ToString());
                i++;
            }
            gridSelectorJSSC.choices = choices;
            gridSelectorJSSC.displayChoices = displayChoices;
        }

        private static void RefreshButtonChoices()
        {
            List<string> choices = new List<string>();

            int i = 1;
            foreach (UIAButton uiaButton in currentUIAButtonGrid.buttonList)
            {
                choices.Add(uiaButton.buttonRef);

                if (currentUIAButtonGrid.buttonCount <= i) break;
                i++;

            }
            buttonSelectorJSSC.choices = choices;
        }

        private static void RefreshButtonOpChoices()
        {
            List<string> choices = new List<string>();

            int i = 1;
            foreach (UIAButtonOperation uiaButton in currentUIAButton.buttonOperations)
            {
                choices.Add(i.ToString());
                i++;
            }
            buttonOperationSelectorJSSC.choices = choices;
        }

        public void ComponentTabSelected(int tab)
        {
            currentComponentTab = tab;
            if (currentComponentTab == ButtonComponentTypes.grid) CustomTreeNavUIElement.PluginUIGridComponentTabSelected();
            RefreshGroupUI(LeftRight.both, 3);
        }

        public static void GridSelected(string gridRef)
        {

            if (gridRef == "QL") currentUIAButtonGrid = UIAStorables.quickLaunchButtonGrid;
            else currentUIAButtonGrid = UIAStorables.buttonGridsList[int.Parse(gridRef) - 1];

            buttonSelectorJSSC.valNoCallback = "R1C1";
            RefreshButtonChoices();
            ButtonSelected("R1C1");

            CustomTreeNavUIElement.PluginUIGridChange();

            RelativePositionComponentUITab.DeactivateSpawnTest();
        }

        private static void ButtonSelected(string buttonRCRef)
        {

            currentUIAButton = currentUIAButtonGrid.buttonList[currentUIAButtonGrid.GetButtonIndexFromRCRef(buttonRCRef)];
            currentUIAButtonOperation = currentUIAButton.buttonOperations[0];
            buttonOperationSelectorJSSC.valNoCallback = "1";
            RefreshButtonOpChoices();
            OperationSelected("1");

            CustomTreeNavUIElement.PluginUIButtonChange();

            RelativePositionComponentUITab.DeactivateSpawnTest();
        }

        public static void OperationSelected(string operationRef)
        {

            currentUIAButtonOperation = currentUIAButton.buttonOperations[int.Parse(operationRef) - 1];

            RelativePositionComponentUITab.DeactivateSpawnTest();
        }
    }

    public class ButtonComponentUITab : BaseUIButtonTreeTab
    {
        private UIAButton currentUIAButton { get { return ButtonSetupPluginUI.currentUIAButton; } }
        private UIAButtonOperation currentButtonOperation { get { return ButtonSetupPluginUI.currentUIAButtonOperation; } }

        private UIAButtonGrid currentUIAButtonGrid { get { return ButtonSetupPluginUI.currentUIAButtonGrid; } }
        private TargetButtonComponent targetButtonComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.targetComponent; } }

        private JSONStorableStringChooser switchUIAGridJSSC;

        private CustomRowUIElement addButtonOpUI;

        private string targetNameHelpText
        {
            get
            {
                int targetCategory = currentButtonOperation.targetComponent.targetCategoryJSEnum.val;
                if (targetCategory == TargetCategory.gazeSelectedAtom) return ButtonComponentUIHelp.targetNameGaze;
                if (targetCategory == TargetCategory.vamSelectedAtom) return ButtonComponentUIHelp.targetNameVAMSelect;
                if (targetCategory == TargetCategory.atomGroup) return ButtonComponentUIHelp.targetNameAtomGroup;
                if (targetCategory == TargetCategory.specificAtom) return ButtonComponentUIHelp.targetNameSpecificAtom;
                if (targetCategory == TargetCategory.userChosenAtom) return ButtonComponentUIHelp.targetNameUserChosen;
                return "";
            }
        }

        private string specificOpTypeHelpText
        {
            get
            {
                int buttonOpType = currentButtonOperation.buttonOpTypeJSEnum.val;
                int buttonOpCategory = currentButtonOperation.buttonOpCategoryJSEnum.val;

                if (buttonOpType == UIAButtonOpType.spawnAtom) return SpecificOpTypeUIHelp.spawnAtom;
                if (buttonOpType == UIAButtonOpType.toggleAtomOn) return SpecificOpTypeUIHelp.toggleAtomOn;
                if (buttonOpType == UIAButtonOpType.hideAtom) return SpecificOpTypeUIHelp.hideAtom;
                if (buttonOpType == UIAButtonOpType.deleteAtom) return SpecificOpTypeUIHelp.deleteAtom;
                if (buttonOpType == UIAButtonOpType.selectAtom) return SpecificOpTypeUIHelp.selectAtom;
                if (buttonOpType == UIAButtonOpType.moveAtom) return SpecificOpTypeUIHelp.moveAtom;
                if (buttonOpType == UIAButtonOpType.teleportAtom) return SpecificOpTypeUIHelp.teleportAtom;
                if (buttonOpType == UIAButtonOpType.toggleAtomCollision) return SpecificOpTypeUIHelp.toggleAtomCollision;
                if (buttonOpType == UIAButtonOpType.detachAtomRoot) return SpecificOpTypeUIHelp.detachAtomRoot;
                if (buttonOpType == UIAButtonOpType.togglePresetLocks) return SpecificOpTypeUIHelp.togglePresetLocks;
                if (buttonOpType == UIAButtonOpType.changeAtomName) return SpecificOpTypeUIHelp.changeAtomName;

                if (buttonOpType == UIAButtonOpType.nodePositionState) return SpecificOpTypeUIHelp.nodePositionState;
                if (buttonOpType == UIAButtonOpType.nodeRotationState) return SpecificOpTypeUIHelp.nodeRotationState;
                if (buttonOpType == UIAButtonOpType.toggleNodePositionState) return SpecificOpTypeUIHelp.toggleNodePositionState;
                if (buttonOpType == UIAButtonOpType.toggleNodeRotationState) return SpecificOpTypeUIHelp.toggleNodeRotationState;
                if (buttonOpType == UIAButtonOpType.nodePhysics) return SpecificOpTypeUIHelp.nodePhysics;
                if (buttonOpType == UIAButtonOpType.toggleNodePhysics) return SpecificOpTypeUIHelp.toggleNodePhysics;

                if (buttonOpType == UIAButtonOpType.resetScale) return SpecificOpTypeUIHelp.resetScale;
                if (buttonOpType == UIAButtonOpType.resetAppearance) return SpecificOpTypeUIHelp.resetAppearance;
                if (buttonOpType == UIAButtonOpType.decalMakerToggle) return SpecificOpTypeUIHelp.decalMakerToggle;
                if (buttonOpType == UIAButtonOpType.setHairColor) return SpecificOpTypeUIHelp.setHairColor;
                if (buttonOpType == UIAButtonOpType.suppressScaleLoad) return SpecificOpTypeUIHelp.suppressScaleLoad;
                if (buttonOpType == UIAButtonOpType.suppressClothingLoad) return SpecificOpTypeUIHelp.suppressClothingLoad;

                if (buttonOpType == UIAButtonOpType.removeAllClothing) return SpecificOpTypeUIHelp.removeAllClothing;
                if (buttonOpType == UIAButtonOpType.undressAllClothing) return SpecificOpTypeUIHelp.undressAllClothing;
                if (buttonOpType == UIAButtonOpType.resetSimAllClothing) return SpecificOpTypeUIHelp.resetSimAllClothing;
                if (buttonOpType == UIAButtonOpType.removeClothingPreset) return SpecificOpTypeUIHelp.removeClothingPreset;
                if (buttonOpType == UIAButtonOpType.mergeClothingPreset) return SpecificOpTypeUIHelp.mergeClothingPreset;
                if (buttonOpType == UIAButtonOpType.undressClothingPreset) return SpecificOpTypeUIHelp.undressClothingPreset;
                if (buttonOpType == UIAButtonOpType.resetSimClothingPreset) return SpecificOpTypeUIHelp.resetSimClothingPreset;
                if (buttonOpType == UIAButtonOpType.activeClothingEditor) return SpecificOpTypeUIHelp.activeClothingEditor;

                if (buttonOpCategory == UIAButtonCategory.presets && buttonOpType != UIAButtonOpType.loadGeneralPreset) return "'" + currentButtonOperation.buttonOpTypeJSEnum.displayVal + "'" + SpecificOpTypeUIHelp.loadPresets;
                if (buttonOpCategory == UIAButtonCategory.presets && buttonOpType == UIAButtonOpType.loadGeneralPreset) return SpecificOpTypeUIHelp.loadGeneralPreset;

                if (buttonOpCategory == UIAButtonCategory.legacyPresets)
                {
                    if (buttonOpType == UIAButtonOpType.loadLegacyLook || buttonOpType == UIAButtonOpType.loadLegacyPose || buttonOpType == UIAButtonOpType.loadLegacyPreset) return "'" + currentButtonOperation.buttonOpTypeJSEnum.displayVal + "'" + SpecificOpTypeUIHelp.loadLegacyPresets;
                    else return "'" + currentButtonOperation.buttonOpTypeJSEnum.displayVal + "'" + SpecificOpTypeUIHelp.saveLegacyPresets;
                }

                if (buttonOpType == UIAButtonOpType.pluginAction) return SpecificOpTypeUIHelp.pluginAction;
                if (buttonOpType == UIAButtonOpType.pluginActionToggle) return SpecificOpTypeUIHelp.pluginActionToggle;
                if (buttonOpType == UIAButtonOpType.pluginBoolToggle) return SpecificOpTypeUIHelp.pluginBoolToggle;
                if (buttonOpType == UIAButtonOpType.pluginBoolTrue) return SpecificOpTypeUIHelp.pluginBoolTrue;
                if (buttonOpType == UIAButtonOpType.pluginBoolFalse) return SpecificOpTypeUIHelp.pluginBoolFalse;
                if (buttonOpType == UIAButtonOpType.pluginMultiSettings) return SpecificOpTypeUIHelp.pluginMultiSettings;
                if (buttonOpType == UIAButtonOpType.pluginMultiSettingsToggle) return SpecificOpTypeUIHelp.pluginMultiSettingsToggle;

                if (buttonOpType == UIAButtonOpType.loadPlugins) return SpecificOpTypeUIHelp.loadPlugins;
                if (buttonOpType == UIAButtonOpType.unLoadPlugins) return SpecificOpTypeUIHelp.unLoadPlugins;
                if (buttonOpType == UIAButtonOpType.pluginAssistFind) return SpecificOpTypeUIHelp.pluginAssistFind;
                if (buttonOpType == UIAButtonOpType.pluginAssistLoad) return SpecificOpTypeUIHelp.pluginAssistLoad;

                if (buttonOpType == UIAButtonOpType.loadUIAProfile) return SpecificOpTypeUIHelp.loadUIAProfile;
                if (buttonOpType == UIAButtonOpType.gazeAssistedSelect) return SpecificOpTypeUIHelp.gazeAssistedSelect;
                if (buttonOpType == UIAButtonOpType.hideInactiveTargets) return SpecificOpTypeUIHelp.hideInactiveTargets;
                if (buttonOpType == UIAButtonOpType.raiseByHAHeight) return SpecificOpTypeUIHelp.raiseByHAHeight;
                if (buttonOpType == UIAButtonOpType.suppressPresetLocks) return SpecificOpTypeUIHelp.suppressPresetLocks;
                if (buttonOpType == UIAButtonOpType.switchUIAGrid) return SpecificOpTypeUIHelp.switchUIAGrid;
                if (buttonOpType == UIAButtonOpType.switchUIAGrid) return SpecificOpTypeUIHelp.hideUGameControlUI;

                if (buttonOpType == UIAButtonOpType.startPlayback) return SpecificOpTypeUIHelp.startPlayback;
                if (buttonOpType == UIAButtonOpType.stopPlayback) return SpecificOpTypeUIHelp.stopPlayback;
                if (buttonOpType == UIAButtonOpType.resetAnimation) return SpecificOpTypeUIHelp.resetAnimation;
                if (buttonOpType == UIAButtonOpType.toggleLoopPlayback) return SpecificOpTypeUIHelp.toggleLoopPlayback;
                if (buttonOpType == UIAButtonOpType.beginRecordMode) return SpecificOpTypeUIHelp.beginRecordMode;
                if (buttonOpType == UIAButtonOpType.stopRecording) return SpecificOpTypeUIHelp.stopRecording;
                if (buttonOpType == UIAButtonOpType.loadMotionCapture) return SpecificOpTypeUIHelp.loadMotionCapture;
                if (buttonOpType == UIAButtonOpType.loadELMotionCapture) return SpecificOpTypeUIHelp.loadELMotionCapture;

                if (buttonOpType == UIAButtonOpType.softBodyPhysicsToggle) return SpecificOpTypeUIHelp.softBodyPhysicsToggle;
                if (buttonOpType == UIAButtonOpType.desktopVSyncToggle) return SpecificOpTypeUIHelp.desktopVSyncToggle;
                if (buttonOpType == UIAButtonOpType.perfMon) return SpecificOpTypeUIHelp.perfMon;
                if (buttonOpType == UIAButtonOpType.showHiddenAtoms) return SpecificOpTypeUIHelp.showHiddenAtoms;

                if (buttonOpType == UIAButtonOpType.loadScene) return SpecificOpTypeUIHelp.loadScene;
                if (buttonOpType == UIAButtonOpType.mergeLoadScene) return SpecificOpTypeUIHelp.mergeLoadScene;
                if (buttonOpType == UIAButtonOpType.loadSubScene) return SpecificOpTypeUIHelp.loadSubScene;

                if (buttonOpType == UIAButtonOpType.showVRHands) return SpecificOpTypeUIHelp.showVRHands;
                if (buttonOpType == UIAButtonOpType.leapMotionEnable) return SpecificOpTypeUIHelp.leapMotionEnable;
                if (buttonOpType == UIAButtonOpType.leapMotionToggle) return SpecificOpTypeUIHelp.leapMotionToggle;
                if (buttonOpType == UIAButtonOpType.vrHeadCollider) return SpecificOpTypeUIHelp.vrHeadCollider;

                if (buttonOpType == UIAButtonOpType.freezeMotionSound) return SpecificOpTypeUIHelp.freezeMotionSound;
                if (buttonOpType == UIAButtonOpType.togglePlayEdit) return SpecificOpTypeUIHelp.togglePlayEdit;
                if (buttonOpType == UIAButtonOpType.setWorldScale) return SpecificOpTypeUIHelp.setWorldScale;
                if (buttonOpType == UIAButtonOpType.offsetWorldScale) return SpecificOpTypeUIHelp.offsetWorldScale;
                if (buttonOpType == UIAButtonOpType.teleportPlayer) return SpecificOpTypeUIHelp.teleportPlayer;
                if (buttonOpType == UIAButtonOpType.blank) return SpecificOpTypeUIHelp.blank;

                if (buttonOpType == UIAButtonOpType.shaderQaulity) return SpecificOpTypeUIHelp.shaderQaulity;
                if (buttonOpType == UIAButtonOpType.msaaLevel) return SpecificOpTypeUIHelp.msaaLevel;
                if (buttonOpType == UIAButtonOpType.pixelLightCount) return SpecificOpTypeUIHelp.pixelLightCount;
                if (buttonOpType == UIAButtonOpType.smoothPasses) return SpecificOpTypeUIHelp.smoothPasses;
                if (buttonOpType == UIAButtonOpType.glowEffects) return SpecificOpTypeUIHelp.glowEffects;
                if (buttonOpType == UIAButtonOpType.realtimeReflectionProbesToggle) return SpecificOpTypeUIHelp.realtimeReflectionProbesToggle;
                if (buttonOpType == UIAButtonOpType.mirrorReflectionsToggle) return SpecificOpTypeUIHelp.mirrorReflectionsToggle;
                if (buttonOpType == UIAButtonOpType.physicsRate) return SpecificOpTypeUIHelp.physicsRate;
                if (buttonOpType == UIAButtonOpType.physicsUpdateCap) return SpecificOpTypeUIHelp.physicsUpdateCap;
                if (buttonOpType == UIAButtonOpType.hqPhysicsToggle) return SpecificOpTypeUIHelp.hqPhysicsToggle;

                if (buttonOpType == UIAButtonOpType.softBodyPhysics) return SpecificOpTypeUIHelp.softBodyPhysics;
                if (buttonOpType == UIAButtonOpType.desktopVSync) return SpecificOpTypeUIHelp.desktopVSync;
                if (buttonOpType == UIAButtonOpType.hqPhysics) return SpecificOpTypeUIHelp.hqPhysics;
                if (buttonOpType == UIAButtonOpType.realtimeReflectionProbes) return SpecificOpTypeUIHelp.realtimeReflectionProbes;
                if (buttonOpType == UIAButtonOpType.mirrorReflections) return SpecificOpTypeUIHelp.mirrorReflections;
                if (buttonOpType == UIAButtonOpType.pluginOpenUI) return SpecificOpTypeUIHelp.pluginOpenUI;
                if (buttonOpType == UIAButtonOpType.pluginToggleOpenUI) return SpecificOpTypeUIHelp.pluginToggleOpenUI;

                return "";
            }
        }

        private string specificOpCategoryHelpText { get
            {
                int buttonOpCategory = currentButtonOperation.buttonOpCategoryJSEnum.val;
                if (buttonOpCategory == UIAButtonCategory.atomControl) return SpecificOpCategoryUIHelp.atomControlOphelp;
                if (buttonOpCategory == UIAButtonCategory.nodeControl) return SpecificOpCategoryUIHelp.nodeControlOphelp;
                if (buttonOpCategory == UIAButtonCategory.appearance) return SpecificOpCategoryUIHelp.appearanceOphelp;
                if (buttonOpCategory == UIAButtonCategory.clothing) return SpecificOpCategoryUIHelp.clothingOphelp;
                if (buttonOpCategory == UIAButtonCategory.presets) return SpecificOpCategoryUIHelp.presetsOphelp;
                if (buttonOpCategory == UIAButtonCategory.legacyPresets) return SpecificOpCategoryUIHelp.legacyPresetsOphelp;
                if (buttonOpCategory == UIAButtonCategory.pluginSettings) return SpecificOpCategoryUIHelp.pluginSettingsOphelp;
                if (buttonOpCategory == UIAButtonCategory.pluginsLoading) return SpecificOpCategoryUIHelp.pluginsLoadingOphelp;
                if (buttonOpCategory == UIAButtonCategory.uiAssist) return SpecificOpCategoryUIHelp.uiAssistOphelp;
                if (buttonOpCategory == UIAButtonCategory.sceneAnim) return SpecificOpCategoryUIHelp.sceneAnimOphelp;
                if (buttonOpCategory == UIAButtonCategory.vamUserSettings) return SpecificOpCategoryUIHelp.vamUserSettingsOphelp;
                if (buttonOpCategory == UIAButtonCategory.sceneLoading) return SpecificOpCategoryUIHelp.sceneLoadingOphelp;
                if (buttonOpCategory == UIAButtonCategory.vrSettings) return SpecificOpCategoryUIHelp.vrSettingsOphelp;
                if (buttonOpCategory == UIAButtonCategory.misc) return SpecificOpCategoryUIHelp.miscOphelp;

                return "";
            }
        }

        public ButtonComponentUITab() : base()
        {
            uiGroupCount = 6;
            defaultHelpText = ButtonComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            switchUIAGridJSSC = new JSONStorableStringChooser("switchUIAGrid", null, "1", "", SwitchUIAGridUpdated);

            addButtonOpUI = new CustomRowUIElement(80, false);
        }


        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            defaultHelpText = ButtonComponentUIHelp.defaultHelpText + "\n\n" + specificOpTypeHelpText;
            currentHelpTextJSS.val = defaultHelpText;
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                UIDynamicPopup popup = CreatePopup(groupID, currentButtonOperation.buttonOpCategoryJSEnum, false, 820, null, "Operation Category", false, helpText: ButtonComponentUIHelp.opCategoryGenericText + specificOpCategoryHelpText);
                popup.popup.onValueChangeHandlers += ButtonCategoryChanged;
                ForceUIRefreshOnValChange(currentButtonOperation.buttonOpCategoryJSEnum, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);
            }
            if (groupID == 2 && !isRight)
            {
                ForceButtonTypeEnumRefresh();
                if (currentButtonOperation.buttonOpTypeJSEnum.val != UIAButtonOpType.blank)
                {
                    CreatePopup(groupID, currentButtonOperation.buttonOpTypeJSEnum, false, 600, null, "Operation Type", currentButtonOperation.buttonOpTypeJSEnum.choices.Count > 12, helpText: ButtonComponentUIHelp.opTypeGenericText + specificOpTypeHelpText);
                    ForceUIRefreshOnValChange(currentButtonOperation.buttonOpTypeJSEnum, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);

                    if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.changeAtomName)
                    {
                        CreateTextField(groupID, currentButtonOperation.targetComponent.changeAtomNameJSS, input: true, inputFieldType: InputField.LineType.SingleLine, placeHolderText: "Insert new name for target Atom", helpText: ButtonComponentUIHelp.changeAtomNameNewName);
                    }
                }
                else currentButtonOperation.buttonOpCategoryJSEnum.popup.currentValueNoCallback = "Select.....";
            }
            if (groupID == 3 && !isRight)
            {
                if (currentButtonOperation.buttonOpCategoryJSEnum.val == UIAButtonCategory.presets)
                {
                    CreateToggle(groupID, currentButtonOperation.savePresetJSB, false, "Switch Operation to Save Preset", helpText: SpecificOpTypeUIHelp.savePresetToggle);
                }
                if (currentButtonOperation.IsComponentInButtonOpType(ButtonComponentTypes.targetComponent))
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, targetButtonComponent.targetCategoryJSEnum, false, 850f, null, "Target Category", false, helpText: ButtonComponentUIHelp.targetCategory);
                    ForceUIRefreshOnValChange(targetButtonComponent.targetCategoryJSEnum, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);
                }
                

            }
            if (groupID == 4 && !isRight)
            {
                if (currentButtonOperation.IsComponentInButtonOpType(ButtonComponentTypes.targetComponent))
                {
                    if (targetButtonComponent.targetCategoryJSEnum.val != TargetCategory.sessionPlugins && targetButtonComponent.targetCategoryJSEnum.val != TargetCategory.scenePlugins)
                    {
                        CreatePopup(groupID, targetButtonComponent.targetNameJSMultiEnum, false, 550f, RefreshTargetNameChoices, "Target", targetButtonComponent.targetNameJSMultiEnum.choices.Count > 9, helpText: targetNameHelpText);

                        if (targetButtonComponent.targetCategoryJSEnum.val == TargetCategory.userChosenAtom && currentButtonOperation.buttonOpTypeJSEnum.val != UIAButtonOpType.activeClothingEditor)
                        {
                            CreateToggle(groupID, targetButtonComponent.keepOpenAtomSelectorJSB, false, "Keep Target Atom Selector open after each selection",80f, ButtonComponentUIHelp.keepOpenAtomSelector);
                            CreateToggle(groupID, targetButtonComponent.keepOpenOptionAtomSelectorJSB, false, "Display 'Keep Open' option on Target Atom Selector",80f, ButtonComponentUIHelp.keepOpenOptionAtomSelector);
                        }

                        if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.removeAllClothing && UIAGlobals.isRealClothingAvailable)
                        {
                            CreateSpacer(groupID);
                            CreateToggle(groupID, currentButtonOperation.clothingComponent.onlyRemoveRealClothingJSB, false, "Only remove 'Real' Clothing", helpText: ButtonComponentUIHelp.onlyRemoveRealClothing);
                        }

                        ForceUIRefreshOnValChange(targetButtonComponent.targetNameJSMultiEnum, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);
                    }


                }

            }
            if (groupID == 5 && !isRight)
            {
                if (targetButtonComponent.SpawnAtomIfTargetMissingToggleAvailable())
                {
                    CreateToggle(groupID, targetButtonComponent.spawnAtomIfTargetMissingJSBool, false, "Spawn Atom if target missing", helpText: ButtonComponentUIHelp.spawnAtomIfMissing);
                    ForceUIRefreshOnValChange(targetButtonComponent.spawnAtomIfTargetMissingJSBool, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);
                }

                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.togglePlayEdit)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.vamPlayEditModeComponent.closeGameUIonPlayModeJSBool, false, "Close Game UI for Play Mode", helpText: ButtonComponentUIHelp.closeGameUIPlay);
                    CreateToggle(groupID, currentButtonOperation.vamPlayEditModeComponent.openGameUIonEditModeJSBool, false, "Open Game UI for Edit Mode", helpText: ButtonComponentUIHelp.openGameUIEdit);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.offsetWorldScale)
                {
                    CreateSpacer(groupID);
                    CreateSlider(groupID, currentButtonOperation.worldScaleComponent.worldScaleIncrementJSF, false, "Increment/Decrement World Scale", helpText: ButtonComponentUIHelp.offsetWorldScale);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.setWorldScale)
                {
                    CreateSpacer(groupID);
                    CreateSlider(groupID, currentButtonOperation.worldScaleComponent.worldScaleAbsJSF, false, "Set World Scale", helpText: ButtonComponentUIHelp.setWorldScale);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.showVRHands)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.showVRHandsComponent.toggleShowLeftVRHandJSB, false, "Toggle Show Left VR Hand", helpText: ButtonComponentUIHelp.showLeftVRHand);
                    CreateToggle(groupID, currentButtonOperation.showVRHandsComponent.toggleShowRightVRHandJSB, false, "Toggle Show Right VR Hand", helpText: ButtonComponentUIHelp.showRightVRHand);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.leapMotionEnable)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.showVRHandsComponent.leapHandsControlJSE, helpText: ButtonComponentUIHelp.leapHandsControl,scrollable:false);
                    CreateToggle(groupID, currentButtonOperation.showVRHandsComponent.enableLeapHandsJSB, false, "Enable Leap Hands Control", helpText: ButtonComponentUIHelp.enableLeapHands);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.leapMotionToggle)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.showVRHandsComponent.leapHandsControlJSE, helpText: ButtonComponentUIHelp.leapHandsControl, scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.switchUIAGrid)
                {
                    RefreshUIAGridChoices();
                    switchUIAGridJSSC.val = currentButtonOperation.switchUIAGridComponent.switchUIAGridTargetJSI.val.ToString();
                    CreateSpacer(groupID);
                    CreatePopup(groupID, switchUIAGridJSSC, false, 650f, RefreshUIAGridChoices, "Select UIA Target Grid", switchUIAGridJSSC.choices.Count > 10, helpText: ButtonComponentUIHelp.uiaGrid);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.shaderQaulity)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.shaderQualityJSEnum, false, 250f, null, "Shader Quality", helpText: "This button operation will configure the VAM User Performance setting 'Shader Quality' to the value selected in the drop down.", scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.msaaLevel)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.msaaLevelJSEnum, false, 300f, null, "MSAA Level", helpText: "This button operation will configure the VAM User Performance setting '' to the value selected in the drop down.", scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.pixelLightCount)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.pixelLightCountJSEnum, false, 450f, null, "Pixel Light Count", helpText: "This button operation will configure the VAM User Performance setting '' to the value selected in the drop down.", scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.smoothPasses)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.smoothPassesJSEnum, false, 350f, null, "Smooth Passes", helpText: "This button operation will configure the VAM User Performance setting 'Smooth Passes' to the value selected in the drop down.", scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.glowEffects)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.glowEffectsJSEnum, false, 250f, null, "Glow Effects", helpText: "This button operation will configure the VAM User Performance setting 'Glow Effects' to the value selected in the drop down.", scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.physicsRate)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.physicsRateJSEnum, false, 550f, null, "Physics Rate", helpText: "This button operation will configure the VAM User Performance setting 'Physics Rate' to the value selected in the drop down.", scrollable: false);
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.physicsUpdateCap)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID, currentButtonOperation.userPreferencesComponent.physicsUpdateCapJSEnum, false, 250f, null, "Physics Update Cap", helpText: "This button operation will configure the VAM User Performance setting 'Physics Update Cap' to the value selected in the drop down.", scrollable: false);
                }

                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.softBodyPhysics)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.userPreferencesComponent.softbodyPhysicsJSB, false, "Soft Body Physics Enabled/Disabled", 80f, helpText: "If this toggle is active, then the VAM User Performance setting 'Soft Body Physics' will be enabled. Otherwise the setting will be disabled by this UIAssist button operation.");
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.desktopVSync)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.userPreferencesComponent.desktopVSyncJSB, false, "Desktop VSync Enabled/Disabled", 80f, helpText: "If this toggle is active, then the VAM User Performance setting 'Desktop VSync' will be enabled. Otherwise the setting will be disabled by this UIAssist button operation.");
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.realtimeReflectionProbes)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.userPreferencesComponent.realtimeReflectionProbesJSB, false, "Realtime Reflection Probes Enabled/Disabled", 80f, helpText: "If this toggle is active, then the VAM User Performance setting 'Realtime Reflection Probes' will be enabled. Otherwise the setting will be disabled by this UIAssist button operation.");
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.mirrorReflections)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.userPreferencesComponent.mirrorReflectionsJSB, false, "Mirror Reflections Enabled/Disabled", 80f, helpText: "If this toggle is active, then the VAM User Performance setting 'Mirror Reflections' will be enabled. Otherwise the setting will be disabled by this UIAssist button operation.");
                }
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.hqPhysics)
                {
                    CreateSpacer(groupID);
                    CreateToggle(groupID, currentButtonOperation.userPreferencesComponent.hqPhysicsJSB, false, "High Quality Physics Enabled/Disabled", 80f, helpText: "If this toggle is active, then the VAM User Performance setting 'High Quality Physics' will be enabled. Otherwise the setting will be disabled by this UIAssist button operation.");
                }
                else if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.renderScale)
                {
                    CreateSpacer(groupID);
                    CreateSlider(groupID, currentButtonOperation.userPreferencesComponent.renderScaleJSF, false, "Render Scale", "This button operation will configure the VAM User Performance setting 'Render Scale' to the value selected in the slider.");
                }
                else if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.activeClothingEditor)
                {
                    CreateSpacer(groupID);
                    CreatePopup(groupID,targetButtonComponent.aceModeJSEnum,popupHeight:160f,scrollable:false, helpText:"Select whether to use the Enhanced or Orginal Active Clothing Editor. The Enhanced version has more options including being able to select clothing item presets but uses more screen space.");
                }


                string labelPostfix = "";
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodePositionState || currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodeRotationState) labelPostfix = " ON";
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.nodePositionState || currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodePositionState || currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.nodeRotationState || currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodeRotationState)
                {
                    CreateSpacer(groupID);
                    if (currentButtonOperation.targetComponent.isPersonTargetType)
                    {
                        CreatePopup(groupID, currentButtonOperation.nodeControlOnComponent.multiNodeControlStateJSEnum, scrollable: false, label: "Control Preset" + labelPostfix);
                        ForceUIRefreshOnValChange(currentButtonOperation.nodeControlOnComponent.multiNodeControlStateJSEnum, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);
                    }
                    else CreatePopup(groupID, currentButtonOperation.nodeControlOnComponent.GetRootNodeControlStates(), scrollable: false, label: "Control State" + labelPostfix);
                }
                labelPostfix = "";
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodePositionState || currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodeRotationState) labelPostfix = " OFF";
                if (currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodePositionState || currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.toggleNodeRotationState)
                {
                    if (currentButtonOperation.targetComponent.isPersonTargetType)
                    {
                        CreatePopup(groupID, currentButtonOperation.nodeControlOffComponent.multiNodeControlStateJSEnum, label: "Control Preset" + labelPostfix);
                        ForceUIRefreshOnValChange(currentButtonOperation.nodeControlOffComponent.multiNodeControlStateJSEnum, LeftRight.left, 2, UIAPluginUI.buttonSetupUI);
                    }
                    else CreatePopup(groupID, currentButtonOperation.nodeControlOffComponent.GetRootNodeControlStates(), label: "Control State" + labelPostfix);
                }
            }

            if (groupID == 6 && !isRight && CustomTreeNavUIElement.treeNavMode == CustomTreeNavUIElement.NORMAL)
            {

                {
                    CreateSpacer(groupID, (1000f - GetTotalUIElementHeights(false) - 120f));
                    if (PatreonFeatures.patreonContentEnabled)
                    {
                        addButtonOpUI.InitUI(1, false, false, false, false);
                        var button = addButtonOpUI.AddGridlessButtonToRow("Add Operation", 3, AddOperation, 300f, 0f);
                        button.buttonColor = Color.green;
                        RegisterContextualHelp(button.gameObject, helpText: ButtonComponentUIHelp.addOperation);

                    }
                    RegisterCustomUIElement(groupID, addButtonOpUI);
                }

            }
        }

        private void AddOperation(int buttonID)
        {
            int currentOpIndex = currentButtonOperation.buttonOpIndex;

            UIAButtonOperation newOp = new UIAButtonOperation(currentUIAButton);
            currentUIAButton.buttonOperations.Insert(currentOpIndex + 1, newOp);
            ButtonSetupPluginUI.buttonOperationSelectorJSSC.valNoCallback = newOp.buttonOpRef;
            ButtonSetupPluginUI.OperationSelected(newOp.buttonOpRef);

            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);
        }


        private void RefreshTargetNameChoices()
        {
            targetButtonComponent.RefreshTargetNameChoices(false);
        }

        private void SwitchUIAGridUpdated(string gridRef)
        {
            currentButtonOperation.switchUIAGridComponent.switchUIAGridTargetJSI.valNoCallback = int.Parse(gridRef);
        }

        private void RefreshUIAGridChoices()
        {
            List<string> choices = new List<string>();
            List<string> displayChoices = new List<string>();

            int index = 1;
            foreach (UIAButtonGrid bg in UIAStorables.buttonGridsList)
            {
                choices.Add(index.ToString());
                if (bg.gridLabelJSS.val != "") displayChoices.Add(bg.gridLabelJSS.val);
                else displayChoices.Add(index.ToString());
                index++;
            }

            switchUIAGridJSSC.choices = choices;
            switchUIAGridJSSC.displayChoices = displayChoices;

            if (!switchUIAGridJSSC.choices.Contains(switchUIAGridJSSC.val)) switchUIAGridJSSC.val = switchUIAGridJSSC.choices[0];
        }

        private void ForceButtonTypeEnumRefresh()
        {
            currentButtonOperation.buttonOpTypeJSEnum.SetEnumChoices(UIAButtonOpType.enumManifestName, UIAButtonOpType.GetButtonTypeExclusions(currentButtonOperation.buttonOpCategoryJSEnum.val));
        }

        private void ButtonCategoryChanged(string category)
        {
            if (currentButtonOperation.buttonOpCategoryJSEnum.val == UIAButtonCategory.misc && currentButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.blank) currentButtonOperation.buttonOpTypeJSEnum.val = UIAButtonOpType.freezeMotionSound;
        }
    }

    public class GridComponentUITab : BaseUIButtonTreeTab
    {
        public JSONStorableStringChooser gridLayoutJSSC;

        private CustomRowUIElement addGridUI;

        UIAButtonGrid currentUIAGrid { get { return ButtonSetupPluginUI.currentUIAButtonGrid; } }
        public GridComponentUITab() : base()
        {
            uiGroupCount = 2;
            defaultHelpText = GridComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            gridLayoutJSSC = new JSONStorableStringChooser("gridLayout", GetButtonGridLayoutChoices(), "", "", SwitchGridLayout);
            addGridUI = new CustomRowUIElement(80, false);
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                UpdateGridLayoutChoices();
                CreatePopup(groupID, gridLayoutJSSC, false, 800, null, "Grid Layout", gridLayoutJSSC.choices.Count > 14, helpText: GridComponentUIHelp.gridLayout);
                gridLayoutJSSC.valNoCallback = currentUIAGrid.GetGridLayout();
                ForceUIRefreshOnValChange(gridLayoutJSSC, LeftRight.right, 1);
                CreatePopup(groupID, currentUIAGrid.buttonSizeJSEnum, false, 300, ButtonSizePopupOpen, "Button Size", false, helpText: GridComponentUIHelp.buttonSize);
                if (!currentUIAGrid.quickLaunchBarJSBool.val) CreateTextField(groupID, currentUIAGrid.gridLabelJSS, false, 100f, true, InputField.LineType.SingleLine, "Grid Label", helpText: GridComponentUIHelp.gridLabel);
            }
            if (groupID == 2 && !isRight && CustomTreeNavUIElement.treeNavMode == CustomTreeNavUIElement.NORMAL)
            {
                CreateSpacer(groupID, (1000f - GetTotalUIElementHeights(false) - 120f));
                addGridUI.InitUI(1, false, false, false, false);

                var button = addGridUI.AddGridlessButtonToRow("Add New Grid", 1, AddGrid, 300f, 0f);
                button.buttonColor = Color.green;
                RegisterContextualHelp(button.gameObject, helpText: GridComponentUIHelp.addNewGrid);
                RegisterCustomUIElement(groupID, addGridUI);
            }

        }

        private void AddGrid(int buttonRef)
        {
            int currentGridIndex = currentUIAGrid.gridIndex;

            UIAButtonGrid newGrid = new UIAButtonGrid(3, 3);
            UIAStorables.buttonGridsList.Insert(currentGridIndex + 1, newGrid);
            ButtonSetupPluginUI.gridSelectorJSSC.val = newGrid.gridRef;

            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);

            if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit) GameControlUI.RefreshWristUIButtonGrid();
        }

        private void ButtonSizePopupOpen()
        {
            List<int> exclsions = new List<int>();
            int gridColumns = currentUIAGrid.gridColumns;
            if (gridColumns > 5)
            {
                exclsions.Add(ButtonSize.large);
                exclsions.Add(ButtonSize.medium);
            }
            if (gridColumns > 6) exclsions.Add(ButtonSize.small);
            currentUIAGrid.buttonSizeJSEnum.SetEnumChoices(ButtonSize.enumManifestName, exclsions);
        }
        private void UpdateGridLayoutChoices()
        {
            gridLayoutJSSC.choices = GetButtonGridLayoutChoices();
        }
        private List<string> GetButtonGridLayoutChoices()
        {
            List<string> buttonGridLayoutChoices = new List<string>();
            if (currentUIAGrid != UIAStorables.quickLaunchButtonGrid)
            {
                buttonGridLayoutChoices.Add("2x2"); buttonGridLayoutChoices.Add("2x3"); buttonGridLayoutChoices.Add("3x2"); buttonGridLayoutChoices.Add("3x3");
                buttonGridLayoutChoices.Add("3x4");
                buttonGridLayoutChoices.Add("4x3");
                buttonGridLayoutChoices.Add("4x4");
                if (PatreonFeatures.patreonContentEnabled) PatreonFeatures.GridLayoutFeature(buttonGridLayoutChoices);
            }
            else
            {
                buttonGridLayoutChoices.Add("4x1");
                buttonGridLayoutChoices.Add("5x1");
                buttonGridLayoutChoices.Add("6x1");
                buttonGridLayoutChoices.Add("7x1");
                buttonGridLayoutChoices.Add("8x1");
                buttonGridLayoutChoices.Add("9x1");
                buttonGridLayoutChoices.Add("4x2");
                buttonGridLayoutChoices.Add("5x2");
                buttonGridLayoutChoices.Add("6x2");
                buttonGridLayoutChoices.Add("7x2");
                buttonGridLayoutChoices.Add("8x2");
                buttonGridLayoutChoices.Add("9x2");
                buttonGridLayoutChoices.Add("4x3");
                buttonGridLayoutChoices.Add("5x3");
                buttonGridLayoutChoices.Add("6x3");
                buttonGridLayoutChoices.Add("7x3");
                buttonGridLayoutChoices.Add("8x3");
                buttonGridLayoutChoices.Add("9x3");

            }

            return (buttonGridLayoutChoices);
        }

        public void SwitchGridLayout(string newLayout)
        {
            currentUIAGrid.gridColumns = int.Parse(newLayout.Substring(0, 1));
            currentUIAGrid.gridRows = int.Parse(newLayout.Substring(2, 1));

            int buttonCols = currentUIAGrid.gridColumns;
            if (buttonCols > 6 && currentUIAGrid.buttonSizeJSEnum.val != ButtonSize.micro && currentUIAGrid.buttonSizeJSEnum.val != ButtonSize.mini) currentUIAGrid.buttonSizeJSEnum.val = ButtonSize.mini;
            else if (buttonCols > 5 && (currentUIAGrid.buttonSizeJSEnum.val == ButtonSize.medium || currentUIAGrid.buttonSizeJSEnum.val == ButtonSize.large)) currentUIAGrid.buttonSizeJSEnum.val = ButtonSize.small;

            if (currentUIAGrid.buttonCount < ButtonSetupPluginUI.currentUIAButton.buttonGridIndex + 1) ButtonSetupPluginUI.buttonSelectorJSSC.val = "R" + currentUIAGrid.gridRows.ToString() + "C" + currentUIAGrid.gridColumns.ToString();

            GameControlUI.RefreshWristUIButtonGrid();
        }
    }


    public class SpawnAtomComponentUITab : BaseUIButtonTreeTab
    {
        SpawnAtomComponent spawnAtomComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.spawnAtomComponent; } }

        public SpawnAtomComponentUITab() : base()
        {
            uiGroupCount = 3;
            defaultHelpText = SpawnAtomComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (isRight) CreateHelpTextUI(groupID, 975);
            else if (groupID == 1 && !isRight)
            {

                CreatePopup(groupID, spawnAtomComponent.atomCategoryJSEnum, false, 400f, null, "New Atom Category", false, helpText: SpawnAtomComponentUIHelp.newAtomCategory);
                ForceUIRefreshOnValChange(spawnAtomComponent.atomCategoryJSEnum, LeftRight.left, 2);

            }
            else if (groupID == 2 && !isRight)
            {
                CreatePopup(groupID, spawnAtomComponent.atomTypeJSEnum, false, 600f, null, "New Atom Type", helpText: SpawnAtomComponentUIHelp.newAtomType);
                CreateToggle(groupID, spawnAtomComponent.atomSpawnSpecifyNameJSBool, false, "Specify new Atom name", helpText: SpawnAtomComponentUIHelp.specifyName);
                ForceUIRefreshOnValChange(spawnAtomComponent.atomSpawnSpecifyNameJSBool, LeftRight.left, 3);
            }
            else if (groupID == 3 && !isRight)
            {
                if (spawnAtomComponent.atomSpawnSpecifyNameJSBool.val) CreateTextField(groupID, spawnAtomComponent.atomSpawnNameJSString, false, 100f, true, InputField.LineType.SingleLine, "New atom name", helpText: SpawnAtomComponentUIHelp.newAtomName);
            }
        }

    }

    public class RelativePositionComponentUITab : BaseUIButtonTreeTab
    {
        static UIAButtonOperation buttonOp { get { return ButtonSetupPluginUI.currentUIAButtonOperation; } }
        static SpawnAtomComponent spawnAtomComponent { get { return buttonOp.spawnAtomComponent; } }
        static RelativePositionComponent relativePositionComponent { get { return buttonOp.relativePositionComponent; } }

        JSONStorableStringChooser testAtomPositionSelectorJSSC;
        JSONStorableBool testAtomPositionDisablePhysicsJSB;

        public static bool testSpawnModeActive = false;
        public static string testSpawnAtom = "";
        private static Vector3 testSpawnAtomOriginalPosition = new Vector3();
        private static Quaternion testSpawnAtomOriginalRotation = new Quaternion();
        private static bool testSpawnAtomOriginalCollision;
        public static int testSpawnMode = TestSpawnMode.spawnAtomMode;

        private static Vector3 testSpawnAtomLastPosition = new Vector3();
        private static Quaternion testSpawnAtomLastRotation = new Quaternion();
        private static string testSpawnAtomRelativeAtom = "";
        private static FreeControllerV3.PositionState testSpawnAtomOriginalPositionState;
        private static FreeControllerV3.RotationState testSpawnAtomOriginalRotationState;


        public RelativePositionComponentUITab() : base()
        {
            uiGroupCount = 7;
            defaultHelpText = RelativePositionComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            testAtomPositionSelectorJSSC = new JSONStorableStringChooser("testAtomPositionSelector", null, "Select...", "", SwitchAtomSpawnTestTarget);
            testAtomPositionDisablePhysicsJSB = new JSONStorableBool("testAtomPositionDisablePhysics", false);
        }

        public static void AtomNameUpdate(string oldName, string newName)
        {
            if (testSpawnAtom == oldName) testSpawnAtom = newName;
        }

        public static void AtomRemovedUpdate(Atom atom)
        {
            string atomName = atom.name;
            if (testSpawnModeActive && testSpawnAtom == atomName) DeactivateSpawnTest();
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                CreatePopup(groupID, relativePositionComponent.atomPositionRelativeToJSEnum, false, 850f, null, "Relative to", false, helpText: RelativePositionComponentUIHelp.relativeTo);
                ForceUIRefreshOnValChange(relativePositionComponent.atomPositionRelativeToJSEnum, LeftRight.left, 2);

            }
            else if (groupID == 2 && !isRight)
            {
                if (relativePositionComponent.atomPositionRelativeToJSEnum.val == RelativePositionMode.lastViewedFemale || relativePositionComponent.atomPositionRelativeToJSEnum.val == RelativePositionMode.lastViewedMale || relativePositionComponent.atomPositionRelativeToJSEnum.val == RelativePositionMode.lastViewedPerson)
                {
                    CreatePopup(groupID, relativePositionComponent.targetNodeJSSC, false, 600f, null, "Target Node", true, helpText: RelativePositionComponentUIHelp.targetNode);
                    if (ButtonSetupPluginUI.currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.spawnAtom) CreateToggle(groupID, spawnAtomComponent.parentLinkToTargetJSBool, false, "Parent link Atom to target node", helpText: RelativePositionComponentUIHelp.parentLink);


                }
                CreateSpacer(groupID);
                CreateSlider(groupID, relativePositionComponent.atomPositionXJSFloat, false, "Position X (relative)", helpText: RelativePositionComponentUIHelp.xPosition);
                ForceUIRefreshOnValChange(relativePositionComponent.atomPositionRelativeToJSEnum, LeftRight.left, 3);

            }
            else if (groupID == 3 && !isRight)
            {
                if (relativePositionComponent.atomPositionRelativeToJSEnum.val != RelativePositionMode.sceneOrigin)
                {
                    CreateToggle(groupID, relativePositionComponent.atomAbsoluteHeightJSBool, false, "Fixed Height (Y Position)", helpText: RelativePositionComponentUIHelp.fixedYPos);
                    ForceUIRefreshOnValChange(relativePositionComponent.atomAbsoluteHeightJSBool, LeftRight.left, 4);
                }

            }
            else if (groupID == 4 && !isRight)
            {
                if (relativePositionComponent.atomPositionRelativeToJSEnum.val != RelativePositionMode.sceneOrigin && relativePositionComponent.atomAbsoluteHeightJSBool.val)
                {
                    CreateSlider(groupID, relativePositionComponent.atomPositionYJSFloat, false, "Position Y (absolute)", helpText: RelativePositionComponentUIHelp.yPosition);
                }
                else CreateSlider(groupID, relativePositionComponent.atomPositionYJSFloat, false, "Position Y (relative)", helpText: RelativePositionComponentUIHelp.yPosition);
                CreateSlider(groupID, relativePositionComponent.atomPositionZJSFloat, false, "Position Z (relative)", helpText: RelativePositionComponentUIHelp.zPosition);
                if (relativePositionComponent.atomPositionRelativeToJSEnum.val != RelativePositionMode.sceneOrigin && buttonOp.buttonOpTypeJSEnum.val != UIAButtonOpType.teleportPlayer)
                {
                    CreateToggle(groupID, relativePositionComponent.atomAbsolutePitchJSBool, false, "Fixed Pitch (X Rotation)", helpText: RelativePositionComponentUIHelp.fixedXRot);
                    ForceUIRefreshOnValChange(relativePositionComponent.atomAbsolutePitchJSBool, LeftRight.left, 5);
                }
            }
            else if (groupID == 5 && !isRight)
            {
                if (buttonOp.buttonOpTypeJSEnum.val != UIAButtonOpType.teleportPlayer)
                {
                    if (relativePositionComponent.atomPositionRelativeToJSEnum.val != RelativePositionMode.sceneOrigin && relativePositionComponent.atomAbsolutePitchJSBool.val)
                    {
                        CreateSlider(groupID, relativePositionComponent.atomRotationXJSFloat, false, "Rotation X (absolute)", helpText: RelativePositionComponentUIHelp.xRotation);
                    }
                    else CreateSlider(groupID, relativePositionComponent.atomRotationXJSFloat, false, "Rotation X (relative)", helpText: RelativePositionComponentUIHelp.xRotation);
                }

                if (buttonOp.buttonOpTypeJSEnum.val != UIAButtonOpType.teleportPlayer || relativePositionComponent.atomPositionRelativeToJSEnum.val == RelativePositionMode.sceneOrigin) CreateSlider(groupID, relativePositionComponent.atomRotationYJSFloat, false, "Rotation Y (relative)", helpText: RelativePositionComponentUIHelp.yRotation);

                if (buttonOp.buttonOpTypeJSEnum.val != UIAButtonOpType.teleportPlayer)
                {
                    if (relativePositionComponent.atomPositionRelativeToJSEnum.val != RelativePositionMode.sceneOrigin)
                    {
                        CreateToggle(groupID, relativePositionComponent.atomAbsoluteRollJSBool, false, "Fixed Roll (Z Rotation)", helpText: RelativePositionComponentUIHelp.fixedZRot);
                        ForceUIRefreshOnValChange(relativePositionComponent.atomAbsoluteRollJSBool, LeftRight.left, 6);
                    }
                }


            }
            else if (groupID == 6 && !isRight)
            {
                if (buttonOp.buttonOpTypeJSEnum.val != UIAButtonOpType.teleportPlayer)
                {
                    if (relativePositionComponent.atomPositionRelativeToJSEnum.val != RelativePositionMode.sceneOrigin && relativePositionComponent.atomAbsoluteRollJSBool.val)
                    {
                        CreateSlider(groupID, relativePositionComponent.atomRotationZJSFloat, false, "Rotation Z (absolute)", helpText: RelativePositionComponentUIHelp.zRotation);
                    }
                    else CreateSlider(groupID, relativePositionComponent.atomRotationZJSFloat, false, "Rotation Z (relative)", helpText: RelativePositionComponentUIHelp.zRotation);
                }


                string label = "Test Atom Spawn Position on";
                if (buttonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.teleportAtom) label = "Test Atom Teleport Position on";
                CreatePopup(groupID, testAtomPositionSelectorJSSC, false, 400f, UpdateTestSpawnPosTargetChoices, label, true, helpText: RelativePositionComponentUIHelp.testAtomOn);
                ForceUIRefreshOnValChange(testAtomPositionSelectorJSSC, LeftRight.left, 7);
            }
            else if (groupID == 7 && !isRight)
            {
                if (testAtomPositionSelectorJSSC.val != "Select...")
                {
                    CreateToggle(groupID, testAtomPositionDisablePhysicsJSB, false, "Disable Atom Collision for test", helpText: RelativePositionComponentUIHelp.disableCollision);
                    if (!testSpawnModeActive) CreateButton(groupID, "Start Test", false, ToggleSpawnTest, helpText: RelativePositionComponentUIHelp.startTest);
                    else CreateButton(groupID, "End Test", false, ToggleSpawnTest, helpText: RelativePositionComponentUIHelp.endTest);
                }
                CreateSpacer(groupID, 400f);
            }
        }

        private void UpdateTestSpawnPosTargetChoices()
        {
            List<string> popupChoices = new List<string>();

            bool isSpawnAtom = buttonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.spawnAtom;
            bool isTeleportAtom = buttonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.teleportAtom;
            bool isGenPresetSpawn = buttonOp.targetComponent.SpawnAtomIfTargetMissing();
            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                if (isSpawnAtom)
                {
                    if (atom.type == spawnAtomComponent.atomTypeJSEnum.displayVal) popupChoices.Add(atom.name);
                }
                else if (isGenPresetSpawn)
                {
                    if (atom.type == buttonOp.targetComponent.targetNameJSMultiEnum.mainVal) popupChoices.Add(atom.name);
                }
                else if (isTeleportAtom) popupChoices.Add(atom.name);
            }
            if (popupChoices.Count == 0) popupChoices.Add("No Atoms of Spawn Type available");

            testAtomPositionSelectorJSSC.choices = popupChoices;

        }

        private void SwitchAtomSpawnTestTarget(string targetAtom)
        {
            if (targetAtom == "No Atoms of Spawn Type available") testAtomPositionSelectorJSSC.val = "Select...";
            if (testSpawnModeActive) DeactivateSpawnTest();

        }

        public void ToggleSpawnTest()
        {
            if (testSpawnModeActive)
            {
                DeactivateSpawnTest();
                RefreshGroupUI(LeftRight.left, 7);
                return;
            }
            Atom atom = SuperController.singleton.GetAtomByUid(testAtomPositionSelectorJSSC.val);
            if (atom != null)
            {
                testSpawnModeActive = true;
                testSpawnAtom = atom.name;
                testSpawnAtomOriginalPosition = atom.mainController.transform.position;
                testSpawnAtomOriginalRotation = atom.mainController.transform.rotation;
                testSpawnAtomOriginalPositionState = atom.mainController.currentPositionState;
                testSpawnAtomOriginalRotationState = atom.mainController.currentRotationState;

                atom.mainController.currentPositionState = FreeControllerV3.PositionState.On;
                atom.mainController.currentRotationState = FreeControllerV3.RotationState.On;
                testSpawnAtomOriginalCollision = atom.collisionEnabledJSON.val;
                ToggleSpawnTestCollisionEnabled(!testAtomPositionDisablePhysicsJSB.val);

                testSpawnAtomRelativeAtom = relativePositionComponent.GetRelativePositionAtomName();
                relativePositionComponent.TeleportAtomToSpawnPoint(testSpawnAtom);
                testSpawnAtomLastPosition = atom.mainController.transform.position;
                testSpawnAtomLastRotation = atom.mainController.transform.rotation;

                bool isSpawnAtom = buttonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.spawnAtom;
                bool isTeleportAtom = buttonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.teleportAtom;

                if (isSpawnAtom) testSpawnMode = TestSpawnMode.spawnAtomMode;
                else if (isTeleportAtom) testSpawnMode = TestSpawnMode.teleportAtomMode;

                RefreshGroupUI(LeftRight.left, 7);
            }
        }
        public static void ToggleSpawnTestCollisionEnabled(bool collisionEnabled)
        {
            if (testSpawnModeActive)
            {
                Atom atom = SuperController.singleton.GetAtomByUid(testSpawnAtom);
                if (atom != null) atom.collisionEnabledJSON.val = collisionEnabled;
            }
        }

        public static void DeactivateSpawnTest()
        {
            if (testSpawnModeActive)
            {
                Vector3 position = new Vector3(0f, 0f, 0f);
                Quaternion rotation = new Quaternion();

                position = testSpawnAtomOriginalPosition;
                rotation = testSpawnAtomOriginalRotation;
                relativePositionComponent.TeleportAtomToSpawnPoint(testSpawnAtom, position, rotation, false);
                ToggleSpawnTestCollisionEnabled(testSpawnAtomOriginalCollision);

                testSpawnModeActive = false;
                Atom atom = SuperController.singleton.GetAtomByUid(testSpawnAtom);
                SuperController.singleton.helpHUDText.text = "";
                if (atom != null)
                {
                    atom.mainController.currentPositionState = testSpawnAtomOriginalPositionState;
                    atom.mainController.currentRotationState = testSpawnAtomOriginalRotationState;
                }

            }

        }
        public static void UpdateSpawnTestAtomPosition()
        {
            if (testSpawnModeActive)
            {
                Atom atom = SuperController.singleton.GetAtomByUid(testSpawnAtom);
                {
                    Vector3 currentPosition = atom.mainController.transform.position;
                    Quaternion currentRotation = atom.mainController.transform.rotation;
                    float distanceChange = Vector3.Distance(currentPosition, testSpawnAtomLastPosition);

                    if (distanceChange > 0.000001f || currentRotation != testSpawnAtomLastRotation)
                    {
                        Transform relativeToTransform = relativePositionComponent.GetSpawnRelativeToTransform(testSpawnAtomRelativeAtom);
                        Vector3 relativePosition = relativeToTransform.InverseTransformDirection(currentPosition - relativeToTransform.position);
                        Quaternion relativeRotation = new Quaternion();

                        relativeRotation = currentRotation * Quaternion.Inverse(relativeToTransform.rotation);

                        //                        relativeRotation.eulerAngles = new Vector3(currentRotation.eulerAngles.x - relativeToTransform.rotation.eulerAngles.x, currentRotation.eulerAngles.y - relativeToTransform.rotation.eulerAngles.y, currentRotation.eulerAngles.z - relativeToTransform.rotation.eulerAngles.z);

                        relativePositionComponent.atomPositionXJSFloat.val = relativePosition.x;
                        if (relativePositionComponent.atomAbsoluteHeightJSBool.val) relativePositionComponent.atomPositionYJSFloat.val = currentPosition.y;
                        else relativePositionComponent.atomPositionYJSFloat.val = relativePosition.y;
                        relativePositionComponent.atomPositionZJSFloat.val = relativePosition.z;

                        if (relativePositionComponent.atomAbsolutePitchJSBool.val) relativePositionComponent.atomRotationXJSFloat.val = AngleDisplay(currentRotation.eulerAngles.x);
                        else relativePositionComponent.atomRotationXJSFloat.val = AngleDisplay(relativeRotation.eulerAngles.x);

                        relativePositionComponent.atomRotationYJSFloat.val = AngleDisplay(relativeRotation.eulerAngles.y);

                        if (relativePositionComponent.atomAbsoluteRollJSBool.val) relativePositionComponent.atomRotationZJSFloat.val = AngleDisplay(currentRotation.eulerAngles.z);
                        else relativePositionComponent.atomRotationZJSFloat.val = AngleDisplay(relativeRotation.eulerAngles.z);


                        testSpawnAtomLastPosition = atom.mainController.transform.position;
                        testSpawnAtomLastRotation = atom.mainController.transform.rotation;
                    }
                    else
                    {
                        Vector3 targetPosition = new Vector3();
                        Quaternion targetRotation = new Quaternion();

                        relativePositionComponent.UpdateSpawnPosition(ref targetPosition, ref targetRotation, testSpawnAtomRelativeAtom);

                        if (targetPosition != testSpawnAtomLastPosition || targetRotation != testSpawnAtomLastRotation)
                        {
                            atom.mainController.transform.position = targetPosition;
                            atom.mainController.transform.rotation = targetRotation;
                            testSpawnAtomLastPosition = targetPosition;
                            testSpawnAtomLastRotation = targetRotation;
                        }
                    }
                }
            }
        }

        public static float AngleDisplay(float eulerAngle)
        {
            if (eulerAngle > 180f) return -360f + eulerAngle;
            return eulerAngle;
        }


    }

    public class MoveAtomComponentUITab : BaseUIButtonTreeTab
    {
        MoveAtomComponent moveAtomComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.moveAtomComponent; } }

        public MoveAtomComponentUITab() : base()
        {
            uiGroupCount = 1;
            defaultHelpText = MoveAtomComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                CreateToggle(groupID, moveAtomComponent.lockedPositionXJSB, false, "Lock X Axis Position", helpText: MoveAtomComponentUIHelp.lockXPos);
                CreateToggle(groupID, moveAtomComponent.lockedPositionYJSB, false, "Lock Y Axis Position (Height)", helpText: MoveAtomComponentUIHelp.lockYPos);
                CreateToggle(groupID, moveAtomComponent.lockedPositionZJSB, false, "Lock Z Axis Position", helpText: MoveAtomComponentUIHelp.lockZPos);

                CreateToggle(groupID, moveAtomComponent.lockedRotationXJSB, false, "Lock X Axis Rotation (Pitch)", helpText: MoveAtomComponentUIHelp.lockXRot);
                CreateToggle(groupID, moveAtomComponent.lockedRotationYJSB, false, "Lock Y Axis Rotation (Yaw)", helpText: MoveAtomComponentUIHelp.lockYRot);
                CreateToggle(groupID, moveAtomComponent.lockedRotationZJSB, false, "Lock Z Axis Rotation (Roll)", helpText: MoveAtomComponentUIHelp.lockZRot);

            }
        }

    }
    public class FileReferencesComponentUITab : BaseUIButtonTreeTab
    {
        Dictionary<int, FileReference> fileReferences { get { return currentUIAButtonOperation.fileReferenceDict; } }

        UIAButtonOperation currentUIAButtonOperation { get { return ButtonSetupPluginUI.currentUIAButtonOperation; } }

        SkinPresetDecalComponent skinDecalComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.skinPresetDecalComponent; } }

        AppearancePresetComponent appearancePresetComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.appearancePresetComponent; } }

        private JSONStorableEnumStringChooser baFilterParameterSelector;
        private JSONStorableEnumStringChooser baFilterParamJSEnum;
        private JSONStorableStringChooser baFilterUserTagCreatorJSSC;
        private UIDynamicPopup userTagCreatorNamePopup;

        public FileReferencesComponentUITab() : base()
        {
            uiGroupCount = 2;
            defaultHelpText = FileReferencesComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
            baFilterParameterSelector = new JSONStorableEnumStringChooser("baFilterParameterSelector", BAFilterParams.enumManifestName, BAFilterParams.tagGroupType,"Select BA Filter");
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            List<int> fileReferenceTypes = currentUIAButtonOperation.GetFileReferenceTypes();

            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                foreach (int fileRefType in fileReferenceTypes)
                {
                    FileReference fileRef = fileReferences[fileRefType];
                    string fileRefDescription = fileRef.GetFileRefTypeDescription();
                    bool isAddonFolderRef = fileRef.fileReferenceType == FileReferenceTypes.addonPackagesFolder;

                    FileReference mainFileRef = null;
                    if (isAddonFolderRef) mainFileRef = currentUIAButtonOperation.fileReferenceDict[currentUIAButtonOperation.GetFileReferenceTypes().First()];

                    if (PatreonFeatures.patreonContentEnabled && !isAddonFolderRef && currentUIAButtonOperation.buttonOpTypeJSEnum.val != UIAButtonOpType.loadMotionCapture)
                    {
                        CreatePopup(groupID, fileRef.fileSelectionModeJSEnum, false, 400, null, fileRefDescription + "\nSelect Mode", false, FileReferencesComponentUIHelp.selectMode, 250f);
                        ForceUIRefreshOnValChange(fileRef.fileSelectionModeJSEnum, LeftRight.left, 1);

                    }
                    if ((fileRef.fileSelectionModeJSEnum.val != FileSelectionMode.none && fileRef.fileSelectionModeJSEnum.val != FileSelectionMode.randomFromBA && fileRef.fileSelectionModeJSEnum.val != FileSelectionMode.sequentialFromBA && fileRef.fileSelectionModeJSEnum.val != FileSelectionMode.chooseFromBA && !isAddonFolderRef) || (isAddonFolderRef && mainFileRef.fileSelectionModeJSEnum.val != FileSelectionMode.none && mainFileRef.fileSelectionModeJSEnum.val != FileSelectionMode.singleFile ))
                    {

                        string selectionType = "folder";
                        if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.singleFile) selectionType = "file";
                        if (isAddonFolderRef) selectionType = "filter";
                        CreateButton(groupID, "Select " + fileRefDescription + " " + selectionType, false, fileRef.GetFileFolderDialog, helpText: FileReferencesComponentUIHelp.fileFolderSelection);
                        CreateTextField(groupID, fileRef.filePathJSString, helpText: FileReferencesComponentUIHelp.filePath);
                        if ((!PatreonFeatures.patreonContentEnabled || isAddonFolderRef) && fileRef.filePathJSString.val != "") CreateButton(groupID, "Clear " + selectionType + " reference", helpText: FileReferencesComponentUIHelp.clearFileReference);
                        if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.randomFromFolder || fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.alphabeticalFromFolder) CreateToggle(groupID, fileRef.includeSubFoldersJSB, false, "Include Sub-Folders", helpText: FileReferencesComponentUIHelp.includeSubFolders);
                    }
                    if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.randomFromBA || fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.sequentialFromBA)
                    {
                        if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.randomFromBA) baFilterParameterSelector.SetEnumChoices(BAFilterParams.enumManifestName,new List<int> { BAFilterParams.sortMode});
                        else baFilterParameterSelector.SetEnumChoices(BAFilterParams.enumManifestName);

                        CreatePopup(groupID, baFilterParameterSelector,scrollable:false, helpText: FileReferencesComponentUIHelp.baFilterParameterSelector);
                        ForceUIRefreshOnValChange(baFilterParameterSelector, LeftRight.left, 1);

                        if (baFilterParameterSelector.val == BAFilterParams.tagGroupType)
                        {
                            baFilterParamJSEnum = new JSONStorableEnumStringChooser("baFilterParam", BATagGroupTypeFilter.enumManifestName, fileRef.baFilter.tagGroupType, "BA Tags Filter Setting", delegate (int newValue)
                            {
                                fileRef.baFilter.tagGroupType = newValue;
                                RefreshGroupUI(LeftRight.left, 1);
                            });
                            CreatePopup(groupID, baFilterParamJSEnum, scrollable: false, helpText: FileReferencesComponentUIHelp.baTagGroupTypeFilter);
                            if (baFilterParamJSEnum.val == BATagGroupTypeFilter.userTag)
                            {
                                baFilterUserTagCreatorJSSC = new JSONStorableStringChooser("", null, fileRef.baFilter.userTagName, "BA User Tag Name",delegate (string newValue) {
                                    fileRef.baFilter.userTagName = newValue;
                                });
                                 userTagCreatorNamePopup = CreatePopup(groupID, baFilterUserTagCreatorJSSC, helpText: FileReferencesComponentUIHelp.baUserTagNameFilter, callbackMethod: delegate
                                {
                                    RefreshBAFilterUserTagJSSC(fileRef);
                                    
                                });
                                if (baFilterUserTagCreatorJSSC.val == "") userTagCreatorNamePopup.popup.currentValue = "<i>Select...</i>";
                            }
                        }
                        else if (baFilterParameterSelector.val == BAFilterParams.hidden)
                        {
                            baFilterParamJSEnum = new JSONStorableEnumStringChooser("baFilterParam", BAShowHiddenFilter.enumManifestName, fileRef.baFilter.showHidden, "BA Show Hidden Filter Setting", delegate (int newValue)
                            {
                                fileRef.baFilter.showHidden = newValue;
                                RefreshGroupUI(LeftRight.left, 1);
                            });
                            CreatePopup(groupID, baFilterParamJSEnum,labelWidth:225f, scrollable: false, helpText: FileReferencesComponentUIHelp.baShowHiddenFilter);
                        }
                        else if (baFilterParameterSelector.val == BAFilterParams.favorite)
                        {
                            baFilterParamJSEnum = new JSONStorableEnumStringChooser("baFilterParam", BAOnlyFavoriteFilter.enumManifestName, fileRef.baFilter.onlyFav, "BA Only Favorite Filter Setting", delegate (int newValue)
                            {
                                fileRef.baFilter.onlyFav = newValue;
                                RefreshGroupUI(LeftRight.left, 1);
                            });
                            CreatePopup(groupID, baFilterParamJSEnum, labelWidth: 225f, scrollable: false, helpText: FileReferencesComponentUIHelp.baOnlyFavoriteFilter);
                        }
                        else if (baFilterParameterSelector.val == BAFilterParams.newUpdated)
                        {
                            baFilterParamJSEnum = new JSONStorableEnumStringChooser("baFilterParam", BAOnlyNewFilter.enumManifestName, fileRef.baFilter.onlyNew, "BA Only New Filter Setting", delegate (int newValue)
                            {
                                fileRef.baFilter.onlyNew = newValue;
                                RefreshGroupUI(LeftRight.left, 1);
                            });
                            CreatePopup(groupID, baFilterParamJSEnum, labelWidth: 225f, scrollable: false, helpText: FileReferencesComponentUIHelp.baOnlyNewFilter);
                        }
                        else if (baFilterParameterSelector.val == BAFilterParams.creator)
                        {
                            baFilterParamJSEnum = new JSONStorableEnumStringChooser("baFilterParam", BACreatorTypeFilter.enumManifestName, fileRef.baFilter.creatorType, "BA Creator Filter Setting", delegate (int newValue)
                            {
                                fileRef.baFilter.creatorType = newValue;
                                RefreshGroupUI(LeftRight.left, 1);
                            });
                            CreatePopup(groupID, baFilterParamJSEnum, labelWidth: 225f, scrollable: false, helpText: FileReferencesComponentUIHelp.baCreatorLocalFilter);
                            if (baFilterParamJSEnum.val == BACreatorTypeFilter.specificCreator)
                            {
                                baFilterUserTagCreatorJSSC = new JSONStorableStringChooser("", null, fileRef.baFilter.creatorName, "BA Creator Name", delegate (string newValue) {
                                    fileRef.baFilter.creatorName = newValue;
                                });
                                userTagCreatorNamePopup = CreateFilterablePopup(groupID, baFilterUserTagCreatorJSSC, helpText: FileReferencesComponentUIHelp.baCreatorNameFilter, callbackMethod: delegate
                                {
                                    RefreshBAFilterCreatorJSSC(fileRef);

                                });
                                if (baFilterUserTagCreatorJSSC.val == "") userTagCreatorNamePopup.popup.currentValue = "<i>Select...</i>";


                            }
                            else if (baFilterParamJSEnum.val == BACreatorTypeFilter.localFolders)
                            {
                                var defaultLocalFolder = fileRef.GetDefaultFolder();
                                if (fileRef.baFilter.localFolder == "") fileRef.baFilter.localFolder = defaultLocalFolder;
                                var localFolder = fileRef.baFilter.localFolder;
                                var localFolderTextField = CreateTextField(groupID, new JSONStorableString("",localFolder), helpText: FileReferencesComponentUIHelp.baLocalFolderFilter);
                                CreateToggle(groupID, new JSONStorableBool("", fileRef.baFilter.includeSubFolders, delegate (bool newValue)
                                {
                                    fileRef.baFilter.includeSubFolders = newValue;
                                    RefreshGroupUI(LeftRight.left, 1);
                                }), false,"Include subfolders in BA filter", helpText: FileReferencesComponentUIHelp.baIncludeSubFolderFilter);
                                if (localFolder != defaultLocalFolder)
                                {
                                    CreateButton(groupID, "Reset BA Local Folder",false, delegate
                                    {
                                        fileRef.baFilter.localFolder = defaultLocalFolder;
                                        RefreshGroupUI(LeftRight.left, 1);
                                    }, helpText: FileReferencesComponentUIHelp.baResetLocalFolderFilter);
                                }
                                CreateButton(groupID, "Select BA Local Folder", false, delegate
                                {
                                    SuperController.singleton.GetDirectoryPathDialog(delegate (string path)
                                    {
                                        if (path.StartsWith(defaultLocalFolder))
                                        {
                                            fileRef.baFilter.localFolder = path;
                                            RefreshGroupUI(LeftRight.left, 1);
                                        }
                                        else if (path!="") SuperController.LogMessage("Selected Folder must be within " + defaultLocalFolder);
                                    }, localFolder, null, false);
                                }, helpText: FileReferencesComponentUIHelp.baSelectLocalFolderFilter);
                            }
                        }
                        else if (baFilterParameterSelector.val == BAFilterParams.sortMode)
                        {
                            baFilterParamJSEnum = new JSONStorableEnumStringChooser("baFilterParam", BAResultsSortModeFilter.enumManifestName, fileRef.baFilter.resourceOrder, "BA Results Sort Order", delegate (int newValue)
                            {
                                fileRef.baFilter.resourceOrder = newValue;
                                RefreshGroupUI(LeftRight.left, 1);
                            });
                            CreatePopup(groupID, baFilterParamJSEnum, labelWidth: 225f, scrollable: false, helpText: FileReferencesComponentUIHelp.baSortModeFilter);
                        }
                    }

                    int buttonType = fileRef.parentButtonOperation.buttonOpTypeJSEnum.val;
                    if (fileRef.fileReferenceCategory == FileReferenceCategory.presetCat && !currentUIAButtonOperation.savePresetJSB.val && (buttonType != UIAButtonOpType.spawnAtom || fileRef.fileReferenceType == FileReferenceTypes.pluginsPreset) && !UIAButtonOpType.IsUIAClothingPresetLoadType(buttonType))
                    {
                        CreateToggle(groupID, fileRef.mergeLoadPresetJSBool, false, "Merge Load Preset", helpText: FileReferencesComponentUIHelp.mergeLoad);
                        if (buttonType == UIAButtonOpType.loadClothPreset) ForceUIRefreshOnValChange(fileRef.mergeLoadPresetJSBool, LeftRight.left, 2);
                    }
                    if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadUIAProfile && fileRef.fileReferenceType == FileReferenceTypes.uiap)
                    {
                        CreateToggle(groupID, fileRef.closeGridOnLoadUIAPJSB, false, "Close Grids on UIAP Load", helpText: FileReferencesComponentUIHelp.closeGridOnLoadUIAP);
                    }
                    if (fileRef.fileReferenceCategory == FileReferenceCategory.presetCat && currentUIAButtonOperation.savePresetJSB.val && fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.singleFile)
                    {
                        CreateToggle(groupID, fileRef.forceUniqueSaveNameJSB, false, "Force unique filename on save", helpText: FileReferencesComponentUIHelp.forceUniqueSaveName);
                    }

                    if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadMorphPreset )
                    {
                        CreateToggle(groupID,fileRef.morphPresetIncludePhysicalJSB, false, "Include Physical (Pose)", helpText: FileReferencesComponentUIHelp.morphPresetIncludePhysical);
                        CreateToggle(groupID, fileRef.morphPresetIncludeAppJSB, false, "Include Appearance", helpText: FileReferencesComponentUIHelp.morphPresetIncludeApp);
                    }
                    if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadGeneralPreset)
                    {
                        CreateToggle(groupID, fileRef.generalPresetIncludeAppJSB, false, "Include Appearance", helpText: FileReferencesComponentUIHelp.generalPresetIncludeApp);
                        CreateToggle(groupID, fileRef.generalPresetIncludePhysicalJSB, false, "Include Physical", helpText: FileReferencesComponentUIHelp.generalPresetIncludePhysical);
                        CreateToggle(groupID, fileRef.generalPresetIncludePoseJSB, false, "Include Pose", helpText: FileReferencesComponentUIHelp.generalPresetIncludePose);
                    }
                    if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadPosePreset && !currentUIAButtonOperation.savePresetJSB.val)
                    {
                        CreateToggle(groupID, fileRef.posePresetSnapBoneToPoseJSB, false, "Snap Bones to Pose", helpText: FileReferencesComponentUIHelp.posePresetSnapBoneToPose);
                    }
                    CreateSpacer(groupID);

                }


                if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadSkinPreset && PatreonFeatures.patreonContentEnabled)
                {
                    CreateToggle(groupID, skinDecalComponent.decalFaceLoadJSBool, false, "Limit preset load to Face Decal", helpText: FileReferencesComponentUIHelp.faceDecalLimit);
                    CreateToggle(groupID, skinDecalComponent.decalBodyLoadJSBool, false, "Limit preset load to Body Decal", helpText: FileReferencesComponentUIHelp.bodyDecalLimit);
                    CreateToggle(groupID, skinDecalComponent.decalGenLoadJSBool, false, "Limit preset load to Gen Decal", helpText: FileReferencesComponentUIHelp.genDecalLimit);
                    CreateToggle(groupID, skinDecalComponent.decalLimbsLoadJSBool, false, "Limit preset load to Limbs Decal", helpText: FileReferencesComponentUIHelp.limbsDecalLimit);
                }

                if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadAppPreset && !currentUIAButtonOperation.savePresetJSB.val)
                {
                    CreateToggle(groupID, appearancePresetComponent.suppressClothingLoadJSBool, false, "Suppress Clothing Load", helpText: FileReferencesComponentUIHelp.suppressClothingLoad);

                    ForceUIRefreshOnValChange(appearancePresetComponent.suppressClothingLoadJSBool, LeftRight.left, 2);

                }
                if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadMotionCapture)
                {
                    MotionCaptureComponent mocapComponent = currentUIAButtonOperation.motionCaptureComponent;
                    CreatePopup(groupID, mocapComponent.motionCaptureAtomNameJSSC, false, 400f, mocapComponent.RefreshMocapAtomChoices, "Select Person from Scene", false, helpText: FileReferencesComponentUIHelp.mocapPerson);
                    CreateToggle(groupID, mocapComponent.mocapLoadHeelAdjustJSB, false, "Apply HeelAdjust foot rotations", helpText: FileReferencesComponentUIHelp.mocapHeelAdjust);
                }

                
            }
            else if (groupID == 2 && !isRight)
            {
                if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadAppPreset && !currentUIAButtonOperation.savePresetJSB.val)
                {

                    if (appearancePresetComponent.suppressClothingLoadJSBool.val && UIAGlobals.isRealClothingAvailable) CreateToggle(groupID, appearancePresetComponent.onlySuppressRealClothingJSB, false, "Only suppress Real Clothing", helpText: FileReferencesComponentUIHelp.onlySuppressRealClothing);
                    if (PatreonFeatures.patreonContentEnabled)
                    {
                        CreateSpacer(groupID);
                        CreateToggle(groupID, appearancePresetComponent.suppressPersonScaleLoadJSBool, false, "Suppress Person Scale Load", helpText: FileReferencesComponentUIHelp.suppressScaleLoad);
                    }
                }
                if (currentUIAButtonOperation.buttonOpTypeJSEnum.val == UIAButtonOpType.loadClothPreset && !fileReferences[FileReferenceTypes.clothingPreset].mergeLoadPresetJSBool.val && UIAGlobals.isRealClothingAvailable) CreateToggle(groupID, appearancePresetComponent.onlyReplaceRealClothingJSB, false, "Only replace Real Clothing",helpText: FileReferencesComponentUIHelp.onlyReplaceRealClothing);
            }
        }

        private void RefreshBAFilterUserTagJSSC(FileReference fileRef)
        {
            baFilterUserTagCreatorJSSC.choices = null;
            BAInterop.GetTagNames(fileRef.fileReferenceType, delegate (List<string> tagNamesList)
            {
                baFilterUserTagCreatorJSSC.choices = tagNamesList;
            },delegate (string message)
            {
                SuperController.LogMessage(message);
            });
        }

        private void RefreshBAFilterCreatorJSSC(FileReference fileRef)
        {
            baFilterUserTagCreatorJSSC.choices = null;
            BAInterop.GetCreatorNames(fileRef.fileReferenceType, delegate (List<string> creatorNamesList)
            {
                baFilterUserTagCreatorJSSC.choices = creatorNamesList;
            }, delegate (string message)
            {
                SuperController.LogMessage(message);
            });
        }

    }

    public class HairColorComponentUITab : BaseUIButtonTreeTab
    {
        HairColorComponent hairColorComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.hairColorComponent; } }

        JSONStorableStringChooser hairColorAtomSelector;
        JSONStorableStringChooser scalpColorAtomSelector;
        JSONStorableString hairSimControlValue;
        JSONStorableString hairScalpControlValue;

        public HairColorComponentUITab() : base()
        {
            uiGroupCount = 3;
            defaultHelpText = HairColorComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            hairColorAtomSelector = new JSONStorableStringChooser("hairColorAtomSelector", null, "Select...", "", SwitchHairColor);
            scalpColorAtomSelector = new JSONStorableStringChooser("scalpColorAtomSelector", null, "Select...", "", SwitchScalpColor);
            hairSimControlValue = new JSONStorableString("hairSimControlValue", "");
            hairScalpControlValue = new JSONStorableString("hairScalpControlValue", "");
        }

        protected override void CreateGroupUI(bool isRightUI, int groupID)
        {
            if (isRightUI) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRightUI)
            {
                UpdateHairSimChoices();
                CreatePopup(groupID, hairColorAtomSelector, false, 400f, UpdateHairSimChoices, "Get Hair Color from:", false, helpText: HairColorComponentUIHelp.hairColorFrom);
                ForceUIRefreshOnValChange(hairColorAtomSelector, LeftRight.left, 2);
            }
            if (groupID == 2 && !isRightUI)
            {
                if (hairColorComponent.hairSimControlJC != null)
                {
                    hairSimControlValue.val = hairColorComponent.hairSimControlJC.ToString();
                    CreateTextField(groupID, hairSimControlValue, helpText: HairColorComponentUIHelp.hairColorJson);
                }
                UpdateHairScalpChoices();
                CreatePopup(groupID, scalpColorAtomSelector, false, 400f, UpdateHairScalpChoices, "Scalp Hair Color from:", false, helpText: HairColorComponentUIHelp.scalpColorFrom);
                ForceUIRefreshOnValChange(scalpColorAtomSelector, LeftRight.left, 3);
            }
            if (groupID == 3 && !isRightUI)
            {
                if (hairColorComponent.hairScalpControlJC != null)
                {
                    hairScalpControlValue.val = hairColorComponent.hairScalpControlJC.ToString();
                    CreateTextField(groupID, hairScalpControlValue, helpText: HairColorComponentUIHelp.scalpColorJson);
                }
            }
        }


        private void UpdateHairSimChoices()
        {
            List<string> popupChoices = new List<string>();

            foreach (Atom atom in AtomUtils.GetAtoms())
            {
                foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                {
                    DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
                    HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

                    if (hairControl != null && simControl != null)
                    {
                        popupChoices.Add(atom.name + ">" + hairGroup.displayName);
                    }
                }
            }
            if (popupChoices.Count == 0) popupChoices.Add("No Sim Hair available");
            if (hairColorComponent.hairSimControlJC != null) popupChoices.Add("Clear Hair Color");

            hairColorAtomSelector.choices = popupChoices;

        }
        private void UpdateHairScalpChoices()
        {
            List<string> popupChoices = new List<string>();

            foreach (Atom atom in AtomUtils.GetAtoms())
            {
                foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                {
                    DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
                    HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

                    if (hairControl != null && simControl != null)
                    {
                        popupChoices.Add(atom.name + ">" + hairGroup.displayName);
                    }
                }
            }

            if (popupChoices.Count == 0) { popupChoices.Add("No Sim Hair available"); }
            if (hairColorComponent.hairScalpControlJC
                != null) popupChoices.Add("Clear Scalp Color");
            scalpColorAtomSelector.choices = popupChoices;
        }

        private void SwitchHairColor(string stringValue)
        {

            if (stringValue == "Clear Hair Color")
            {
                hairColorComponent.hairSimControlJC = null;
                hairSimControlValue.valNoCallback = "";
            }
            else if (stringValue != "No Sim Hair available")
            {
                string atomName = stringValue.Substring(0, stringValue.IndexOf('>'));
                string hairName = stringValue.Substring(atomName.Length + 1);

                Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                if (atom != null)
                {
                    foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                    {
                        HairSimControl hsc = hairGroup.GetComponentInChildren<HairSimControl>();
                        if (hairGroup.displayName == hairName && hsc != null)
                        {
                            if (hairColorComponent.hairSimControlJC == null) hairColorComponent.hairSimControlJC = new JSONClass();
                            hairColorComponent.hairSimControlJC = hsc.GetJSON(true, true, true);
                            break;
                        }
                    }
                }
            }

            hairColorAtomSelector.valNoCallback = "Select...";
        }
        private void SwitchScalpColor(string stringValue)
        {
            if (stringValue == "Clear Scalp Color")
            {
                hairColorComponent.hairScalpControlJC = null;
                hairScalpControlValue.valNoCallback = "";
            }
            else if (stringValue != "No Sim Hair available")
            {
                string atomName = stringValue.Substring(0, stringValue.IndexOf('>'));
                string hairName = stringValue.Substring(atomName.Length + 1);

                Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                if (atom != null)
                {
                    foreach (DAZHairGroup hairGroup in atom.GetComponentsInChildren<DAZHairGroup>())
                    {
                        if (hairGroup.displayName == hairName && hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>() != null)
                        {
                            hairColorComponent.hairScalpControlJC = new JSONClass();
                            hairColorComponent.hairScalpControlJC = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>().GetJSON(true, true, true);
                            break;
                        }
                    }
                }
            }
            scalpColorAtomSelector.valNoCallback = "Select...";
        }

    }
    public class PresetLockComponentUITab : BaseUIButtonTreeTab
    {
        PresetLockComponent presetLockComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.presetLockComponent; } }

        public PresetLockComponentUITab() : base()
        {
            uiGroupCount = 1;
            defaultHelpText = PresetLockComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                CreateToggle(groupID, presetLockComponent.generalPresetLockJSON, false, "Toggle General Preset Lock");
                CreateToggle(groupID, presetLockComponent.appPresetLockJSON, false, "Toggle Appearance Preset Lock");
                CreateToggle(groupID, presetLockComponent.posePresetLockJSON, false, "Toggle Pose Preset Lock");
                CreateToggle(groupID, presetLockComponent.animationPresetLockJSON, false, "Toggle Animation Preset Lock");
                CreateToggle(groupID, presetLockComponent.glutePhysPresetLockJSON, false, "Toggle Glute Physics Preset Lock");
                CreateToggle(groupID, presetLockComponent.breastPhysPresetLockJSON, false, "Toggle Breast Physics Preset Lock");
                CreateToggle(groupID, presetLockComponent.pluginPresetLockJSON, false, "Toggle Plugin Preset Lock");
                CreateToggle(groupID, presetLockComponent.skinPresetLockJSON, false, "Toggle Skin Preset Lock");
                CreateToggle(groupID, presetLockComponent.morphPresetLockJSON, false, "Toggle Morph Preset Lock");
                CreateToggle(groupID, presetLockComponent.hairPresetLockJSON, false, "Toggle Hair Preset Lock");
                CreateToggle(groupID, presetLockComponent.clothingPresetLockJSON, false, "Toggle Clothing Preset Lock");
                CreateSpacer(groupID);

                var toggle = CreateToggle(groupID, presetLockComponent.toggleOnKeepPresetLockJSON, false, "Toggle On: 'Keep Preset Locks between Scene Loads'", helpText: PresetLockComponentUIHelp.toggleOnKeep);
                toggle.height = 80f;
                toggle = CreateToggle(groupID, presetLockComponent.toggleOffKeepPresetLockJSON, false, "Toggle Off: Disable 'Keep Preset Locks between Scene Loads'", helpText: PresetLockComponentUIHelp.toggleOffDisable);
                toggle.height = 80f;
            }

        }

    }

    public class PluginLoadComponentUITab : BaseUIButtonTreeTab
    {
        private PluginsLoadComponent pluginsLoadComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.pluginsLoadComponent; } }

        private JSONStorableStringChooser pluginSelectorJSSC;

        private JSONStorableStringChooser pluginDataSelectorJSSC;

        private ButtonPluginLoad _currentButtonPluginLoad;
        private ButtonPluginLoad currentButtonPluginLoad { get
            {
                if (_currentButtonPluginLoad == null) _currentButtonPluginLoad = pluginsLoadComponent.pluginLoadList[0];
                if (!pluginsLoadComponent.pluginLoadList.Contains(_currentButtonPluginLoad)) _currentButtonPluginLoad = pluginsLoadComponent.pluginLoadList[0];
                return _currentButtonPluginLoad;
            }
            set
            {
                _currentButtonPluginLoad = value;
            }
        }

        public PluginLoadComponentUITab() : base()
        {
            uiGroupCount = 4;
            defaultHelpText = PluginLoadComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            pluginSelectorJSSC = new JSONStorableStringChooser("pluginSelector", null, "1", "", PluginSelectorUpdated);
            pluginDataSelectorJSSC = new JSONStorableStringChooser("pluginDataSelector", null, "Select...", "", SaveDataPluginSelectorUpdated);
        }


        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                PluginSelectorPopupOpen();
                PluginSelectorUpdated(pluginSelectorJSSC.val);
                List<string> ignoreVals = new List<string>();
                ignoreVals.Add("Add Plugin...");
                List<string> ignoreValPatterns = new List<string>();
                ignoreValPatterns.Add("Remove Plugin");
                CreatePopup(groupID, pluginSelectorJSSC, false, 600, PluginSelectorPopupOpen, "Plugin\nLoad #", true, PluginLoadComponentUIHelp.pluginLoadSelector, 300f, true, ignoreVals, ignoreValPatterns);
                ForceUIRefreshOnValChange(pluginSelectorJSSC, LeftRight.left, 2);
            }
            else if (groupID == 2 && !isRight)
            {
                CreateButton(groupID, "Select Plugin File", false, SelectPluginFile, helpText: PluginLoadComponentUIHelp.pluginFileSelect);
                CreateTextField(groupID, currentButtonPluginLoad.filePathJSString, helpText: PluginLoadComponentUIHelp.pluginSelectedFile);
                CreateToggle(groupID, pluginsLoadComponent.useLatestVARJSBool, false, "Always use latest VAR package", helpText: PluginLoadComponentUIHelp.useLatestVAR);

                if (pluginsLoadComponent.pluginLoadList.First()==currentButtonPluginLoad) CreateToggle(groupID, pluginsLoadComponent.openPluginUIOnLoadJSB, false, "Open Plugin UI on Load",helpText: PluginLoadComponentUIHelp.openPluginUIOnLoad);

                CreateToggle(groupID, currentButtonPluginLoad.singleInstancePerAtomJSBool, false, "Limit to 1 plugin instance per Atom", helpText: PluginLoadComponentUIHelp.loadLimiter);
                ForceUIRefreshOnValChange(currentButtonPluginLoad.singleInstancePerAtomJSBool, LeftRight.left, 3);
            }
            else if (groupID == 3 && !isRight)
            {
                if (currentButtonPluginLoad.singleInstancePerAtomJSBool.val)
                {
                    CreateToggle(groupID, currentButtonPluginLoad.restorePluginDataIfLoadedJSBool, false, "Restore plugin data if plugin loaded", helpText: PluginLoadComponentUIHelp.dataRestore);
                }
                CreateSpacer(groupID);
                CreatePopup(groupID, pluginDataSelectorJSSC, false, 300f, PluginDataSelectorPopupOpen, "Capture initial data", false, helpText: PluginLoadComponentUIHelp.initialPluginDataSelector);
                CreateTextField(groupID, currentButtonPluginLoad.pluginSaveDataJSString, helpText: PluginLoadComponentUIHelp.initialData);

            }

        }

        private void PluginDataSelectorPopupOpen()
        {
            pluginDataSelectorJSSC.choices = GetPluginDataSelectorChoices();
        }
        private List<string> GetPluginDataSelectorChoices()
        {
            List<string> pluginChoices = new List<string>();
            foreach (Atom atom in AtomUtils.GetAtoms(includeCoreControl:true))
            {
                List<string> pluginRefsInAtom = new List<string>();
                MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;

                pluginRefsInAtom = PluginUtils.GetPluginRefsOfType(manager, currentButtonPluginLoad.filePathJSString.val);
                foreach (string pluginRef in pluginRefsInAtom)
                {
                    if (atom.name =="CoreControl") pluginChoices.Add(pluginRef + "[SCENE]");
                    else pluginChoices.Add(pluginRef + "[" + atom.name + "]");
                }
            }

            List<string> pluginRefsInSessionPluginMgr = PluginUtils.GetPluginRefsOfType(UIAGlobals.mvrScript.manager, currentButtonPluginLoad.filePathJSString.val);
            foreach (string pluginRef in pluginRefsInSessionPluginMgr) pluginChoices.Add(pluginRef + "[SESSION]");

            if (currentButtonPluginLoad.pluginSaveDataJSString.val != "" && currentButtonPluginLoad.pluginSaveJSON != null) pluginChoices.Add("<Clear Plugin Data>");
            if (pluginChoices.Count == 0) pluginChoices.Add("<None available>");

            return pluginChoices;
        }



        private void SaveDataPluginSelectorUpdated(string pluginChoice)
        {

            if (pluginChoice == "<Clear Plugin Data>")
            {
                currentButtonPluginLoad.pluginSaveDataJSString.val = "";
                currentButtonPluginLoad.pluginSaveJSON = null;
            }
            else if (pluginChoice != "<None available>")
            {
                string pluginName = pluginChoice.Substring(0, pluginChoice.IndexOf('['));
                string atomName = pluginChoice.Substring(pluginChoice.IndexOf('[') + 1);
                JSONArray pluginDataJSON = PluginUtils.GetPluginDataByName(atomName.Remove(atomName.Length - 1), pluginName);
                string pluginData = pluginDataJSON.ToString();
                currentButtonPluginLoad.pluginSaveDataJSString.val = pluginData.Substring(0, Math.Min(pluginData.Length, 500));
                currentButtonPluginLoad.pluginSaveJSON = pluginDataJSON;

            }

            pluginDataSelectorJSSC.valNoCallback = "Select...";
        }


        private void SelectPluginFile()
        {
            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Scripts", false, false, true);
            SuperController.singleton.GetMediaPathDialog(PluginFileSelected, "cs|cslist|dll", "Custom\\Scripts", false, true, false, null, false, shortcuts);
        }
        private void PluginFileSelected(string fileName)
        {
            if (fileName != "")
            {
                if (fileName.IndexOf(':') == 1)
                {
                    currentButtonPluginLoad.filePathJSString.val = fileName.Substring(UIAConsts._vamFilePath.Length, fileName.Length - UIAConsts._vamFilePath.Length);
                }
                else currentButtonPluginLoad.filePathJSString.val = fileName;
            }
        }

        private List<string> GetPluginSelectorChoices()
        {
            List<string> choices = new List<string>();


            for (int i = 1; i <= pluginsLoadComponent.pluginLoadList.Count; i++)
            {
                choices.Add(i.ToString());
            }

            choices.Add("Add Plugin...");
            if (pluginsLoadComponent.pluginLoadList.Count > 1)
            {
                if (pluginSelectorJSSC == null) choices.Add("Remove Plugin 1");
                else choices.Add("Remove Plugin " + pluginSelectorJSSC.val);
            }

            return choices;
        }

        private void PluginSelectorPopupOpen()
        {
            pluginSelectorJSSC.choices = GetPluginSelectorChoices();

        }

        private void PluginSelectorUpdated(string pluginRef)
        {
            if (pluginRef == "Add Plugin...")
            {
                ButtonPluginLoad newPluginLoad = new ButtonPluginLoad();
                pluginsLoadComponent.pluginLoadList.Add(newPluginLoad);
                pluginSelectorJSSC.val = pluginsLoadComponent.pluginLoadList.Count.ToString();
            }
            else if (pluginRef.StartsWith("Remove Plugin"))
            {
                pluginsLoadComponent.pluginLoadList.RemoveAt(int.Parse(pluginRef.Substring(14)) - 1);
                pluginSelectorJSSC.val = "1";
            }
            else currentButtonPluginLoad = pluginsLoadComponent.pluginLoadList[int.Parse(pluginRef) - 1];

            PluginSelectorPopupOpen();
        }

    }

    public class PluginSettingsUITab : BaseUIButtonTreeTab
    {
        PluginSettingComponent pluginSettingComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.pluginSettingComponent; } }
        UIAButtonOperation buttonOperation { get { return ButtonSetupPluginUI.currentUIAButtonOperation; } }

        JSONStorableStringChooser buttonOnSettingParamSelectorJSSC;
        JSONStorableStringChooser buttonOffSettingParamSelectorJSSC;
        JSONStorableBool varSet;
        JSONStorableBool offVarSet;

        int buttonType { get { return buttonOperation.buttonOpTypeJSEnum.val; } }
        bool isActionType { get { return (buttonType == UIAButtonOpType.pluginAction || buttonType == UIAButtonOpType.pluginActionToggle); } }
        bool isBoolType { get { return (buttonType == UIAButtonOpType.pluginBoolFalse || buttonType == UIAButtonOpType.pluginBoolTrue || buttonType == UIAButtonOpType.pluginBoolToggle); } }
        bool isToggleType { get { return (buttonType == UIAButtonOpType.pluginActionToggle || buttonType == UIAButtonOpType.pluginMultiSettingsToggle || buttonType == UIAButtonOpType.pluginBoolToggle); } }

        public PluginSettingsUITab() : base()
        {
            uiGroupCount = 5;
            defaultHelpText = PluginSettingsUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            buttonOnSettingParamSelectorJSSC = new JSONStorableStringChooser("buttonOnSettingParamSelector", null, "Select Parameter...", "");
            buttonOffSettingParamSelectorJSSC = new JSONStorableStringChooser("buttonOffSettingParamSelector", null, "Select Parameter...", "");
            varSet = new JSONStorableBool("varSet", false, SwitchPluginOnVariableSet);
            offVarSet = new JSONStorableBool("offVarSet", false, SwitchPluginOffVariableSet);
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            string varType = "";
            string label = "";
            string helpText;
            if (isBoolType) varType = "Bool";
            else if (isActionType) varType = "Action";
            bool validTargetParamNameOn = (pluginSettingComponent.targetParamNameOnJSSC.val != "No Actions available" && pluginSettingComponent.targetParamNameOnJSSC.val != "");
            bool validTargetParamNameOff = (pluginSettingComponent.targetParamNameOffJSSC.val != "No Actions available" && pluginSettingComponent.targetParamNameOffJSSC.val != "");
            bool isSettingType = !isBoolType && !isActionType;

            if (isRight) CreateHelpAndButtonTree(groupID, 945);           
            else if (groupID == 1 && !isRight )
            {
                buttonOnSettingParamSelectorJSSC.val = "Select Parameter...";
                buttonOffSettingParamSelectorJSSC.val = "Select Parameter...";

                string pluginSelectorLabel = "Plugin Type";

                if (buttonType == UIAButtonOpType.pluginOpenUI || buttonType == UIAButtonOpType.pluginToggleOpenUI) pluginSelectorLabel = "Plugin UI Type";

                CreatePopup(groupID, pluginSettingComponent.pluginTypeJSSC, isRight, 400f, RefreshPluginTypeChoices, pluginSelectorLabel, helpText: PluginSettingsUIHelp.pluginSelector);
                ForceUIRefreshOnValChange(pluginSettingComponent.pluginTypeJSSC, LeftRight.left, 1);

                if (buttonType != UIAButtonOpType.pluginOpenUI && buttonType != UIAButtonOpType.pluginToggleOpenUI)
                {
                    if (isBoolType || isActionType)
                    {
                        helpText = PluginSettingsUIHelp.pluginAction;
                        if (isBoolType)
                        {
                            helpText = PluginSettingsUIHelp.pluginBool;
                            if (!isToggleType)
                            {
                                helpText = PluginSettingsUIHelp.pluginBoolSetTrue;
                                if (buttonType == UIAButtonOpType.pluginBoolFalse) helpText = PluginSettingsUIHelp.pluginBoolSetFalse;
                            }
                        }
                        if (!isToggleType || isBoolType) label = "Select " + varType;
                        else
                        {
                            label = "Button ON " + varType;
                            helpText = PluginSettingsUIHelp.pluginActionOn;
                        }
                        CreatePopup(groupID, pluginSettingComponent.targetParamNameOnJSSC, false, 400f, pluginSettingComponent.RefreshPluginParamNameChoices, label, helpText: helpText);
                        ForceUIRefreshOnValChange(pluginSettingComponent.targetParamNameOnJSSC, LeftRight.left, 1);
                    }
                    if (isSettingType || validTargetParamNameOn)
                    {
                        label = "Plugin Settings";
                        if (isToggleType) label = "Button ON Settings";

                        helpText = PluginSettingsUIHelp.pluginSettingsParamSelector;
                        if (isToggleType) helpText = PluginSettingsUIHelp.pluginSettingsONParamSelector;
                        CreatePopup(groupID, buttonOnSettingParamSelectorJSSC, false, 400, RefreshVariableOnChoices, label, helpText: helpText);
                        ForceUIRefreshOnValChange(buttonOnSettingParamSelectorJSSC, LeftRight.left, 2);
                    }

                }


            }
            else if (groupID == 2 && !isRight && (isSettingType || validTargetParamNameOn))
            {
                if (buttonOnSettingParamSelectorJSSC.val != "Select Parameter...")
                {
                    varSet.valNoCallback = pluginSettingComponent.pluginVariablesButtonOn.GetVarSet(buttonOnSettingParamSelectorJSSC.val);

                    CreateToggle(groupID, varSet, isRight, "Modify plugin setting", helpText: PluginSettingsUIHelp.modifyPluginSetting);
                    ForceUIRefreshOnValChange(varSet, LeftRight.left, 3);
                }
            }
            else if (groupID == 3 && !isRight )
            {
                if (pluginSettingComponent.pluginVariablesButtonOn.GetVarSet(buttonOnSettingParamSelectorJSSC.val) && (isSettingType || validTargetParamNameOn))
                {
                    CreatePluginSettingUI(isRight, groupID, buttonOnSettingParamSelectorJSSC.val, pluginSettingComponent.pluginVariablesButtonOn);
                }

                if (isToggleType)
                {
                    CreateSpacer(groupID);
                    if (isActionType)
                    {
                        label = "Button OFF " + varType;
                        CreatePopup(groupID, pluginSettingComponent.targetParamNameOffJSSC, false, 400f, pluginSettingComponent.RefreshPluginParamNameChoices, label, helpText: PluginSettingsUIHelp.pluginActionOn);
                        ForceUIRefreshOnValChange(pluginSettingComponent.targetParamNameOffJSSC, LeftRight.left, 3);
                    }

                    if (isSettingType || validTargetParamNameOff)
                    {
                        label = "Button OFF Settings";

                        CreatePopup(groupID, buttonOffSettingParamSelectorJSSC, false, 400, RefreshVariableOffChoices, label, helpText: PluginSettingsUIHelp.pluginSettingsOFFParamSelector);
                        ForceUIRefreshOnValChange(buttonOffSettingParamSelectorJSSC, LeftRight.left, 4);
                    }
                    
                }
            }
            else if (groupID == 4 && !isRight && (isSettingType || validTargetParamNameOff))
            {
                if (buttonOffSettingParamSelectorJSSC.val != "Select Parameter...")
                {
                    offVarSet.valNoCallback = pluginSettingComponent.pluginVariablesButtonOff.GetVarSet(buttonOffSettingParamSelectorJSSC.val);

                    CreateToggle(groupID, offVarSet, isRight, "Modify plugin setting", helpText: PluginSettingsUIHelp.modifyPluginSetting);
                    ForceUIRefreshOnValChange(offVarSet, LeftRight.left, 5);
                }
            }
            else if (groupID == 5 && !isRight && (isSettingType || validTargetParamNameOff))
            {
                if (pluginSettingComponent.pluginVariablesButtonOff.GetVarSet(buttonOffSettingParamSelectorJSSC.val))
                {
                    CreatePluginSettingUI(isRight, groupID, buttonOffSettingParamSelectorJSSC.val, pluginSettingComponent.pluginVariablesButtonOff);
                }
            }
        }

        private void CreatePluginSettingUI(bool isRight, int groupID, string selectedParam, PluginVariables pluginVariables)
        {
            PluginVariable pv = pluginVariables._pluginVarDict[selectedParam];
            int varType = pv.pluginVarType.val;

            switch (varType)
            {
                case PluginVariableType.vFloat:
                    JSONStorableFloat pvFloat = pv.pluginFloatValue;
                    pvFloat.min = pluginVariables.GetMinFloatValue(selectedParam);
                    pvFloat.max = pluginVariables.GetMaxFloatValue(selectedParam);
                    pvFloat.constrained = false;
                    CreateSlider(groupID, pvFloat, isRight, selectedParam, helpText: PluginSettingsUIHelp.floatPluginSetting);
                    break;
                case PluginVariableType.vBool:
                    JSONStorableBool pvBool = pv.pluginBoolValue;
                    CreateToggle(groupID, pvBool, isRight, selectedParam, helpText: PluginSettingsUIHelp.boolPluginSetting);
                    break;
                case PluginVariableType.vStringChooser:
                    JSONStorableString pvString = pv.pluginStringValue;
                    JSONStorableStringChooser pvStringChooser =null;
                    if (pluginVariables== pluginSettingComponent.pluginVariablesButtonOn)
                    {
                        pvStringChooser = new JSONStorableStringChooser("pvStringChooser", null, pvString.val, "", OnPVStringChooserCallback);
                    }
                    else
                    {
                        pvStringChooser = new JSONStorableStringChooser("pvStringChooser", null, pvString.val, "", OffPVStringChooserCallback);
                    }
                    pvStringChooser.choices = pluginVariables.GetStringChooserChoices(selectedParam);
                    CreatePopup(groupID, pvStringChooser, isRight, 400f, null, "Set '" + selectedParam + "'", helpText: PluginSettingsUIHelp.stringChooserPluginSetting);
                    break;
                case PluginVariableType.vString:
                case PluginVariableType.vURL:
                    JSONStorableString pvString1 = pv.pluginStringValue;
                    CreateTextField(groupID, pvString1, isRight, 100, true, InputField.LineType.MultiLineNewline, "Set '" + selectedParam + "'", helpText: PluginSettingsUIHelp.stringPluginSetting);
                    break;
                case PluginVariableType.vColor:
                    JSONStorableColor pvColor = pv.pluginColorValue;
                    CreateColorPicker(groupID, pvColor, isRight, selectedParam, helpText: PluginSettingsUIHelp.colorPluginSetting);
                    break;

            }
        }

        private void OnPVStringChooserCallback(string value)
        {
            PluginVariable pv = pluginSettingComponent.pluginVariablesButtonOn._pluginVarDict[buttonOnSettingParamSelectorJSSC.val];
            pv.pluginStringValue.val = value;
        }
        private void OffPVStringChooserCallback(string value)
        {
            PluginVariable pv = pluginSettingComponent.pluginVariablesButtonOff._pluginVarDict[buttonOffSettingParamSelectorJSSC.val];
            pv.pluginStringValue.val = value;
        }
        private void SwitchPluginOnVariableSet(bool setVal)
        {
            PluginVariables pluginVariables = pluginSettingComponent.pluginVariablesButtonOn;
            SwitchPluginVariablesSet(setVal, buttonOnSettingParamSelectorJSSC.val, pluginVariables);
        }

        private void SwitchPluginOffVariableSet(bool setVal)
        {
            PluginVariables pluginVariables = pluginSettingComponent.pluginVariablesButtonOff;
            SwitchPluginVariablesSet(setVal, buttonOffSettingParamSelectorJSSC.val, pluginVariables);
        }

        private void SwitchPluginVariablesSet(bool setVal, string selectedParam, PluginVariables pluginVariables)
        {
            if (setVal && !pluginVariables._pluginVarDict.ContainsKey(selectedParam))
            {
                PluginVariable pluginVar = new PluginVariable(selectedParam, pluginVariables.GetVarTypeFromPlugin(selectedParam));

                pluginVariables._pluginVarDict.Add(selectedParam, pluginVar);
            }
            else if (!setVal && pluginVariables._pluginVarDict.ContainsKey(selectedParam))
            {
                pluginVariables._pluginVarDict.Remove(selectedParam);
            }
        }
        private void RefreshPluginTypeChoices()
        {
            pluginSettingComponent.pluginTypeJSSC.choices = pluginSettingComponent.GetPluginTypeChoices();
        }

        private void RefreshVariableOnChoices()
        {
            pluginSettingComponent.pluginVariablesButtonOn.SetPluginParamChoices(buttonOnSettingParamSelectorJSSC);
        }

        private void RefreshVariableOffChoices()
        {
            pluginSettingComponent.pluginVariablesButtonOff.SetPluginParamChoices(buttonOffSettingParamSelectorJSSC);
        }
    }

    public class ButtonSkinComponentUITab : BaseUIButtonTreeTab
    {
        protected virtual ButtonSkinComponent buttonSkinComponent { get
            {
                return ButtonSetupPluginUI.currentUIAButton.buttonSkinComponent;
            }
        }

        private UIAButton currentUIAButton { get
            {
                return ButtonSetupPluginUI.currentUIAButton;
            }
        }

        CustomRowUIElement skinUIOptionsTabSelector;

        static string lastLoadedThumbnailFolder = "";

        const int textTabEV = 1;
        const int colorOnTabEV = 2;
        const int colorOffTabEV = 3;
        const int thumbnailTabEV = 4;

        private int currentSkinTab = textTabEV;

        public ButtonSkinComponentUITab() : base()
        {
            uiGroupCount = 6;
            defaultHelpText = ButtonSkinComponentUIHelp.defaultHelpText;

            currentHelpTextJSS.val = defaultHelpText;
            skinUIOptionsTabSelector = new CustomRowUIElement();
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight)
            {
                if (groupID == 1)
                {
                    if (currentUIAButton.maxButtonStates > 1) skinUIOptionsTabSelector.InitUI(4);
                    else skinUIOptionsTabSelector.InitUI(3);

                    var button = skinUIOptionsTabSelector.AddButtonToRow("Text", textTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, ButtonSkinComponentUIHelp.textTab);
                    if (currentUIAButton.maxButtonStates > 1)
                    {
                        button = skinUIOptionsTabSelector.AddButtonToRow("On Colors", colorOnTabEV, TabSelected);
                        RegisterContextualHelp(button.gameObject, ButtonSkinComponentUIHelp.onColorsTab);
                        button = skinUIOptionsTabSelector.AddButtonToRow("Off Colors", colorOffTabEV, TabSelected);
                        RegisterContextualHelp(button.gameObject, ButtonSkinComponentUIHelp.offColorsTab);
                    }
                    else
                    {
                        button = skinUIOptionsTabSelector.AddButtonToRow("Colors", colorOnTabEV, TabSelected);
                        RegisterContextualHelp(button.gameObject, ButtonSkinComponentUIHelp.colorsTab);
                    }
                    button = skinUIOptionsTabSelector.AddButtonToRow("Thumbnail", thumbnailTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, ButtonSkinComponentUIHelp.thumbnailTab);
                    RegisterCustomUIElement(groupID, skinUIOptionsTabSelector, isRight);

                }
                CreateHelpAndButtonTree(groupID, 880);
            }

            if (!isRight) skinUIOptionsTabSelector.SetButtonColorActive(currentSkinTab);
            switch (currentSkinTab)
            {
                case textTabEV:
                    if (groupID == 2 && !isRight)
                    {
                        if (currentUIAButton.buttonOperations.Count == 1)
                        {
                            CreateToggle(groupID, currentUIAButton.autoLabelJSBool, false, "Auto Label", 50f, helpText: ButtonSkinComponentUIHelp.autoLabel);
                            ForceUIRefreshOnValChange(currentUIAButton.autoLabelJSBool, LeftRight.left, 3);
                        }
                    }
                    else if (groupID == 3 && !isRight)
                    {
                        if (!currentUIAButton.autoLabelJSBool.val || currentUIAButton.buttonOperations.Count > 1)
                        {
                            string placeHolderText = " Button On Label";
                            if (currentUIAButton.maxButtonStates == 1) placeHolderText = "Button Label";

                            CreateTextField(groupID, currentUIAButton.buttonLabelOnJSString, false, 100f, true, InputField.LineType.SingleLine, placeHolderText, helpText: ButtonSkinComponentUIHelp.buttonLabel);

                            if (currentUIAButton.maxButtonStates > 1) CreateTextField(groupID, currentUIAButton.buttonLabelOffJSString, false, 100f, true, InputField.LineType.SingleLine, " Button Off Label", helpText: ButtonSkinComponentUIHelp.buttonOffLabel);
                        }
                        CreateToggle(groupID, buttonSkinComponent.fontSizeDefault, false, "Use default Font Size", helpText: ButtonSkinComponentUIHelp.useDefaultFontSize);
                        ForceUIRefreshOnValChange(buttonSkinComponent.fontSizeDefault, LeftRight.left, 4);
                    }
                    else if (groupID == 4 && !isRight)
                    {
                        if (!buttonSkinComponent.fontSizeDefault.val)
                        {
                            CreateSlider(groupID, buttonSkinComponent.fontSize.minFontSize, false, "Min Font Size", helpText: ButtonSkinComponentUIHelp.fontSizeMin);
                            CreateSlider(groupID, buttonSkinComponent.fontSize.maxFontSize, false, "Max Font Size", helpText: ButtonSkinComponentUIHelp.fontSizeMax);
                        }
                        CreateToggle(groupID, buttonSkinComponent.buttonFontDefault, false, "Use default Font", helpText: ButtonSkinComponentUIHelp.useDefaultFont);
                        ForceUIRefreshOnValChange(buttonSkinComponent.buttonFontDefault, LeftRight.left, 5);
                    }
                    else if (groupID == 5 && !isRight)
                    {
                        if (!buttonSkinComponent.buttonFontDefault.val)
                        {
                            CreatePopup(groupID, buttonSkinComponent.buttonFont, false, 200f, null, "Button Text Font", false, helpText: ButtonSkinComponentUIHelp.textFont);
                        }
                        CreateToggle(groupID, buttonSkinComponent.textAlignmentDefault, false, "Use default Text Alignment", helpText: ButtonSkinComponentUIHelp.useDefaultTextAlignment);
                        ForceUIRefreshOnValChange(buttonSkinComponent.textAlignmentDefault, LeftRight.left, 6);
                    }
                    else if (groupID == 6 && !isRight)
                    {
                        if (!buttonSkinComponent.textAlignmentDefault.val)
                        {
                            CreatePopup(groupID, buttonSkinComponent.textAlignment, false, 200f, null, "Button Text Alignment", false, helpText: ButtonSkinComponentUIHelp.textAlignment);
                        }
                        float spacerHeight = 0f;
                        if (!buttonSkinComponent.buttonFontDefault.val) spacerHeight = 200f;
                        if (!buttonSkinComponent.textAlignmentDefault.val) spacerHeight = 550f;
                        CreateSpacer(groupID, spacerHeight);
                    }

                    break;
                case colorOnTabEV:
                case colorOffTabEV:
                    ButtonSkinToggleParam colorParam = buttonSkinComponent.buttonSkinOnParams;
                    string labelTextInsert = "";
                    string useDefaultButtonColorHelp = ButtonSkinComponentUIHelp.useDefaultButtonColor;
                    string buttonColorHelp = ButtonSkinComponentUIHelp.buttonColor;
                    string useDefaultTextColorHelp = ButtonSkinComponentUIHelp.useDefaultTextColor;
                    string textColorHelp = ButtonSkinComponentUIHelp.textColor;
                    if (currentUIAButton.maxButtonStates > 1)
                    {
                        if (currentSkinTab == colorOffTabEV)
                        {
                            labelTextInsert = "Off ";
                            useDefaultButtonColorHelp = ButtonSkinComponentUIHelp.useDefaultButtonOffColor;
                            buttonColorHelp = ButtonSkinComponentUIHelp.buttonOffColor;
                            useDefaultTextColorHelp = ButtonSkinComponentUIHelp.useDefaultTextOffColor;
                            textColorHelp = ButtonSkinComponentUIHelp.textOffColor;
                        }
                        else
                        {
                            labelTextInsert = "On ";
                            useDefaultButtonColorHelp = ButtonSkinComponentUIHelp.useDefaultButtonOnColor;
                            buttonColorHelp = ButtonSkinComponentUIHelp.buttonOnColor;
                            useDefaultTextColorHelp = ButtonSkinComponentUIHelp.useDefaultTextOnColor;
                            textColorHelp = ButtonSkinComponentUIHelp.textOnColor;
                        }
                    }

                    if (currentSkinTab == colorOffTabEV) colorParam = buttonSkinComponent.buttonSkinOffParams;

                    if (groupID == 2 && !isRight)
                    {
                        CreateToggle(groupID, colorParam.buttonColorUseDefault, false, "Use default Button Color", helpText: useDefaultButtonColorHelp);
                        ForceUIRefreshOnValChange(colorParam.buttonColorUseDefault, LeftRight.left, 3);
                    }
                    else if (groupID == 3 && !isRight)
                    {
                        if (!colorParam.buttonColorUseDefault.val)
                        {
                            CreateColorPicker(groupID, colorParam.buttonColor, false, "Button " + labelTextInsert + "Color", helpText: buttonColorHelp);
                        }
                        CreateToggle(groupID, colorParam.textColorUseDefault, false, "Use default Text Color", helpText: useDefaultTextColorHelp);
                        ForceUIRefreshOnValChange(colorParam.textColorUseDefault, LeftRight.left, 4);
                    }
                    else if (groupID == 4 && !isRight)
                    {
                        if (!colorParam.textColorUseDefault.val)
                        {
                            CreateColorPicker(groupID, colorParam.textColor, false, "Text " + labelTextInsert + "Color", helpText: textColorHelp);
                        }
                    }
                    break;
                case thumbnailTabEV:
                    labelTextInsert = "";
                    if (currentUIAButton.maxButtonStates > 1) labelTextInsert = "Button On ";
                    if (groupID == 2 && !isRight)
                    {

                        if (currentUIAButton.GetAllThumbnailFileReferences().Count > 0)
                        {
                            CreateToggle(groupID, buttonSkinComponent.usePresetButtonTexture, false, "Use Preset Thumbnail", helpText: ButtonSkinComponentUIHelp.usePresetThumbnail);
                            ForceUIRefreshOnValChange(buttonSkinComponent.usePresetButtonTexture, LeftRight.left, 3);
                        }
                    }
                    else if (groupID == 3 && !isRight)
                    {
                        if (currentUIAButton.GetAllThumbnailFileReferences().Count == 0 || !buttonSkinComponent.usePresetButtonTexture.val)
                        {
                            CreateToggle(groupID, buttonSkinComponent.buttonSkinOnParams.customTextureURLUseDefault, false, "Use Default " + labelTextInsert + "Thumbnail", helpText: ButtonSkinComponentUIHelp.useDefaultThumbnailOn);
                            ForceUIRefreshOnValChange(buttonSkinComponent.buttonSkinOnParams.customTextureURLUseDefault, LeftRight.left, 4);
                        }
                    }
                    else if (groupID == 4 && !isRight)
                    {
                        if (currentUIAButton.GetAllThumbnailFileReferences().Count == 0 || !buttonSkinComponent.usePresetButtonTexture.val)
                        {
                            if (!buttonSkinComponent.buttonSkinOnParams.customTextureURLUseDefault.val)
                            {
                                CreateTextField(groupID, buttonSkinComponent.buttonSkinOnParams.customThumbnailURL, helpText: ButtonSkinComponentUIHelp.onThumbnailURL);
                                CreateButton(groupID, "Select " + labelTextInsert + "Thumbnail Image", false, SelectButtonOnThumbnail, helpText: ButtonSkinComponentUIHelp.onSelectThumbnail);
                            }
                            if (currentUIAButton.maxButtonStates > 1)
                            {
                                labelTextInsert = "Button Off ";
                                CreateToggle(groupID, buttonSkinComponent.buttonSkinOffParams.customTextureURLUseDefault, false, "Use Default " + labelTextInsert + "Thumbnail", helpText: ButtonSkinComponentUIHelp.useDefaultThumbnailOff);
                                ForceUIRefreshOnValChange(buttonSkinComponent.buttonSkinOffParams.customTextureURLUseDefault, LeftRight.left, 5);
                            }

                        }
                    }
                    else if (groupID == 5 && !isRight)
                    {
                        labelTextInsert = "Button Off ";
                        if ((currentUIAButton.GetAllThumbnailFileReferences().Count == 0 || !buttonSkinComponent.usePresetButtonTexture.val) && !buttonSkinComponent.buttonSkinOffParams.customTextureURLUseDefault.val && currentUIAButton.maxButtonStates > 1)
                        {
                            CreateTextField(groupID, buttonSkinComponent.buttonSkinOffParams.customThumbnailURL, helpText: ButtonSkinComponentUIHelp.offThumbnailURL);
                            CreateButton(groupID, "Select " + labelTextInsert + "Thumbnail Image", false, SelectButtonOffThumbnail, helpText: ButtonSkinComponentUIHelp.offSelectThumbnail);
                        }
                    }
                    break;
            }

        }

        private void TabSelected(int tab)
        {
            currentSkinTab = tab;
            RefreshGroupUI(LeftRight.both, 2);
        }

        protected void SelectButtonOnThumbnail()
        {
            string folder = "Saves";
            if (lastLoadedThumbnailFolder != "") folder = lastLoadedThumbnailFolder;
            SuperController.singleton.GetMediaPathDialog(OnThumbnailSelected, "jpg|png", folder, true, true, false);
        }

        protected void OnThumbnailSelected(string fileName)
        {
            if (fileName != "")
            {
                lastLoadedThumbnailFolder = FileManagerSecure.GetDirectoryName(fileName);
                buttonSkinComponent.buttonSkinOnParams.customThumbnailURL.val = fileName;
            }
        }

        protected void SelectButtonOffThumbnail()
        {
            string folder = "Saves";
            if (lastLoadedThumbnailFolder != "") folder = lastLoadedThumbnailFolder;
            SuperController.singleton.GetMediaPathDialog(OffThumbnailSelected, "jpg|png", folder, true, true, false);
        }

        protected void OffThumbnailSelected(string fileName)
        {
            if (fileName != "")
            {
                lastLoadedThumbnailFolder = FileManagerSecure.GetDirectoryName(fileName);
                buttonSkinComponent.buttonSkinOffParams.customThumbnailURL.val = fileName;
            }
        }
    }

    public class ButtonDefaultSkinComponentUITab : ButtonSkinComponentUITab
    {
        protected override ButtonSkinComponent buttonSkinComponent
        {
            get
            {
                return UIAStorables.defaultButtonSkin;
            }
        }

        CustomRowUIElement skinUIOptionsTabSelector;

        static string lastLoadedThumbnailFolder = "";

        const int textTabEV = 1;
        const int colorOnTabEV = 2;
        const int colorOffTabEV = 3;
        const int thumbnailTabEV = 4;

        private int currentSkinTab = textTabEV;

        public ButtonDefaultSkinComponentUITab() : base()
        {
            uiGroupCount = 2;

            defaultHelpText = ButtonDisplaySettingsUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            skinUIOptionsTabSelector = new CustomRowUIElement();
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight)
            {
                if (groupID == 1)
                {
                    skinUIOptionsTabSelector.InitUI(4);

                    UIDynamicButton button = skinUIOptionsTabSelector.AddButtonToRow("Default Text", textTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, ButtonDisplaySettingsUIHelp.defaultTextTab);

                    button = skinUIOptionsTabSelector.AddButtonToRow("Default On Colors", colorOnTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, ButtonDisplaySettingsUIHelp.defaultOnColorsTab);

                    button = skinUIOptionsTabSelector.AddButtonToRow("Default Off Colors", colorOffTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, ButtonDisplaySettingsUIHelp.defaultOffColorsTab);

                    button = skinUIOptionsTabSelector.AddButtonToRow("Default Thumbnail", thumbnailTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, ButtonDisplaySettingsUIHelp.defaultThumbnailTab);

                    RegisterCustomUIElement(groupID, skinUIOptionsTabSelector, isRight);
                    skinUIOptionsTabSelector.SetButtonColorActive(currentSkinTab);
                }
                CreateHelpTextUI(groupID, 970);
            }

            switch (currentSkinTab)
            {
                case textTabEV:
                    if (groupID == 2 && !isRight)
                    {
                        CreatePopup(groupID, buttonSkinComponent.buttonFont, false, 200f, null, "Default Text Font", false, ButtonDisplaySettingsUIHelp.defaultTextFont);
                        CreatePopup(groupID, buttonSkinComponent.textAlignment, false, 200f, null, "Default Text Alignment", false, ButtonDisplaySettingsUIHelp.defaultTextAlignment, 200f);
                        CreateSpacer(groupID, 5);
                        CreateSlider(groupID, buttonSkinComponent.defaultMicroFontSize.minFontSize, false, "Micro Button\nDefault Font Size (Min)", ButtonDisplaySettingsUIHelp.microButtonFontSizeMin);
                        CreateSlider(groupID, buttonSkinComponent.defaultMicroFontSize.maxFontSize, false, "Micro Button\nDefault Font Size (Max)", ButtonDisplaySettingsUIHelp.microButtonFontSizeMax);
                        CreateSpacer(groupID, 5);
                        CreateSlider(groupID, buttonSkinComponent.defaultMiniFontSize.minFontSize, false, "Mini Button\nDefault Font Size (Min)", ButtonDisplaySettingsUIHelp.miniButtonFontSizeMin);
                        CreateSlider(groupID, buttonSkinComponent.defaultMiniFontSize.maxFontSize, false, "Mini Button\nDefault Font Size (Max)", ButtonDisplaySettingsUIHelp.miniButtonFontSizeMax);
                        CreateSpacer(groupID, 5);
                        CreateSlider(groupID, buttonSkinComponent.defaultSmallFontSize.minFontSize, false, "Small Button\nDefault Font Size (Min)", ButtonDisplaySettingsUIHelp.smallButtonFontSizeMin);
                        CreateSlider(groupID, buttonSkinComponent.defaultSmallFontSize.maxFontSize, false, "Small Button\nDefault Font Size (Max)", ButtonDisplaySettingsUIHelp.smallButtonFontSizeMax);
                        CreateSpacer(groupID, 5);
                        CreateSlider(groupID, buttonSkinComponent.defaultMediumFontSize.minFontSize, false, "Medium Button\nDefault Font Size (Min)", ButtonDisplaySettingsUIHelp.mediumButtonFontSizeMin);
                        CreateSlider(groupID, buttonSkinComponent.defaultMediumFontSize.maxFontSize, false, "Medium Button\nDefault Font Size (Max)", ButtonDisplaySettingsUIHelp.mediumButtonFontSizeMax);
                        CreateSpacer(groupID, 5);
                        CreateSlider(groupID, buttonSkinComponent.defaultLargeFontSize.minFontSize, false, "Large Button\nDefault Font Size (Min)", ButtonDisplaySettingsUIHelp.largeButtonFontSizeMin);
                        CreateSlider(groupID, buttonSkinComponent.defaultLargeFontSize.maxFontSize, false, "Large Button\nDefault Font Size (Max)", ButtonDisplaySettingsUIHelp.largeButtonFontSizeMax);
                    }

                    break;
                case colorOnTabEV:
                case colorOffTabEV:
                    ButtonSkinToggleParam colorParam = buttonSkinComponent.buttonSkinOnParams;
                    if (currentSkinTab == colorOffTabEV) colorParam = buttonSkinComponent.buttonSkinOffParams;

                    string labelTextInsert = "On ";
                    string buttonColorHelpText = ButtonDisplaySettingsUIHelp.buttonOnColor;
                    string buttonTextHelpText = ButtonDisplaySettingsUIHelp.textOnColor;
                    if (currentSkinTab == colorOffTabEV)
                    {
                        labelTextInsert = "Off ";
                        buttonColorHelpText = ButtonDisplaySettingsUIHelp.buttonOffColor;
                        buttonTextHelpText = ButtonDisplaySettingsUIHelp.textOffColor;
                    }

                    if (groupID == 2 && !isRight)
                    {
                        CreateColorPicker(groupID, colorParam.buttonColor, false, "Button " + labelTextInsert + "Default Color", buttonColorHelpText);
                        CreateColorPicker(groupID, colorParam.textColor, false, "Text " + labelTextInsert + "Default Color", buttonTextHelpText);

                    }
                    break;
                case thumbnailTabEV:
                    if (groupID == 2 && !isRight)
                    {

                        CreateTextField(groupID, buttonSkinComponent.buttonSkinOnParams.customThumbnailURL, helpText: ButtonDisplaySettingsUIHelp.onThumbnailURL);
                        CreateButton(groupID, "Select Button On\nDefault Thumbnail Image...", false, SelectButtonOnThumbnail, 80f, helpText: ButtonDisplaySettingsUIHelp.onSelectThumbnail);
                        CreateButton(groupID, "Clear Button On Thumbnail", false, ClearButtonOnThumbnail, helpText: ButtonDisplaySettingsUIHelp.onClearThumnail);

                        CreateSpacer(groupID, 10f);

                        CreateTextField(groupID, buttonSkinComponent.buttonSkinOffParams.customThumbnailURL, helpText: ButtonDisplaySettingsUIHelp.offThumbnailURL);
                        CreateButton(groupID, "Select Button Off\nDefault Thumbnail Image...", false, SelectButtonOffThumbnail, 80f, helpText: ButtonDisplaySettingsUIHelp.offSelectThumbnail);
                        CreateButton(groupID, "Clear Button Off Thumbnail", false, ClearButtonOffThumbnail, helpText: ButtonDisplaySettingsUIHelp.offClearThumnail);
                    }
                    break;
            }

        }

        private void ClearButtonOffThumbnail()
        {
            buttonSkinComponent.buttonSkinOffParams.customThumbnailURL.val = "";
        }

        private void ClearButtonOnThumbnail()
        {
            buttonSkinComponent.buttonSkinOnParams.customThumbnailURL.val = "";
        }

        private void TabSelected(int tab)
        {
            currentSkinTab = tab;
            RefreshGroupUI(LeftRight.both, 2);
        }
    }

    public class MergeClothingComponentUITab : BaseUIButtonTreeTab
    {
        ClothingComponent clothingButtonComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.clothingComponent; } }

        public MergeClothingComponentUITab() : base()
        {
            uiGroupCount = 1;
            defaultHelpText = MergeClothingUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpAndButtonTree(groupID, 945);
            else if (groupID == 1 && !isRight)
            {
                foreach (JSONStorableBool jsb in clothingButtonComponent.GetRegionTagBools())
                {
                    string tagName = jsb.name.Substring(0, jsb.name.Length - 3);
                    CreateToggle(groupID, jsb, false, "Remove " + tagName + " Clothing");
                }
                CreateSpacer(groupID);
                foreach (JSONStorableBool jsb in clothingButtonComponent.GetTypeTagBools())
                {
                    string tagName = jsb.name.Substring(0, jsb.name.Length - 3);
                    CreateToggle(groupID, jsb, false, "Remove " + tagName + " Clothing");
                }
            }

        }

    }
    public class NodePhysicsComponentUITab : BaseUIButtonTreeTab
    {
        private NodePhysicsComponent nodePhysicsButtonComponent { get {
                if (isOffToggle) return ButtonSetupPluginUI.currentUIAButtonOperation.nodePhysicsOffComponent;
                return ButtonSetupPluginUI.currentUIAButtonOperation.nodePhysicsOnComponent;
            } }

        private bool isOffToggle = false;

        private CustomRowUIElement nodePhysicsTypeTabSelector;

        int currentPhysicsTypeTab = NodePhysicsType.hold;
        public NodePhysicsComponentUITab(bool _isOffToggle) : base()
        {
            isOffToggle = _isOffToggle;
            uiGroupCount = 4;
            defaultHelpText = NodePhysicsComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            nodePhysicsTypeTabSelector = new CustomRowUIElement();
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            var targetComponent = nodePhysicsButtonComponent.parentButtonOperation.targetComponent;
            if (isRight)
            {
                if (groupID == 1)
                {
                    if (targetComponent.isPersonTargetType) nodePhysicsTypeTabSelector.InitUI(5);
                    else nodePhysicsTypeTabSelector.InitUI(3);
                    var button = nodePhysicsTypeTabSelector.AddButtonToRow("Hold", NodePhysicsType.hold, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodePhysicsComponentUIHelp.hold);
                    button = nodePhysicsTypeTabSelector.AddButtonToRow("Link", NodePhysicsType.link, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodePhysicsComponentUIHelp.link);
                    button = nodePhysicsTypeTabSelector.AddButtonToRow("Comply", NodePhysicsType.compliance, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodePhysicsComponentUIHelp.comply);

                    if (targetComponent.isPersonTargetType)
                    {
                        button = nodePhysicsTypeTabSelector.AddButtonToRow("Joint Drive", NodePhysicsType.jointDrive, TabSelected);
                        RegisterContextualHelp(button.gameObject, NodePhysicsComponentUIHelp.jointDrive);
                       
                    }
                    button = nodePhysicsTypeTabSelector.AddButtonToRow("Max Velocity", NodePhysicsType.maxVelocity, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodePhysicsComponentUIHelp.maxVelocity);

                    RegisterCustomUIElement(groupID, nodePhysicsTypeTabSelector, true);

                }
                CreateHelpAndButtonTree(groupID, 880);
            }
            else if (groupID == 1 && !isRight)
            {
                if (currentPhysicsTypeTab == NodePhysicsType.hold || currentPhysicsTypeTab == NodePhysicsType.link)
                {
                    var nodePhysics = nodePhysicsButtonComponent.holdNodePhysics;
                    if (currentPhysicsTypeTab == NodePhysicsType.link) nodePhysics = nodePhysicsButtonComponent.linkNodePhysics;
                    CreatePositionRotationSDFUI(groupID, nodePhysics);
                }
                if (currentPhysicsTypeTab == NodePhysicsType.compliance) CreateCompliancePhysicsUI(groupID, nodePhysicsButtonComponent.nodePhysicsCompliance);
                if (currentPhysicsTypeTab == NodePhysicsType.jointDrive)
                {
                    CreatePopup(groupID, nodePhysicsButtonComponent.jointDriveSDFPhysics.modifyJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText:NodePhysicsComponentUIHelp.modifyJointDrive);
                    ForceUIRefreshOnValChange(nodePhysicsButtonComponent.jointDriveSDFPhysics.modifyJSEnum, LeftRight.left, 1);
                    if (nodePhysicsButtonComponent.jointDriveSDFPhysics.modifyJSEnum.val == NodePhysicsModificationMode.customValue) CreateSDFUI(groupID, nodePhysicsButtonComponent.jointDriveSDFPhysics);
                    CreateSpacer(groupID, 5);
                    CreateXYZUI(groupID, nodePhysicsButtonComponent.jointDriveXYZPhysics);
                    CreateSpacer(groupID, 150);
                }
                if (currentPhysicsTypeTab == NodePhysicsType.maxVelocity)
                {
                    CreatePopup(groupID, nodePhysicsButtonComponent.modifyMaxVelocityJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyMaxVelocity);
                    ForceUIRefreshOnValChange(nodePhysicsButtonComponent.modifyMaxVelocityJSEnum, LeftRight.left, 1);
                    if (nodePhysicsButtonComponent.modifyMaxVelocityJSEnum.val == NodePhysicsModificationMode.customValue)
                    {
                        CreateToggle(groupID, nodePhysicsButtonComponent.maxVelocityEnabledJSB, helpText: NodePhysicsComponentUIHelp.maxVelocityEnabled);
                        CreateSlider(groupID, nodePhysicsButtonComponent.maxVelocityJSF, helpText: NodePhysicsComponentUIHelp.maxVelocityValue);
                    }
                }

            }
            else if (groupID == 2 && !isRight)
            {
                if (!targetComponent.isPersonTargetType && (currentPhysicsTypeTab == NodePhysicsType.jointDrive || currentPhysicsTypeTab == NodePhysicsType.maxVelocity)) currentPhysicsTypeTab = NodePhysicsType.hold;
                nodePhysicsTypeTabSelector.SetButtonColorActive(currentPhysicsTypeTab);

            }

        }

        private void CreateXYZUI(int groupID, NodePhysicsXYZAngleElement nodePhysicsXYZ)
        {
            CreatePopup(groupID, nodePhysicsXYZ.modifyXJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 230, helpText: NodePhysicsComponentUIHelp.modifyXJointDrive);
            ForceUIRefreshOnValChange(nodePhysicsXYZ.modifyXJSEnum, LeftRight.left, 1);
            if (nodePhysicsXYZ.modifyXJSEnum.val == NodePhysicsModificationMode.customValue) CreateSlider(groupID, nodePhysicsXYZ.xAngleJSF, helpText: NodePhysicsComponentUIHelp.jointDriveX);

            CreatePopup(groupID, nodePhysicsXYZ.modifyYJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 230, helpText: NodePhysicsComponentUIHelp.modifyYJointDrive);
            ForceUIRefreshOnValChange(nodePhysicsXYZ.modifyYJSEnum, LeftRight.left, 1);
            if (nodePhysicsXYZ.modifyYJSEnum.val == NodePhysicsModificationMode.customValue) CreateSlider(groupID, nodePhysicsXYZ.yAngleJSF, helpText: NodePhysicsComponentUIHelp.jointDriveY);

            CreatePopup(groupID, nodePhysicsXYZ.modifyZJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 230, helpText: NodePhysicsComponentUIHelp.modifyZJointDrive);
            ForceUIRefreshOnValChange(nodePhysicsXYZ.modifyZJSEnum, LeftRight.left, 1);
            if (nodePhysicsXYZ.modifyZJSEnum.val == NodePhysicsModificationMode.customValue) CreateSlider(groupID, nodePhysicsXYZ.zAngleJSF, helpText: NodePhysicsComponentUIHelp.jointDriveZ);
        }
        private void CreateSDTUI(int groupID, NodePhysicsThesholdSprintDamperElement nodePhysicsTSD)
        {
            var slider = CreateSlider(groupID, nodePhysicsTSD.thresholdJSF, helpText: NodePhysicsComponentUIHelp.threshold);
            if (nodePhysicsTSD.thresholdJSF.name.StartsWith("complyPosition")) slider.valueFormat = "F4";
            else slider.valueFormat = "F1";
            var slider2 = CreateSlider(groupID, nodePhysicsTSD.springJSF, helpText: NodePhysicsComponentUIHelp.spring);
            slider2.valueFormat = "F0";
            CreateSlider(groupID, nodePhysicsTSD.damperJSF, helpText: NodePhysicsComponentUIHelp.damper);

        }
        private void CreateCompliancePhysicsUI(int groupID, NodePhysicsCompliance nodePhysics)
        {
            CreatePopup(groupID, nodePhysics.positionComplianceNodePhysics.modifyJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyPositionCompliance);
            ForceUIRefreshOnValChange(nodePhysics.positionComplianceNodePhysics.modifyJSEnum, LeftRight.left, 1);
            if (nodePhysics.positionComplianceNodePhysics.modifyJSEnum.val == NodePhysicsModificationMode.customValue) CreateSDTUI(groupID, nodePhysics.positionComplianceNodePhysics);

            CreateSpacer(groupID, 5);
            CreatePopup(groupID, nodePhysics.rotationComplianceNodePhysics.modifyJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyRotationCompliance);
            ForceUIRefreshOnValChange(nodePhysics.rotationComplianceNodePhysics.modifyJSEnum, LeftRight.left, 1);
            if (nodePhysics.rotationComplianceNodePhysics.modifyJSEnum.val == NodePhysicsModificationMode.customValue) CreateSDTUI(groupID, nodePhysics.rotationComplianceNodePhysics);

            CreateSpacer(groupID, 5);
            CreatePopup(groupID, nodePhysics.modifyComplianceSpeedJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyComplianceSpeed);
            ForceUIRefreshOnValChange(nodePhysics.modifyComplianceSpeedJSEnum, LeftRight.left, 1);
            if (nodePhysics.modifyComplianceSpeedJSEnum.val == NodePhysicsModificationMode.customValue)
            {
                var slider = CreateSlider(groupID, nodePhysics.complianceSpeedJSF, helpText: NodePhysicsComponentUIHelp.complianceSpeed);
                slider.valueFormat = "F1";
            }

            var targetComponent = nodePhysicsButtonComponent.parentButtonOperation.targetComponent;

            if (targetComponent.isPersonTargetType)
            {
                CreateSpacer(groupID, 5);
                CreatePopup(groupID, nodePhysics.modifyJointDriveSpringJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyJointDriveSpring);
                ForceUIRefreshOnValChange(nodePhysics.modifyJointDriveSpringJSEnum, LeftRight.left, 1);
                if (nodePhysics.modifyJointDriveSpringJSEnum.val == NodePhysicsModificationMode.customValue)
                {
                    var slider = CreateSlider(groupID, nodePhysics.jointDriveSpringJSF, helpText: NodePhysicsComponentUIHelp.jointDriveSpring);
                    slider.valueFormat = "F1";
                }
            }
            CreateSpacer(groupID, 150);
        }
        private void CreatePositionRotationSDFUI(int groupID, NodePhysicsPositionRotationSDF nodePhysics)
        {
            CreatePopup(groupID, nodePhysics.positionSpringDamperForce.modifyJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyPositionSDF);
            ForceUIRefreshOnValChange(nodePhysics.positionSpringDamperForce.modifyJSEnum, LeftRight.left, 1);
            if (nodePhysics.positionSpringDamperForce.modifyJSEnum.val == NodePhysicsModificationMode.customValue) CreateSDFUI(groupID, nodePhysics.positionSpringDamperForce);

            CreateSpacer(groupID, 5);
            CreatePopup(groupID, nodePhysics.rotationSpringDamperForce.modifyJSEnum, scrollable: false, popupHeight: 70f, labelWidth: 200, helpText: NodePhysicsComponentUIHelp.modifyRotationSDF);
            ForceUIRefreshOnValChange(nodePhysics.rotationSpringDamperForce.modifyJSEnum, LeftRight.left, 1);
            if (nodePhysics.rotationSpringDamperForce.modifyJSEnum.val == NodePhysicsModificationMode.customValue) CreateSDFUI(groupID, nodePhysics.rotationSpringDamperForce);

        }
        private void CreateSDFUI(int groupID, NodePhysicsSpringDamperForceElement nodePhysicsSDF)
        {
            bool isJointDrive = (nodePhysicsSDF.springJSF.name == "jointDriveSpring");
            var slider = CreateSlider(groupID, nodePhysicsSDF.springJSF, helpText: NodePhysicsComponentUIHelp.spring);
            if (!isJointDrive) slider.valueFormat = "F0";

            CreateSlider(groupID, nodePhysicsSDF.damperJSF, helpText: NodePhysicsComponentUIHelp.damper);

            var slider3 = CreateSlider(groupID, nodePhysicsSDF.maxForceJSF, helpText: NodePhysicsComponentUIHelp.maxForce);
            if (!isJointDrive) slider3.valueFormat = "F0";
        }
        private void TabSelected(int tab)
        {
            currentPhysicsTypeTab = tab;
            RefreshGroupUI(LeftRight.both, 1);
        }
    }

    public class NodeSelectionComponentUITab : BaseUIButtonTreeTab
    {
        private NodeSelectionComponent nodeSelectionComponent { get
            { return ButtonSetupPluginUI.currentUIAButtonOperation.nodeSelectionComponent;
            } }


        private CustomRowUIElement nodeSelectBodyRegionTabSelector;
        private CustomRowUIElement bulkSelectSettingButtons;

        private int currentBodyNodeControlTab = NodeBodyRegion.head;

        public NodeSelectionComponentUITab() : base()
        {

            uiGroupCount = 2;
            defaultHelpText = NodeSelectionComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            nodeSelectBodyRegionTabSelector = new CustomRowUIElement();
            bulkSelectSettingButtons = new CustomRowUIElement(80, false);
        }
        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight)
            {
                if (groupID == 1)
                {
                    nodeSelectBodyRegionTabSelector.InitUI(4);
                    var button = nodeSelectBodyRegionTabSelector.AddButtonToRow("Head Nodes", NodeBodyRegion.head, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeSelectionComponentUIHelp.head);
                    button = nodeSelectBodyRegionTabSelector.AddButtonToRow("Torso Nodes", NodeBodyRegion.torso, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeSelectionComponentUIHelp.torso);
                    button = nodeSelectBodyRegionTabSelector.AddButtonToRow("Limb Nodes", NodeBodyRegion.limbs, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeSelectionComponentUIHelp.limbs);
                    button = nodeSelectBodyRegionTabSelector.AddButtonToRow("Genital Nodes", NodeBodyRegion.gens, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeSelectionComponentUIHelp.genitals);
                    RegisterCustomUIElement(groupID, nodeSelectBodyRegionTabSelector, true);

                }
                CreateHelpAndButtonTree(groupID, 880);
            }
            else if (groupID == 1 && !isRight)
            {
                var nodeControlStates = nodeSelectionComponent.GetNodeControlSelections(currentBodyNodeControlTab);
                foreach (var nodeControl in nodeControlStates)
                {
                    if (nodeControl.name!="eyeTargetControl") CreateToggle(groupID, nodeControl, helpText: NodeSelectionComponentUIHelp.selectNode);
                }

                CreateSpacer(groupID, 10f);
                bulkSelectSettingButtons.InitUI(2, true, false, false, false);
                var button = bulkSelectSettingButtons.AddButtonToRow("Select All", 1, SetAllNodes);
                RegisterContextualHelp(button.gameObject, NodeSelectionComponentUIHelp.selectAll);

                button = bulkSelectSettingButtons.AddButtonToRow("Unselect All", 0, SetAllNodes);
                RegisterCustomUIElement(groupID, bulkSelectSettingButtons, false);
                RegisterContextualHelp(button.gameObject, NodeSelectionComponentUIHelp.DeSelectAll);

            }
            else if (groupID == 2 && !isRight)
            {
                nodeSelectBodyRegionTabSelector.SetButtonColorActive(currentBodyNodeControlTab);

            }
        }
        private void TabSelected(int tab)
        {
            currentBodyNodeControlTab = tab;
            RefreshGroupUI(LeftRight.both, 1);
        }

        private void SetAllNodes(int nodeControlState)
        {
            var nodeControlStates = nodeSelectionComponent.GetNodeControlSelections(currentBodyNodeControlTab);
            foreach (var nodeControl in nodeControlStates)
            {
                if (nodeControlState == 1) nodeControl.val = true;
                else nodeControl.val = false;
            }
        }
    }
    public class NodeControlComponentUITab : BaseUIButtonTreeTab
    {
        private NodeControlComponent nodeContolButtonComponent
        {
            get
            {
                if (isOffToggle) return ButtonSetupPluginUI.currentUIAButtonOperation.nodeControlOffComponent;
                return ButtonSetupPluginUI.currentUIAButtonOperation.nodeControlOnComponent;
            }
        }

        private bool isOffToggle = false;

        CustomRowUIElement nodeControlBodyRegionTabSelector;
        CustomRowUIElement bulkControlSettingButtons;

        int currentBodyNodeControlTab = NodeBodyRegion.head;

        public NodeControlComponentUITab(bool _isOffToggle) : base()
        {
            isOffToggle = _isOffToggle;

            uiGroupCount = 2;
            defaultHelpText = NodeControlComponentUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            nodeControlBodyRegionTabSelector = new CustomRowUIElement();
            bulkControlSettingButtons = new CustomRowUIElement(80, false);
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight)
            {
                if (groupID == 1)
                {
                    nodeControlBodyRegionTabSelector.InitUI(4);
                    var button = nodeControlBodyRegionTabSelector.AddButtonToRow("Head Nodes", NodeBodyRegion.head, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.head);
                    button = nodeControlBodyRegionTabSelector.AddButtonToRow("Torso Nodes", NodeBodyRegion.torso, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.torso);
                    button = nodeControlBodyRegionTabSelector.AddButtonToRow("Limb Nodes", NodeBodyRegion.limbs, TabSelected);
                     RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.limbs);
                    button = nodeControlBodyRegionTabSelector.AddButtonToRow("Genital Nodes", NodeBodyRegion.gens, TabSelected);
                    RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.genitals);
                    RegisterCustomUIElement(groupID, nodeControlBodyRegionTabSelector, true);
                }
                CreateHelpAndButtonTree(groupID, 880);
            }
            else if (groupID == 1 && !isRight)
            {
                var nodeControlStates = nodeContolButtonComponent.GetNodeControlStates(currentBodyNodeControlTab);
                foreach (var nodeControl in nodeControlStates)
                {
                    CreatePopup(groupID, nodeControl, scrollable: false, labelWidth: 250f, popupHeight: 50f, helpText: NodeControlComponentUIHelp.targetControlMode);
                }
                CreateSpacer(groupID, 10f);
                bulkControlSettingButtons.InitUI(4, true, false, false, false);
                var button = bulkControlSettingButtons.AddButtonToRow("All On", NodeControlState.on, SetAllNodes);
                RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.allOn);
                button = bulkControlSettingButtons.AddButtonToRow("All Comply", NodeControlState.comply, SetAllNodes);
                RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.allComply);
                button = bulkControlSettingButtons.AddButtonToRow("All Off", NodeControlState.off, SetAllNodes);
                RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.allOff);
                button = bulkControlSettingButtons.AddButtonToRow("All No Change", NodeControlState.noChange, SetAllNodes);
                RegisterContextualHelp(button.gameObject, NodeControlComponentUIHelp.allNoChange);

                RegisterCustomUIElement(groupID, bulkControlSettingButtons, false);
                CreateSpacer(groupID, 400f);
            }
            else if (groupID == 2 && !isRight)
            {
                nodeControlBodyRegionTabSelector.SetButtonColorActive(currentBodyNodeControlTab);

            }


        }

        private void SetAllNodes(int nodeControlState)
        {
            var nodeControlStates = nodeContolButtonComponent.GetNodeControlStates(currentBodyNodeControlTab);
            foreach (var nodeControl in nodeControlStates)
            {
                nodeControl.val = nodeControlState;
            }
        }

        private void TabSelected(int tab)
        {
            currentBodyNodeControlTab = tab;
            RefreshGroupUI(LeftRight.both, 1);
        }


    }



    public class DecalMakerComponentUITab : BaseUIButtonTreeTab
    {
        DecalMakerComponent decalMakerButtonComponent { get { return ButtonSetupPluginUI.currentUIAButtonOperation.decalMakerComponent; } }

        private CustomRowUIElement decalBodyRegionTabSelector;

        private JSONStorableStringChooser decalSelector;

        const int faceTabEV = 1;
        const int torsoTabEV = 2;
        const int limbsTabEV = 3;
        const int gensTabEV = 4;

        private int currentBodyDecalTab = faceTabEV;

        private DecalMakerTextureCollection _currentDMTC = null;
        private DecalMakerTextureCollection currentDMTC {get
                {
                    if (_currentDMTC==null) _currentDMTC = decalMakerButtonComponent.decalMakerTextureCollectionFace;
                    return _currentDMTC;
                }
            set
            {
                _currentDMTC = value;
            }
        }

        private string currentBodyRegionName = "Face";
        private DecalMakerTexture currentDMT = null;

        static string lastSelectedTextureFolder = "";

        public DecalMakerComponentUITab() : base()
        {
            uiGroupCount = 4;
            defaultHelpText = DecalMakerUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            decalBodyRegionTabSelector = new CustomRowUIElement();
            decalSelector = new JSONStorableStringChooser("decalSelector", null, "1", "",DecalSelected);

            
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight)
            {
                if (groupID == 1)
                {
                    decalBodyRegionTabSelector.InitUI(4);
                    var button = decalBodyRegionTabSelector.AddButtonToRow("Face Decals", faceTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, DecalMakerUIHelp.faceTab);
                    button = decalBodyRegionTabSelector.AddButtonToRow("Torso Decals", torsoTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, DecalMakerUIHelp.torsoTab);
                    button = decalBodyRegionTabSelector.AddButtonToRow("Limb Decals", limbsTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, DecalMakerUIHelp.limbsTab);
                    button = decalBodyRegionTabSelector.AddButtonToRow("Genital Decals", gensTabEV, TabSelected);
                    RegisterContextualHelp(button.gameObject, DecalMakerUIHelp.gensTab);
                    RegisterCustomUIElement(groupID, decalBodyRegionTabSelector,true);

                }
                CreateHelpAndButtonTree(groupID, 880);
            }
            else if (groupID == 1 && !isRight)
            {
                RefreshDMTC();
                RefreshDecalSelectorChoices();
            }
            else if (groupID == 2 && !isRight)
            {
                decalBodyRegionTabSelector.SetButtonColorActive(currentBodyDecalTab);
                if (currentDMTC.decalMakerTexturesList.Count > 0)
                {
                    if (currentDMTC.decalMakerTexturesList.Count > 1)
                    {
                        CreatePopup(groupID, decalSelector, false, 400f, RefreshDecalSelectorChoices, "Select Decal", decalSelector.choices.Count > 8, helpText: DecalMakerUIHelp.selectDecal);
                        ForceUIRefreshOnValChange(decalSelector, LeftRight.left, 3);
                    }                       
                    currentDMT = currentDMTC.decalMakerTexturesList[int.Parse(decalSelector.val) - 1];                

                }
            }
            else if (groupID == 3 && !isRight)
            {
                if (currentDMTC.decalMakerTexturesList.Count > 0)
                {
                    CreateTextField(groupID, currentDMT.decalTexturePathJSS, helpText: DecalMakerUIHelp.decalFilePath);
                    CreateButton(groupID, "Select " + currentBodyRegionName + " Decal " + decalSelector.val + " File", false, SelectDecalFile, helpText: DecalMakerUIHelp.selectDecalFile);
                    CreateColorPicker(groupID, currentDMT.decalTextureColorJSC, false, currentBodyRegionName + " Decal " + decalSelector.val + " Color", helpText: DecalMakerUIHelp.decalColor);
                    CreateSlider(groupID, currentDMT.decalTextureAlphaJSF, false, currentBodyRegionName + " Decal " + decalSelector.val + " Alpha", helpText: DecalMakerUIHelp.decalAlpha);
                }
            }
            else if (groupID == 4 && !isRight)
            {
                CreateButton(groupID, "Add " + currentBodyRegionName + " Decal", false, AddDecal, helpText: DecalMakerUIHelp.addDecal);
                if (currentDMTC.decalMakerTexturesList.Count > 0) CreateButton(groupID, "Remove " + currentBodyRegionName + " Decal " + decalSelector.val, false, RemoveDecal, helpText: DecalMakerUIHelp.deleteDecal);
            }

        }

        private void SelectDecalFile()
        {
            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Atom\\Person\\Textures", false, false, true);

            string folder = "Custom\\Atom\\Person\\Textures";
            if (currentDMT.decalTexturePathJSS.val != "") folder = FileManagerSecure.GetDirectoryName(currentDMT.decalTexturePathJSS.val);
            else if (lastSelectedTextureFolder != "") folder = lastSelectedTextureFolder;

             SuperController.singleton.GetMediaPathDialog(TextureSelected, "jpg|png", folder, false, true, false, null, false, shortcuts);
        }
        private void TextureSelected(string fileName)
        {
            if (fileName != "")
            {
                lastSelectedTextureFolder = FileManagerSecure.GetDirectoryName(fileName);
                currentDMT.decalTexturePathJSS.val = fileName;
            }
        }

        private void TabSelected(int tab)
        {
            currentBodyDecalTab = tab;
            RefreshDMTC();
            RefreshGroupUI(LeftRight.both, 1);
        }

        private void RefreshDMTC()
        {
            
            if (currentBodyDecalTab == faceTabEV)
            {
                currentDMTC = decalMakerButtonComponent.decalMakerTextureCollectionFace;
                currentBodyRegionName = "Face";
            }
            if (currentBodyDecalTab == torsoTabEV)
            {
                currentDMTC = decalMakerButtonComponent.decalMakerTextureCollectionTorso;
                currentBodyRegionName = "Torso";
            }
            if (currentBodyDecalTab == limbsTabEV)
            {
                currentDMTC = decalMakerButtonComponent.decalMakerTextureCollectionLimbs;
                currentBodyRegionName = "Limbs";
            }
            if (currentBodyDecalTab == gensTabEV)
            {
                currentBodyRegionName = "Genitals";
                currentDMTC = decalMakerButtonComponent.decalMakerTextureCollectionGen;
            }
            decalSelector.val = "1";
        }

        private void DecalSelected(string decalVal)
        {
            if (currentDMTC.decalMakerTexturesList.Count > 0) currentDMT = currentDMTC.decalMakerTexturesList[int.Parse(decalVal) - 1];
            else currentDMT = null;
        }

        private void AddDecal()
        {
            currentDMTC.AddDecal();
            decalSelector.val = currentDMTC.decalMakerTexturesList.Count.ToString();
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void RemoveDecal()
        {
            int index = int.Parse(decalSelector.val) - 1;
            currentDMTC.RemoveDecal(index);
            decalSelector.val = "1";
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void RefreshDecalSelectorChoices()
        {
            DecalMakerTextureCollection dmtc = null;
            if (currentBodyDecalTab == faceTabEV) dmtc = decalMakerButtonComponent.decalMakerTextureCollectionFace;
            if (currentBodyDecalTab == torsoTabEV) dmtc = decalMakerButtonComponent.decalMakerTextureCollectionTorso;
            if (currentBodyDecalTab == limbsTabEV) dmtc = decalMakerButtonComponent.decalMakerTextureCollectionLimbs;
            if (currentBodyDecalTab == gensTabEV) dmtc = decalMakerButtonComponent.decalMakerTextureCollectionGen;

            List<string> choices = new List<string>();
            int index = 1;
            foreach (DecalMakerTexture dmt in dmtc.decalMakerTexturesList)
            {
                choices.Add(index.ToString());
                index++;
            }
            decalSelector.choices = choices;
            if (choices.Count>0 && !decalSelector.choices.Contains(decalSelector.val)) decalSelector.val = decalSelector.choices[0];
        }
    }


    public class SettingsPluginUI : BaseUIHelpTab
    {

        private static CustomRowUIElement settingsTabSelectorUI;

        private static int currentSettingsTab;

        private static VRSettingsUITab vrSettingsUITab;
        private static GameControlSettingsUITab gridControlSettingsUITab;
        private static TargetGroupSettingsUITab targetGroupSettingsUITab;
        private static GazeTargettingSettingsUITab gazeTargettingSettingsUITab;
        private static PresetLoadingSettingsUITab presetLoadingSettingsUITab;
        private static AboutSettingsUITab aboutSettingsUITab;
        private static ButtonDefaultSkinComponentUITab defaultSkinUITab;

        public SettingsPluginUI() : base()
        {
            settingsTabSelectorUI = new CustomRowUIElement(80);
            uiGroupCount = 2;           

            currentHelpTextJSS.val = defaultHelpText;

            vrSettingsUITab = new VRSettingsUITab();
            gridControlSettingsUITab = new GameControlSettingsUITab();
            targetGroupSettingsUITab = new TargetGroupSettingsUITab();
            gazeTargettingSettingsUITab = new GazeTargettingSettingsUITab();
            presetLoadingSettingsUITab = new PresetLoadingSettingsUITab();
            aboutSettingsUITab = new AboutSettingsUITab();
            defaultSkinUITab = new ButtonDefaultSkinComponentUITab();

            currentSettingsTab = 7;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (groupID == 1 && !isRight)
            {
                settingsTabSelectorUI.InitUI(7);

                settingsTabSelectorUI.AddButtonToRow("VR", 1,SettingsTabSelected);
                settingsTabSelectorUI.AddButtonToRow("Game Control UI", 2, SettingsTabSelected);
                settingsTabSelectorUI.AddButtonToRow("Target Groups", 3, SettingsTabSelected);
                settingsTabSelectorUI.AddButtonToRow("Gaze Targetting", 4, SettingsTabSelected);
                settingsTabSelectorUI.AddButtonToRow("Preset Loading", 5, SettingsTabSelected);
                settingsTabSelectorUI.AddButtonToRow("Button Display", 6, SettingsTabSelected);
                settingsTabSelectorUI.AddButtonToRow("About", 7, SettingsTabSelected);

                RegisterCustomUIElement(groupID, settingsTabSelectorUI);
            }

            if (groupID == 2 && !isRight)
            {
                settingsTabSelectorUI.SetButtonColorActive(currentSettingsTab);
                switch (currentSettingsTab)
                {
                    case 1:
                        RegisterChildUITab(groupID, vrSettingsUITab, false);
                        vrSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 2:
                        RegisterChildUITab(groupID, gridControlSettingsUITab, false);
                        gridControlSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 3:
                        RegisterChildUITab(groupID, targetGroupSettingsUITab, false);
                        targetGroupSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 4:
                        RegisterChildUITab(groupID, gazeTargettingSettingsUITab, false);
                        gazeTargettingSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 5:
                        RegisterChildUITab(groupID, presetLoadingSettingsUITab, false);
                        presetLoadingSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 6:
                        RegisterChildUITab(groupID, defaultSkinUITab, false);
                        defaultSkinUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 7:
                        RegisterChildUITab(groupID, aboutSettingsUITab, false);
                        aboutSettingsUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                }
            }
        }
        private void SettingsTabSelected(int tab)
        {
            currentSettingsTab = tab;
            RefreshGroupUI(LeftRight.both, 2);
        }
    }

    public class VRSettingsUITab : BaseUIHelpTab
    {
        CustomRowUIElement vrSettingsTabSelector;
        const int handAttachTabEV = 1;
        const int gazeTabEV = 2;
        const int controllersTabEV = 3;

        private int currentVRSettingsTab = handAttachTabEV;
        public VRSettingsUITab() : base()
        {
            uiGroupCount = 4;
            defaultHelpText = VRSettingsUIHelp.handAttachDefaultText;
            currentHelpTextJSS.val = defaultHelpText;

            vrSettingsTabSelector = new CustomRowUIElement();
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (groupID == 1 && isRight)
            {
                vrSettingsTabSelector.InitUI(3);
                UIDynamicButton button = vrSettingsTabSelector.AddButtonToRow("Hand Attach/Detach", handAttachTabEV, TabSelected);

                button = vrSettingsTabSelector.AddButtonToRow("Gaze UI Control", gazeTabEV, TabSelected);

                button = vrSettingsTabSelector.AddButtonToRow("VR Controllers", controllersTabEV, TabSelected);

                RegisterCustomUIElement(groupID, vrSettingsTabSelector, isRight);
                 CreateHelpTextUI(groupID,970);
            }

            if (!isRight) vrSettingsTabSelector.SetButtonColorActive(currentVRSettingsTab);
            switch (currentVRSettingsTab)
            {
                case handAttachTabEV:
                    if (groupID == 2 && !isRight)
                    {
                        CreatePopup(groupID, VRSettings.vrHandControlJSEnum, false, 300, null, null, false, VRSettingsUIHelp.attachToVRHand);
                        ForceUIRefreshOnValChange(VRSettings.vrHandControlJSEnum, LeftRight.left, 3);
                    }
                    else if (groupID == 3 && !isRight)
                    {
                        if (VRSettings.vrHandControlJSEnum.val != VRHandControl.disabled && VRSettings.vrHandControlJSEnum.val != VRHandControl.vamUI)
                        {
                            CreateSlider(groupID, VRSettings.vrWristOpacityJSF, false, "Opacity on hand\n(when inactive)", VRSettingsUIHelp.opactityOnHand);
                            CreateToggle(groupID, VRSettings.hideWristUIWithShowVRHandsJSB, false, "Hide UI with 'Show VR Hands' hide", 50, VRSettingsUIHelp.hideWristsWithShowVRHands);
                        }
                        if(VRSettings.vrHandControlJSEnum.val == VRHandControl.vamUI)
                        {
                            CreatePopup(groupID, VRSettings.vrUIAttachPositionJSEnum, false, 300,null,"UIAssist Launch button position",false, VRSettingsUIHelp.vrUIAttachPosition,250);
                        }
                        CreateSpacer(groupID);

                        CreatePopup(groupID, VRSettings.vrPinModeJSEnum, false, 100, null, null, false, VRSettingsUIHelp.detachedUIMode);
                        ForceUIRefreshOnValChange(VRSettings.vrPinModeJSEnum, LeftRight.left, 4);
                    }
                    else if (groupID == 4 && !isRight)
                    {
                        if (VRSettings.vrPinModeJSEnum.val != VRPinMode.disabled)
                        {

                            if (VRSettings.vrPinModeJSEnum.val == VRPinMode.staticMode && VRSettings.vrHandControlJSEnum.val != VRHandControl.disabled)
                            {
                                CreateSlider(groupID, VRSettings.vrPinMinOffsetJSF, false, "Minimum Distance\n(from player)", VRSettingsUIHelp.detachedMinDistance);
                            }
                            if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudCustomMode)
                            {
                                CreateSlider(groupID, VRSettings.vrHUDHorizontalOffsetJSF, false, "HUD Horizontal position", VRSettingsUIHelp.hudHorizontalOffset);
                                CreateSlider(groupID, VRSettings.vrHUDVerticalOffsetJSF, false, "HUD Vertical position", VRSettingsUIHelp.hudHorizontalOffset);
                                CreateSlider(groupID, VRSettings.vrHUDDepthOffsetJSF, false, "HUD Distance", VRSettingsUIHelp.hudDepthOffset);
                            }

                            if (VRSettings.vrHandControlJSEnum.val != VRHandControl.disabled)
                            {
                                CreateToggle(groupID, VRSettings.autoPinOnGridOpenJSB, false, "Auto detach UI on Grid Open", 50, VRSettingsUIHelp.autoPinOnGridOpen);
                                CreateToggle(groupID, VRSettings.autoUnPinOnGridCloseJSB, false, "Auto attach UI on Grid Close", 50, VRSettingsUIHelp.autoUnPinOnGridClose);
                            }
                        }
                    }
                    break;
                case gazeTabEV:
                    if (groupID == 2 && !isRight)
                    {
                        CreatePopup(groupID, VRSettings.gazeMenuSelectModeJSEnum, false, 0, null, null, false, VRSettingsUIHelp.gazeMenuSelectMode);
                        if (VRSettings.gazeMenuSelectModeJSEnum.val != VRGazeUIControlMode.disabled)
                        {
                            CreateToggle(groupID, VRSettings.gazeActivateOpenVRGridJSB, false, "Gaze Open Grids", 50, VRSettingsUIHelp.gazeActivateOpenVRGrid);
                            CreateToggle(groupID, VRSettings.gazeActivateCloseVRGridJSB, false, "Gaze Close Grids", 50, VRSettingsUIHelp.gazeActivateCloseVRGrid);

                            CreateColorPicker(groupID, VRSettings.gazeSelectButtonColorJSC, false, "Gaze UI Button Highlight Color", VRSettingsUIHelp.gazeSelectButtonColor);

                            CreateSpacer(groupID, 10);
                        }
                    }
                    break;

                case controllersTabEV:
                    if (groupID == 2 && !isRight)
                    {
                        CreatePopup(groupID, VRSettings.LcontrollerDoubleSelectModeJSE, false, 400, null, null, false, VRSettingsUIHelp.leftControllerDoubleSelect, 250);
                        CreatePopup(groupID, VRSettings.RcontrollerDoubleSelectModeJSE, false, 400, null, null, false, VRSettingsUIHelp.rightControllerDoubleSelect, 250);
                    }
                    break;
            }
        }
        private void TabSelected(int tab)
        {
            currentVRSettingsTab = tab;
            if (tab== handAttachTabEV) defaultHelpText = VRSettingsUIHelp.handAttachDefaultText;
            else if (tab == gazeTabEV) defaultHelpText = VRSettingsUIHelp.gazeUIControlDefaultText;
            else if (tab == controllersTabEV) defaultHelpText = VRSettingsUIHelp.vrControllersDefaultText;

            currentHelpTextJSS.val = defaultHelpText;

            RefreshGroupUI(LeftRight.left,2);
        }

    }

    public class GameControlSettingsUITab : BaseUIHelpTab
    {
        public GameControlSettingsUITab():base()
        {
            uiGroupCount = 1;
            defaultHelpText = GameControlUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                CreateToggle(groupID, GameControlSettings.displayGazeSelectionsJSB,false,"Show Gaze Targets on UI",50,GameControlUIHelp.displayGazeSelections);
                CreateToggle(groupID, GameControlSettings.displayGridLabelsJSB, false, "Show Grid Labels on UI", 50, GameControlUIHelp.displayGridLabels);
                if (PatreonFeatures.patreonContentEnabled) CreateToggle(groupID, GameControlSettings.activateQuickLaunchBarJSB, false, "Enable Quick Launch Bar", 50, GameControlUIHelp.activateQuickLaunchBar);
                CreateToggle(groupID, GameControlSettings.i24hClockEnableJSB, false, "Enable 24h clock format", 50, GameControlUIHelp.i24hClockEnable);
                string hideGameControlUILabel = "Hide UIAssist Game Control UI";
                if (UIAGlobals.hideGameControlUI) hideGameControlUILabel = "Show UIAssist Game Control UI";
                CreateButton(groupID, hideGameControlUILabel, false, ToggleHideGameControlUI,50f);
                CreateSpacer(groupID);
                CreateSlider(groupID, GameControlSettings.desktopUIScaleJSF, false, "Desktop Scale", GameControlUIHelp.desktopUIScale);
                CreateSlider(groupID, GameControlSettings.desktopHOffsetJSF, false, "Desktop Horizontal Offset", GameControlUIHelp.desktopHOffset);
                CreateSlider(groupID, GameControlSettings.desktopVOffsetJSF, false, "Desktop Veritcial Offset", GameControlUIHelp.desktopVOffset);
            }
        }

        private void ToggleHideGameControlUI()
        {
            UIAGlobals.hideGameControlUI = !UIAGlobals.hideGameControlUI;
            RefreshGroupUI(LeftRight.both, 1);
        }

    }

    public class TargetGroupSettingsUITab : BaseUIHelpTab
    {
        public static JSONStorableStringChooser customTargetGroupSelectorJSSC;
        private static CustomTargetGroupType currentCTG;
        private CustomRowUIElement addDeleteCTGUI;
        private bool deleteConfirmMode = false;

        public TargetGroupSettingsUITab()
        {
            uiGroupCount = 4;
            defaultHelpText = CTGSettingsUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            customTargetGroupSelectorJSSC = new JSONStorableStringChooser("customTargetGroupSelector",null,"Select...","Custom Group",SelectedCustomTargetGroup);

            currentCTG = null;

            addDeleteCTGUI = new CustomRowUIElement(80, false);
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                
                CreatePopup(groupID, customTargetGroupSelectorJSSC, false, 600, UpdateCTGSelectorChoices, null, CustomTargetGroupSettings.customTargetGroupTypeList.Count() > 10, CTGSettingsUIHelp.ctgSelect);
                ForceUIRefreshOnValChange(customTargetGroupSelectorJSSC, LeftRight.left, 2);
            }
            else if (groupID == 2 && !isRight)
            {
                if (customTargetGroupSelectorJSSC.val != "Select...")
                {
                    CreateTextField(groupID, currentCTG.cgtNameJSS,false,100f,true, InputField.LineType.SingleLine,"Enter Custom Target Group name", CTGSettingsUIHelp.ctgName);
                    CreatePopup(groupID, currentCTG.cgtAtomCategoryJSEnum, false, 400f, null, "Atom Category Filter",false, CTGSettingsUIHelp.ctgAtomCat);
                    ForceUIRefreshOnValChange(currentCTG.cgtAtomCategoryJSEnum, LeftRight.left, 3);
                }

            }
            else if (groupID == 3 && !isRight)
            {
                deleteConfirmMode = false;
                if (customTargetGroupSelectorJSSC.val != "Select...")
                {
                    if (currentCTG.cgtAtomCategoryJSEnum.val != AtomCategories.people) CreatePopup(groupID, currentCTG.cgtAtomTypeJSEnum, false, 100f, null, "Atom Type Filter", false, CTGSettingsUIHelp.ctgAtomType);
                    else CreatePopup(groupID, currentCTG.cgtPersonGenderFilterJSEnum, false, 100f, null, "Person Gender Filter", false, CTGSettingsUIHelp.ctgPersonGender);
                    CreatePopup(groupID, currentCTG.cgtNameFilterTypeJSEnum, false, 100f, null, "Atom Name Filter", false, CTGSettingsUIHelp.ctgAtomNameFilterType);
                    ForceUIRefreshOnValChange(currentCTG.cgtNameFilterTypeJSEnum, LeftRight.left, 4);
                }
            }
            else if (groupID == 4 && !isRight)
            {
                if (customTargetGroupSelectorJSSC.val != "Select..." && currentCTG.cgtNameFilterTypeJSEnum.val!=CGTNameFilterTypes.none)
                {
                    CreateTextField(groupID, currentCTG.cgtNameFilterJSS, false, 100f, true, InputField.LineType.SingleLine, "Enter filter text for Atom name...", CTGSettingsUIHelp.ctgAtomName);
                }
                CreateSpacer(groupID, (1000f - GetTotalUIElementHeights(false) - 80f));

                addDeleteCTGUI.InitUI(2, true, false);

                if (!deleteConfirmMode)
                {
                    UIDynamicButton button = addDeleteCTGUI.AddButtonToRow("Add Custom\nTarget Group", 1, AddDeleteCTGButtonActivated);
                    RegisterContextualHelp(button.gameObject, CTGSettingsUIHelp.addCTG);
                    if (currentCTG != null && CustomTargetGroupSettings.customTargetGroupTypeList.Count > 0)
                    {
                        button = addDeleteCTGUI.AddButtonToRow("Delete Custom\nTarget Group...", 2, AddDeleteCTGButtonActivated);
                        button.buttonColor = Color.red;
                        button.textColor = Color.white;
                        RegisterContextualHelp(button.gameObject, CTGSettingsUIHelp.deleteCTG);
                    }
                }
                else
                {
                    UIDynamicButton button = addDeleteCTGUI.AddButtonToRow("Confirm Delete", 3, AddDeleteCTGButtonActivated);
                    button.buttonColor = Color.red;
                    button.textColor = Color.white;
                    RegisterContextualHelp(button.gameObject, CTGSettingsUIHelp.confirmDeleteCTG);
                    button = addDeleteCTGUI.AddButtonToRow("Cancel Delete", 4, AddDeleteCTGButtonActivated);
                    button.buttonColor = Color.green;
                    button.textColor = Color.black;
                    RegisterContextualHelp(button.gameObject, CTGSettingsUIHelp.cancelDeleteCTG);
                }
                
                RegisterCustomUIElement(groupID, addDeleteCTGUI);
            }
        }

        private void AddDeleteCTGButtonActivated(int buttonRef)
        {
            if (buttonRef==1)
            {
                string ctgName = CustomTargetGroupSettings.AddNewCTG();
                customTargetGroupSelectorJSSC.val = ctgName;
                RefreshGroupUI(LeftRight.left, 1);
            }
            else if (buttonRef == 3)
            {
                CustomTargetGroupSettings.DeleteCTG(customTargetGroupSelectorJSSC.val);
                deleteConfirmMode = false;
                if (CustomTargetGroupSettings.customTargetGroupTypeList.Count() > 0) customTargetGroupSelectorJSSC.val = CustomTargetGroupSettings.customTargetGroupTypeList.First().cgtNameJSS.val;
                else customTargetGroupSelectorJSSC.val = "Select...";
                RefreshGroupUI(LeftRight.left, 1);
            }
            else
            {
                if (buttonRef == 2) deleteConfirmMode = true;
                
                if (buttonRef == 4) deleteConfirmMode = false;
                RefreshGroupUI(LeftRight.left, 4);
            }
            
        }

        private void UpdateCTGSelectorChoices()
        {
            List<string> choices = CustomTargetGroupSettings.GetCurrentCTGNamesList();
            if (choices.Count == 0)
            {
                choices.Add("None Available");
            }

            customTargetGroupSelectorJSSC.choices = choices;
        }

        private void SelectedCustomTargetGroup(string cgtName)
        {
            if (cgtName =="None Available")
            {
                customTargetGroupSelectorJSSC.valNoCallback = "Select...";
                currentCTG = null;
            }
            else currentCTG = CustomTargetGroupSettings.GetCGTFromName(cgtName);
        }
    }

    public class GazeTargettingSettingsUITab : BaseUIHelpTab
    {
        public GazeTargettingSettingsUITab()
        {
            uiGroupCount = 1;
            defaultHelpText = GazeTargetSettingsUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                UIDynamicPopup popup = CreatePopup(groupID, GazeTargetSettings.defaultGazeTargetJSEnum, false, 100, null, null, false, GazeTargetSettingsUIHelp.defaultGazeTarget, 200);
                CreateToggle(groupID, GazeTargetSettings.gazeSelectOffAtomsJSB,false,"Enable 'Off' Atom Gaze Selection", 50,GazeTargetSettingsUIHelp.enableOffAtomSelection);

                CreateSpacer(groupID);
                CreateToggle(groupID, GazeAssistedSelectTool.gazeAssistActiveJSB, false, "Enable Gaze Assisted Atom Select",50, GazeTargetSettingsUIHelp.enableGazeAssistedAtomSelection);
            }

        }
    }

    public class PresetLoadingSettingsUITab : BaseUIHelpTab
    {

        public PresetLoadingSettingsUITab()
        {
            uiGroupCount = 1;
            defaultHelpText = PresetLoadSettingsUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {               
                CreateToggle(groupID, PresetLoadSettings.suppressPresetLocksJSB, false, "Suppress Preset Locks on all UIA Preset Loads", 80f, PresetLoadSettingsUIHelp.suppressPresetLocks);
                if (PatreonFeatures.patreonContentEnabled) CreateToggle(groupID, PresetLoadSettings.suppressScaleLoadJSB, false, "Suppress Scale Load on all UIA Appearance Preset Loads", 80f, PresetLoadSettingsUIHelp.suppressScale);
                CreateToggle(groupID, PresetLoadSettings.suppressClothingLoadJSB, false, "Suppress Clothing Load on all UIA Appearance Preset Loads", 80f, PresetLoadSettingsUIHelp.suppressClothing);

            }
        }
    }

    public class AboutSettingsUITab : BaseUIHelpTab
    {
        public AboutSettingsUITab()
        {
            uiGroupCount = 1;
            defaultHelpText = AboutUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            string pluginDetails;
            if (PatreonFeatures.patreonContentEnabled)
            {
                pluginDetails = UIAConsts.pluginName + " v" + UIAConsts.pluginVersion + "\nby " + UIAConsts.pluginAuthor + "\nPatron version";
            }
            else
            {
                pluginDetails = UIAConsts.pluginName + " v" + UIAConsts.pluginVersion + "\nby " + UIAConsts.pluginAuthor + "\nFree version";
            }
            _pluginDetailsJSON = new JSONStorableString("PluginDetails", pluginDetails);
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                CreateTextField(groupID, _pluginDetailsJSON, false, 200);
                if (UIAGlobals.isPrimaryPlugin) CreateTextField(groupID, new JSONStorableString("primary", "Primary"), false,100,helpText: AboutUIHelp.primarySecondaryState);
                else CreateTextField(groupID, new JSONStorableString("secondary", "Secondary"), false, 100, helpText: AboutUIHelp.primarySecondaryState);
            }
        }

        public JSONStorableString _pluginDetailsJSON;
    }

    public class ToolsPluginUI : BaseUIHelpTab
    {

        private static CustomRowUIElement toolsTabSelectorUI;

        private static int currentToolsTab;

        private static HeelAdjustToolUITab heelAdjustToolUITab;
        private static ForceVAMModesToolUITab forceVAMModesToolUITab;
        private static LeapMotionControlToolUITab leapMotionControlToolUITab;
        private static BulkGridSetupToolUITab bulkGridSetupToolUITab;

        public ToolsPluginUI() : base()
        {
            toolsTabSelectorUI = new CustomRowUIElement(80);
            uiGroupCount = 2;

            heelAdjustToolUITab = new HeelAdjustToolUITab();
            forceVAMModesToolUITab = new ForceVAMModesToolUITab();
            leapMotionControlToolUITab = new LeapMotionControlToolUITab();
            bulkGridSetupToolUITab = new BulkGridSetupToolUITab();

            currentToolsTab = 1;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (groupID == 1 && !isRight)
            {
                toolsTabSelectorUI.InitUI(4);

                toolsTabSelectorUI.AddButtonToRow("Heel Adjust", 1, SettingsTabSelected);
                toolsTabSelectorUI.AddButtonToRow("Force VAM Modes", 2, SettingsTabSelected);

                toolsTabSelectorUI.AddButtonToRow("Leap Motion\nControl", 3, SettingsTabSelected);
                toolsTabSelectorUI.AddButtonToRow("Bulk Grid Setup", 4, SettingsTabSelected);

                RegisterCustomUIElement(groupID, toolsTabSelectorUI);
            }

            if (groupID == 2 && !isRight)
            {
                toolsTabSelectorUI.SetButtonColorActive(currentToolsTab);
                switch (currentToolsTab)
                {
                    case 1:
                        RegisterChildUITab(groupID, heelAdjustToolUITab, false);
                        heelAdjustToolUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 2:
                        RegisterChildUITab(groupID, forceVAMModesToolUITab, false);
                        forceVAMModesToolUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 3:
                        RegisterChildUITab(groupID, leapMotionControlToolUITab, false);
                        leapMotionControlToolUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;
                    case 4:
                        RegisterChildUITab(groupID, bulkGridSetupToolUITab, false);
                        bulkGridSetupToolUITab.RefreshGroupUI(LeftRight.both, 1);
                        break;

                }
            }
        }
        private void SettingsTabSelected(int tab)
        {
            currentToolsTab = tab;
            RefreshGroupUI(LeftRight.both, 2);
        }
    }

    public class HeelAdjustToolUITab : BaseUIHelpTab
    {
        private JSONStorableStringChooser shoeClothingItemSelectorJSSC;

        public HeelAdjustToolUITab()
        {
            uiGroupCount = 2;
            defaultHelpText = HeelAdjustToolUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;

            shoeClothingItemSelectorJSSC = new JSONStorableStringChooser("shoeClothingItemSelector", null, "Select...", "HeelAdjust Shoe Settings", HeelAdjustTool.SwitchSelectedClothingItem);
        }


        private void RefreshShoeSelectorChoices()
        {
            shoeClothingItemSelectorJSSC.choices = HeelAdjustTool.GetActiveShoeNameList();
        }


        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                if (UIAGlobals.isPrimaryPlugin)
                {
                    CreateToggle(groupID, HeelAdjustTool.heelAdjustActiveJSB, false, "Auto apply Heel Adjusts",helpText: HeelAdjustToolUIHelp.haActive);
                    CreateToggle(groupID, HeelAdjustTool.heelAdjustRaisePeopleJSB, false, "Auto raise People by Heel Height", helpText: HeelAdjustToolUIHelp.haRaisePeople);
                    CreatePopup(groupID, shoeClothingItemSelectorJSSC, false, 600f, RefreshShoeSelectorChoices, helpText: HeelAdjustToolUIHelp.shoeSelector);
                    ForceUIRefreshOnValChange(shoeClothingItemSelectorJSSC, LeftRight.left, 2);
                }
                else CreateTextField(groupID, new JSONStorableString("temp","Heel Adjustments can only be applied in the Primary plugin of a dual UIAssit plugin setup."));

            }
            else if (groupID == 2 && !isRight && UIAGlobals.isPrimaryPlugin)
            {
                if (shoeClothingItemSelectorJSSC.val != "Select...")
                {
                    CreateToggle(groupID, HeelAdjustTool._heelCollidersVisibleJSON, helpText: HeelAdjustToolUIHelp.heelColidersVisible);
                    CreateSlider(groupID, HeelAdjustTool._footRotationXJSON, helpText: HeelAdjustToolUIHelp.footRotationX);
                    CreateSlider(groupID, HeelAdjustTool._toeRotationXJSON, helpText: HeelAdjustToolUIHelp.toeRotationX);
                    CreateSlider(groupID, HeelAdjustTool._heelColliderHeightJSON, helpText: HeelAdjustToolUIHelp.heelHeight);
                    CreateSlider(groupID, HeelAdjustTool._heelColliderBackOffsetJSON, helpText: HeelAdjustToolUIHelp.heelBackOffset);
                    CreateSlider(groupID, HeelAdjustTool._toeColliderHeightJSON, helpText: HeelAdjustToolUIHelp.toeHeight);
                    CreateSlider(groupID, HeelAdjustTool._toeColliderOffsetJSON, helpText: HeelAdjustToolUIHelp.toeOffset);
                    CreateSlider(groupID, HeelAdjustTool._toeColliderLengthJSON, helpText: HeelAdjustToolUIHelp.toeLength);
                }
            }
        }
    }

    public class ForceVAMModesToolUITab : BaseUIHelpTab
    {
        public ForceVAMModesToolUITab()
        {
            uiGroupCount = 2;
            defaultHelpText = ForceVAMModesToolUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                CreateToggle(groupID, ForceVAMModesTool.forceEditModeOpenUIJSB, false, "Edit Mode on VAM UI Open", helpText: ForceVAMModesToolUIHelp.forceEditModeOpen);
                CreateToggle(groupID, ForceVAMModesTool.forcePlayModeCloseUIJSB, false, "Play Mode on VAM UI Close", helpText: ForceVAMModesToolUIHelp.forcePlayModeClose);
                CreateSpacer(groupID);

                CreateToggle(groupID, ForceVAMModesTool.forceEditModeLoadSceneJSB, false, "Edit Mode on Scene Load", helpText: ForceVAMModesToolUIHelp.forceEditModeLoadScene);
                CreateSpacer(groupID);

                CreateToggle(groupID, ForceVAMModesTool.forceVSyncOffInVRJSB, false, "'VSync Off' at VR startup", helpText: ForceVAMModesToolUIHelp.forceVSyncOffInVR);
                CreateToggle(groupID, ForceVAMModesTool.forceVSyncOnInDesktopJSB, false, "'VSync On' at Desktop startup", helpText: ForceVAMModesToolUIHelp.forceVSyncOnInDesktop);
                CreateSpacer(groupID);

                CreateToggle(groupID, ForceVAMModesTool.forceDisableUIOnTopInVRJSB, false, "Disable 'UI On Top' at VR Startup", helpText: ForceVAMModesToolUIHelp.forceDisableUIOnTopInVR);
                CreateToggle(groupID, ForceVAMModesTool.forceUIOnTopInDesktopJSB, false, "Enable 'UI On Top' at Desktop Startup",80, helpText: ForceVAMModesToolUIHelp.forceUIOnTopInDesktop);
                CreateSpacer(groupID);

                CreateToggle(groupID, ForceVAMModesTool.forceDisableFreezePhysicsOnGrabNewPersonJSB, false, "Disable 'Freeze Physics while Grabbing' for new Person Atoms",80f, helpText: ForceVAMModesToolUIHelp.forceDisableFreezePhysicsOnGrabNewPerson);
                CreateToggle(groupID, ForceVAMModesTool.forceDisableFreezePhysicsOnGrabSceneLoadJSB, false, "Disable 'Freeze Physics while Grabbing' for all People on Scene Load", 100f, helpText: ForceVAMModesToolUIHelp.forceDisableFreezePhysicsOnGrabSceneLoad);
            }

        }
    }

    public class LeapMotionControlToolUITab : BaseUIHelpTab
    {
        public LeapMotionControlToolUITab()
        {
            uiGroupCount = 3;
            defaultHelpText = LeapMotionControlToolUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                CreateToggle(groupID, LeapMotionControlTool.leapMotionControlActiveJSB, false, "Enable Touch UI Control (VR Only)", helpText: LeapMotionControlToolUIHelp.leapMotionControlActive);
                ForceUIRefreshOnValChange(LeapMotionControlTool.leapMotionControlActiveJSB, LeftRight.left, 2);
            }
            else if (groupID == 2 && !isRight)
            {
                if (LeapMotionControlTool.leapMotionControlActiveJSB.val)
                {
                    CreatePopup(groupID, LeapMotionControlTool.leapMotionAutoEnableDisableJSE, labelWidth:325f, scrollable:false, helpText: LeapMotionControlToolUIHelp.leapMotionAutoEnableDisable);

                    CreateSpacer(groupID);
                    CreateToggle(groupID, LeapMotionControlTool.leapFreeMoveControlActiveJSB, false, "Enable Free Move with 'OK' Hand gesture", 80f, helpText: LeapMotionControlToolUIHelp.leapFreeMoveControlActive);
                    ForceUIRefreshOnValChange(LeapMotionControlTool.leapFreeMoveControlActiveJSB, LeftRight.left, 3);
                }
                
            }
            else if (groupID == 3 && !isRight)
            {
                if (LeapMotionControlTool.leapFreeMoveControlActiveJSB.val && LeapMotionControlTool.leapMotionControlActiveJSB.val)
                {
                    CreateToggle(groupID, LeapMotionControlTool.leapFreeMoveFemaleRotationActiveJSB, false, "Free Move Rotation around Last viewed Female", 80f, helpText: LeapMotionControlToolUIHelp.leapFreeMoveFemaleRotationActive);
                    CreateToggle(groupID, LeapMotionControlTool.leapFreeMoveMaleRotationActiveJSB, false, "Free Move Rotation around Last viewed Male", 80f, helpText: LeapMotionControlToolUIHelp.leapFreeMoveMaleRotationActive);
                }

            }
        }
    }
    public class BulkGridSetupToolUITab : BaseUIHelpTab
    {
        public BulkGridSetupToolUITab()
        {
            uiGroupCount = 2;
            defaultHelpText = BulkGridSetupToolUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                CreatePopup(groupID, BulkGridSetupTool.buttonsTypeJSEnum, false, 400, null,"New Buttons Operation", false, BulkGridSetupToolUIHelp.buttonsType,175);
                ForceUIRefreshOnValChange(BulkGridSetupTool.buttonsTypeJSEnum, LeftRight.left, 2);
                CreatePopup(groupID, BulkGridSetupTool.buttonsTargetTypeJSEnum, false, 400, null, "New Buttons Target", false, BulkGridSetupToolUIHelp.buttonsTargetType, 175);
                CreateToggle(groupID, BulkGridSetupTool.useMergeLoadJSB, false, "Use Merge Load", helpText: BulkGridSetupToolUIHelp.useMergeLoad);
            }
            else if (groupID == 2 && !isRight)
            {
                if (BulkGridSetupTool.buttonsTypeJSEnum.val== UIAButtonOpType.loadAppPreset)
                {
                    CreateToggle(groupID, BulkGridSetupTool.suppressClothingLoadJSB, false, "Suppress Clothing Load", helpText: BulkGridSetupToolUIHelp.suppressClothingLoad);
                    if (PatreonFeatures.patreonContentEnabled) CreateToggle(groupID, BulkGridSetupTool.suppressPersonScaleLoadJSB, false, "Suppress Person Scale Load", helpText: BulkGridSetupToolUIHelp.suppressPersonScaleLoad);
                }
                CreateButton(groupID, "Select folder and create new Grid containing all Presets from folder...",false,LaunchFolderSelection,80f, helpText: BulkGridSetupToolUIHelp.launchFolderSelection);
            }
        }

        private void LaunchFolderSelection()
        {
            string presetFolder = "";
            switch (BulkGridSetupTool.buttonsTypeJSEnum.val)
            {
                case UIAButtonOpType.loadAppPreset:
                    presetFolder = FileReferenceTypes.GetDefaultFolder(FileReferenceTypes.appearancePreset);
                    break;
                case UIAButtonOpType.loadPosePreset:
                    presetFolder = FileReferenceTypes.GetDefaultFolder(FileReferenceTypes.posePreset);
                    break;
                case UIAButtonOpType.loadHairPreset:
                    presetFolder = FileReferenceTypes.GetDefaultFolder(FileReferenceTypes.hairPreset);
                    break;
                case UIAButtonOpType.loadClothPreset:
                    presetFolder = FileReferenceTypes.GetDefaultFolder(FileReferenceTypes.clothingPreset);
                    break;
            }

            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(presetFolder, false, false, true);
            SuperController.singleton.GetDirectoryPathDialog(BulkGridSetupTool.FolderSelected, presetFolder, shortcuts, false);
        }
    }
    
    public class SaveProfileUITab : BaseUIHelpTab
    {
        private bool saveDefaultConfirmationMode = false;
        private CustomRowUIElement confirmDefaultOverwriteUI;
        public SaveProfileUITab()
        {
            uiGroupCount = 2;
            defaultHelpText = SaveProfileUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
            confirmDefaultOverwriteUI = new CustomRowUIElement(80, false);
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID,1130);
            else if (groupID == 1 && !isRight)
            {
                CreateButton(groupID, "Save UIA Profile...", false, GetSaveUIAPFile, helpText: SaveProfileUIHelp.getSaveUIAPFile);
                CreateSpacer(groupID,100);

                
            }
            if (groupID == 2 && !isRight)
            {
                if (!saveDefaultConfirmationMode)
                {
                    string defaultType = "Save current Profile as UIAP Default...";
                    int buttonHeight = 50;
                    if (!UIAGlobals.isPrimaryPlugin)
                    {
                        defaultType = "Save current Profile as Secondary UIAP Default...";
                        buttonHeight = 80;
                    }
                    CreateButton(groupID, defaultType, false, ActivateSaveDefaultConfirmMode, buttonHeight, helpText: SaveProfileUIHelp.getSaveUIAPFile);
                }
                else
                {
                    confirmDefaultOverwriteUI.InitUI(2, true, false);
                    var button = confirmDefaultOverwriteUI.AddButtonToRow( "Confirm overwrite default Profile",1, SaveDefaultUIAPConfirmation);
                    RegisterContextualHelp(button.gameObject, SaveProfileUIHelp.saveDefaultUIAPConfirmation);
                    button.buttonColor = Color.red;
                    button.textColor = Color.white;
                    button= confirmDefaultOverwriteUI.AddButtonToRow("Cancel default Profile save",2, SaveDefaultUIAPConfirmation);
                    RegisterContextualHelp(button.gameObject, SaveProfileUIHelp.cancelSaveDefaultUIAPConfirmation);
                    RegisterCustomUIElement(groupID,confirmDefaultOverwriteUI);
                    button.buttonColor = Color.green;
                }
                
            }
        }

        private void ActivateSaveDefaultConfirmMode()
        {
            saveDefaultConfirmationMode = true;
            currentHelpTextJSS.val = defaultHelpText;
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void SaveDefaultUIAPConfirmation(int buttonRef)
        {
            if (buttonRef==1) UIAPProfile.SaveDefaultUIAP();
            saveDefaultConfirmationMode = false;
            currentHelpTextJSS.val = defaultHelpText;
            RefreshGroupUI(LeftRight.left, 2);
        }
        private void GetSaveUIAPFile()
        {           
            SuperController.singleton.GetMediaPathDialog(UIAPProfile.SaveUIAPSelected, "uiap", FileUtils.GetLocalUIAPPath(), false, true, false);
            // Update the browser to be a Save browser
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), "uiap");
            browser.ActivateFileNameField();
        }


    }

    public class LoadProfileUITab : BaseUIHelpTab
    {
        public LoadProfileUITab()
        {
            uiGroupCount = 3;
            defaultHelpText = LoadProfileUIHelp.defaultHelpText;
            currentHelpTextJSS.val = defaultHelpText;
        }

        protected override void CreateGroupUI(bool isRight, int groupID)
        {
            if (isRight) CreateHelpTextUI(groupID);
            else if (groupID == 1 && !isRight)
            {
                CreateButton(groupID, "Load UIA Profile...",false,GetLoadUIAPFile, helpText: LoadProfileUIHelp.getLoadUIAPFile);
                CreateSpacer(groupID, 100);
                CreateButton(groupID, "Import Grid from UIA Profile...", false, GetImportGridUIAP, helpText: LoadProfileUIHelp.getImportGridUIAP);
            }
            else if (groupID == 2 && !isRight)
            {
                if (ImportUIAPGridTool._uiapFileNameJSON.val != "")
                {
                    CreateTextField(groupID, ImportUIAPGridTool._uiapFileNameJSON, false, 50, helpText: LoadProfileUIHelp.uiapFileName);
                    CreatePopup(groupID, ImportUIAPGridTool._uiapGridOptionsJSON, false, 400, null, "Select Grid to Import", helpText: LoadProfileUIHelp.uiapGridOptions);

                    ForceUIRefreshOnValChange(ImportUIAPGridTool._uiapGridOptionsJSON, LeftRight.left, 3);
                }
            }
            else if (groupID == 3 && !isRight)
            {
                if (ImportUIAPGridTool._uiapFileNameJSON.val != "")
                {
                    if (ImportUIAPGridTool._uiapGridOptionsJSON.val == "Quick Launch Buttons")
                    {
                        var uidButton = CreateButton(groupID, "Overwrite Quick Launch Buttons", false, ImportUIAPGridTool.ImportQLBGrid, helpText: LoadProfileUIHelp.importQLBGrid);
                        uidButton.buttonColor = Color.red;
                        uidButton.textColor = Color.white;
                    }
                    else CreateButton(groupID, "Import Grid", false, ImportUIAPGridTool.ImportGrid, helpText: LoadProfileUIHelp.importGrid);
                }

            }
        }


        private void GetImportGridUIAP()
        {
            SuperController.singleton.GetMediaPathDialog(ImportGridUIAPFileSelected, "uiap", UIAConsts._PluginDataSubfolderName, false, true, false, null, false, FileUtils.GetUIAPShortcutFolders());
        }

        private void ImportGridUIAPFileSelected(string fileName)
        {
            ImportUIAPGridTool._uiapFileNameJSON.val = fileName;
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void GetLoadUIAPFile()
        {
            SuperController.singleton.GetMediaPathDialog(UIAPProfile.LoadUIAPSelected, "uiap", FileUtils.GetLocalUIAPPath(), false, true, false, null, false, FileUtils.GetUIAPShortcutFolders());
        }
    }
}
